#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/kmod.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/vmalloc.h>
#include <linux/mman.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#include <linux/jiffies.h>
#include <linux/workqueue.h>
#include <linux/moduleparam.h>

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#endif
#include "dahdi/kernel.h"

#include "cdcx522apci.h"

static int debug = 0;
static int NDsp = 0;
static int DisalbeAllCh = 0;

static struct shcdc_desc shcdc_des[3] = {{ "CDC-3522A-PCI", 0 }, { "CDC-2522A-PCI", 0 }, { "CDC-1522A-PCI", 0 }};

static struct shcdc *ifaces[WC_MAX_IFACES];

static struct dahdi_transcoder *uencode;
static struct dahdi_transcoder *udecode;
static struct cdc_state *encoders;
static struct cdc_state *decoders;

static int CountREC = 0;
static int CountTRI = 0;

static int IntCount = 0;
static int IntCount1 = 0;

static void plx_dmasgl_transfer_complete(struct shcdc *wc, unsigned char dmaindex);

void PrintDebugMsg(struct shcdc *wc, int ndsp, int num)
{
#if 1
	printk("[%d, %d], ", num, ndsp);
	printk("write=%d, ", wc->tmpwritechunk_offset[ndsp]);
	printk("Swrite=%d, ", wc->Stmpwritechunk_offset[ndsp]);
	printk("pQ0Rw=%d, ", wc->pQ0Rw[ndsp]);
	printk("pQ0Rr=%d\n", wc->pQ0Rr[ndsp]);
	printk("pQ0Tw=%d, ", wc->pQ0Tw[ndsp]);
	printk("pQ0Tr=%d\n", wc->pQ0Tr[ndsp]);
#endif
}

//static inline int zt_tc_sanitycheck(struct zt_transcode_header *zth, unsigned int outbytes);//deleted.................

static inline int pci_getconfig_dword(struct shcdc *wc, int offset, u32 *val)
{
	return pci_read_config_dword(wc->dev, offset, val);
}

static inline int pci_setconfig_dword(struct shcdc *wc, int offset, u32 val)
{
	return pci_write_config_dword(wc->dev, offset, val);
}

static inline unsigned long plx_getreg(struct shcdc *wc, unsigned long offset)
{
	return	readl(wc->plx + offset);
}

static inline void plx_setreg(struct shcdc *wc, unsigned long offset, unsigned long val)
{
	writel(val, wc->plx + offset);
}

static inline void shcdc_wait_ms(long ms)
{
	udelay(ms);
}

static unsigned short dsp_getreg_word(struct shcdc *wc, int offset)
{
	return readw(wc->dsp + offset);
}

static void dsp_setreg_word(struct shcdc *wc, int offset, unsigned short val)
{
	writew(val, wc->dsp + offset);
}

static int shcdc_getdsp_data(struct shcdc *wc, int dspsel, unsigned short addr, unsigned short *data, unsigned long len)
{
	unsigned long i;

	for (i = 0; i < len; i++, addr += 2)
	{
		*(data + i) = dsp_getreg_word(wc, addr);
	}

	return 0;
}

static int shcdc_setdsp_data(struct shcdc *wc, int dspsel, unsigned short addr, unsigned short *data, unsigned long len)
{
	unsigned long i;

	for (i = 0; i < len; i++, addr += 2)
	{
		dsp_setreg_word(wc, addr, *(data + i));
	}
	return 0;

}

static void plx_sync_reg_modify(struct plx_reg_data *reg)
{
	unsigned long regval;

	regval = plx_getreg(reg->wc, reg->offset);
	regval |= reg->bitstoset;
	regval &= ~(reg->bitstoclear);
	plx_setreg(reg->wc, reg->offset, regval);

}

static void shcdc_enable_pci_interrupts(struct shcdc *wc)
{
	struct plx_reg_data regdata;
	unsigned long regval;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoset = (1 << 8);
	regdata.bitstoclear = 0;

	plx_sync_reg_modify(&regdata);

	regval = plx_getreg(regdata.wc, regdata.offset);

}

static void shcdc_disable_pci_interrupts(struct shcdc *wc)
{
	struct plx_reg_data regdata;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoset = 0;
	regdata.bitstoclear = (1 << 8);

	plx_sync_reg_modify(&regdata);

}

static void shcdc_enable_local_interrupts(struct shcdc *wc)
{
	struct plx_reg_data regdata;
	unsigned long regval;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoset = (1 << 11);
	regdata.bitstoclear = 0;

	plx_sync_reg_modify(&regdata);

	regval = plx_getreg(regdata.wc, regdata.offset);

}

static void shcdc_disable_local_interrupts(struct shcdc *wc)
{
	struct plx_reg_data regdata;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoset = 0;
	regdata.bitstoclear = (1 << 11);

	plx_sync_reg_modify(&regdata);

}

static void shcdc_enable_cpld_interrupts(struct shcdc *wc)
{
	unsigned short val = 1;

	shcdc_setdsp_data(wc, 0, 0x48, &val, sizeof(val) / sizeof(unsigned short));
}

static void shcdc_disable_cpld_interrupts(struct shcdc *wc)
{
	unsigned short val = 0;

	shcdc_setdsp_data(wc, 0, 0x48, &val, sizeof(val) / sizeof(unsigned short));
}

static unsigned long plx_chip_type_get(struct shcdc *wc)
{// trunk/Driver/HWSYS/PCIcPCI/platform/linux/CommonApi.c	PlxChipTypeGet
	unsigned long val;
	unsigned long type;
	unsigned long reg_original;

	reg_original = plx_getreg(wc, 0x04);
	plx_setreg(wc, 0x4, reg_original | (0xff << 24));
	val = plx_getreg(wc, 0x4);
	plx_setreg(wc, 0x4, reg_original);

	if ((val >> 24) == 0) {
		printk("shcdc: unkown chip type\n");
		return -1;
	}

	if ((val >> 28) == 0xf) {
		val = plx_getreg(wc, 0x70);
		if (val == 0x905410b5) {
			type = 0x9054;
			return type;
		}
	}
	return -1;

}


static void *plx_dmabuf_alloc(struct shcdc *wc, struct plx_phys_mem_object *memobj)
{
	dma_addr_t busaddress;
	unsigned long virt_addr;

	memobj->kernelva = dma_alloc_coherent(&(wc->dev->dev), memobj->size, &busaddress, GFP_KERNEL | __GFP_NOWARN);

	if (!memobj->kernelva)
		return NULL;

	memobj->busphysical = (u64)busaddress;

	for (virt_addr = (unsigned long)memobj->kernelva; virt_addr < ((unsigned long)memobj->kernelva + memobj->size); virt_addr += PAGE_SIZE)
	{
		SetPageReserved(virt_to_page(virt_addr));
	}

	memobj->cpuphysical = virt_to_phys(memobj->kernelva);
	if (debug)
		printk("cpu phys addr: 0x%08lx, bus phys addr: 0x%08lx, kernel VA : 0x%p\n", (unsigned long)memobj->cpuphysical, (unsigned long)memobj->busphysical, memobj->kernelva);

	if (memobj->size >= (10 << 10))
	{
		if (debug)
			printk("(%d kb)\n", memobj->size >> 10);
	}
	else
	{
		if (debug)
			printk("(%d bytes)\n", memobj->size);
	}

	return memobj->kernelva;
}

static void plx_dmabuf_free(struct shcdc *wc, struct plx_phys_mem_object *memobj)
{
	unsigned long virt_addr;

	for (virt_addr = (unsigned long)memobj->kernelva; virt_addr < ((unsigned long)memobj->kernelva + PAGE_ALIGN(memobj->size)); virt_addr += PAGE_SIZE)
	{
		ClearPageReserved(virt_to_page(virt_addr));
	}
	dma_free_coherent(&(wc->dev->dev), memobj->size, memobj->kernelva, (int)memobj->busphysical);
	if (debug)
		printk("released physical memory at 0x%08lx\n", (unsigned long) memobj->cpuphysical);

	if (memobj->size >= (10 << 10))
	{
		if (debug)
			printk("(%d kb)\n", memobj->size >> 10);
	}
	else
	{
		if (debug)
			printk("(%d bytes)\n", memobj->size);
	}

	memset(memobj, 0x0, sizeof(struct plx_phys_mem_object/*unsigned long*/));

}

static int plx_dmasgl_channnel_open(struct shcdc *wc, DMA_CHANNEL channel, struct dma_channel_desc *desc, void *owner)
{
	u8 i;
	u32 mode;
	u32 regvalue;
	u32 threshold;
	struct plx_reg_data regdata;


	switch (channel) {
		case DMA_CHANNEL0:
			i = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			break;
		default:
			printk("open:Invalid DMA channel\n");
			return -1;

	}

	if (wc->dmainfo[i].state != DMA_STATE_CLOSED)
	{
		printk("DMA channel already open\n");
		return -1;
	}

	wc->dmainfo[i].state = DMA_STATE_SGL;
	wc->dmainfo[i].owner = owner;
	wc->dmainfo[i].pending = 0;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoclear = 0;

	regvalue = plx_getreg(wc, PCI9054_LOCAL_DMA_ARBIT);
	regvalue &= ~((1 << 20) | (1 << 19));

	switch (desc->dmachannelpriority) {
		case CHANNEL_0_HIGHEST:
			plx_setreg(wc, PCI9054_LOCAL_DMA_ARBIT, regvalue | (1 << 19));
			break;
		case CHANNEL_1_HIGHEST:
			plx_setreg(wc, PCI9054_LOCAL_DMA_ARBIT, regvalue | (1 << 20));
			break;
		case ROTATIONAL:
			plx_setreg(wc, PCI9054_LOCAL_DMA_ARBIT, regvalue);
			break;
		default:
			printk("dmachannelopen() invalid priority state\n");
			return -1;
	}

	threshold = 
			(desc->tholdforiopwrites << 0) |
			(desc->tholdforiopreads << 4) |
			(desc->tholdforpciwrites << 8) |
			(desc->tholdforpcireads << 12);
	mode =
			(1 << 9) |
			(1 << 10) |
			(1 << 17) |
			(0 << 18) |
			(desc->iopbuswidth << 0) |
			(desc->waitstates << 2) |
			(desc->enablereadyinput << 6) |
			(desc->enablebterminput << 7) |
			(desc->enableiopburst << 8) |
			(desc->holdiopaddrconst << 11) |
			(desc->demandmode << 12) |		//9054 the DMA1 is reserved
			(desc->enablewriteinvalidmode << 13) |
			(desc->enabledmaeoppin << 14) |
			(desc->dmastoptransfermode << 15);
	if (desc->holdiopaddrconst)
		wc->dmainfo[i].localaddrconstant = 1;
	else 
		wc->dmainfo[i].localaddrconstant = 0;

	regvalue = plx_getreg(wc, PCI9054_DMA_THRESHOLD);
	if (channel == DMA_CHANNEL0)
	{
		plx_setreg(wc, PCI9054_DMA_THRESHOLD, (regvalue & 0xffff0000) | threshold);
		plx_setreg(wc, PCI9054_DMA0_MODE, mode);
		regdata.bitstoset = (1 << 18);
		plx_sync_reg_modify(&regdata);
	}
	else
	{
		plx_setreg(wc, PCI9054_DMA_THRESHOLD, (regvalue & 0xffff0000) | (threshold << 16));
		plx_setreg(wc, PCI9054_DMA1_MODE, mode);
		regdata.bitstoset = (1 << 19);
		plx_sync_reg_modify(&regdata);
		//plx_setreg(wc, PCI9054_DMA1_PCI_DAC, 0);
	}


	return 0;
}

static int plx_dmasgl_channel_close(struct shcdc *wc, DMA_CHANNEL channel, unsigned int checkinprogress)
{
	u8 i;
	u32 regvalue;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			break;
		default:
			printk("close:Invalid DMA channel\n");
			return -1;
	}

	if (checkinprogress)
	{
		regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
		if (channel == DMA_CHANNEL1)
			regvalue = regvalue >> 8;

		if ((regvalue & (1 << 4)) == 0)
		{
			if (regvalue & (1 << 0))
			{
				printk("dma in progress\n");
				return -1;
			}
			else
			{
				printk("dma paused\n");
				return -1;
			}
		}
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
	{
		printk("--DMA channel has not been opened for SGL DMA\n");
		return -1;
	}

	wc->dmainfo[i].state = DMA_STATE_CLOSED;
	wc->dmainfo[i].owner = NULL;
	if (wc->dmainfo[i].pending)
	{
		plx_dmasgl_transfer_complete(wc, i);
	}

	if (wc->dmainfo[i].sglbuffer.kernelva)
	{
		if (debug)
			printk("channel close: releasing memory used for SGL descript...\n");
		plx_dmabuf_free(wc, &wc->dmainfo[i].sglbuffer);
	}
	if (wc->dmainfo[i].dspbuffer.kernelva)
	{
		if (debug)
			printk("channel close: releasing memory used for DSP descript...\n");
		plx_dmabuf_free(wc, &wc->dmainfo[i].dspbuffer);
	}
	return 0;

}

static int plx_dmasgl_transfer_listadd(struct shcdc *wc, DMA_CHANNEL channel, struct dma_transfer_element *dmadata)
{
	u8 i;
	u8 shift;
	u32 regvalue;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			shift = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			shift = 8;
			break;
		default:
			printk("listadd:Invalid DMA channel\n");
			return -1;

	}

	regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
	if ((regvalue & ((1 << 4) << shift)) == 0)
	{
		printk("listadd: dma channel is carrently active\n");
		return -1;
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
	{
		printk("listadd: dma channel has not been opened for SGL DMA\n");
		return -1;
	}

	if (wc->dmainfo[i].pending)
	{
		printk("listadd: a dma transfer is currently in progress\n");
		return -1;
	}

	wc->dmainfo[i].dmadata[wc->dmainfo[i].dmadatacount] = *dmadata;
	
	wc->dmainfo[i].dmadatacount++;

	return 0;
}

static int plx_dmasgl_lockbuffer_and_build(struct shcdc *wc, u8 dmaindex)
{
	u32 k = 0;
	u32 bussgl;
	u32 bussgloriginal;
	u32 sglsize;
	u32 totalpages;

	unsigned long vasgl;
	struct dma_transfer_element *dma;

	totalpages = wc->dmainfo[dmaindex].dmadatacount;

	sglsize = (totalpages * (4 * sizeof(u32))) + 16;

	if (wc->dmainfo[dmaindex].sglbuffer.kernelva)
	{
		if (wc->dmainfo[dmaindex].sglbuffer.size < sglsize)
		{
			printk("err:sglbuffer.size < sglsize return line %d\n",__LINE__);
				return 0;
		}
	}

	if (!wc->dmainfo[dmaindex].sglbuffer.kernelva)
	{
		printk("lockbuf: unable to allocate %d bytes for %d SGL descriptors\n", wc->dmainfo[dmaindex].sglbuffer.size, totalpages);
		return 0;
	}
	else
		vasgl = (unsigned long)wc->dmainfo[dmaindex].sglbuffer.kernelva;



	bussgl = (u32)wc->dmainfo[dmaindex].sglbuffer.busphysical;

	vasgl = (vasgl + 0xf) & ~(0xf);

	bussgl = (bussgl + 0xf) & ~(0xf);

	bussgloriginal = bussgl;


	for (k = 0; k < wc->dmainfo[dmaindex].dmadatacount; k++)
	{
		dma = &wc->dmainfo[dmaindex].dmadata[k];

		totalpages--;
		*(((u32*)vasgl) + 0) = dma->u.pciaddrlow;
		*(((u32*)vasgl) + 1) = dma->localaddr;
		*(((u32*)vasgl) + 2) = dma->transfercount;
		if (totalpages == 0)
		{
			*(((u32*)vasgl) + 3) = (dma->localtopcidma << 3) | (1 << 1) | (1 << 0);
		}
		else
		{
			bussgl += (4 * sizeof(u32));
			*(((u32*)vasgl) + 3) = bussgl | (dma->localtopcidma << 3) | (1 << 0);
			vasgl +=(4 * sizeof(u32));
		}

	}
	return bussgloriginal;
}

static int plx_dmasgl_transfer(struct shcdc *wc, DMA_CHANNEL channel)
{
	u8 i;
	u8 shift;
	u16 offsetdesc;
	u32 regvalue;
	u32 sglpciaddress;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			shift = 0;
			offsetdesc = PCI9054_DMA0_DESC_PTR;
			break;
		case DMA_CHANNEL1:
			i = 1;
			shift = 8;
			offsetdesc = PCI9054_DMA1_DESC_PTR;
			break;
		default:
			printk("dmatransfer:Invalid DMA channel\n");
			return -1;

	}

	regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
	if ((regvalue & ((1 << 4) << shift)) == 0)
	{
		printk("dmatransfer: dma channel is carrently active\n");
		return -1;
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
	{
		printk("dmatransfer: dma channel has not been opened for SGL DMA\n");
		return -1;
	}

	if (wc->dmainfo[i].pending)
	{
		printk("dmatransfer: a dma transfer is currently in progress\n");
		return -1;
	}

	sglpciaddress = plx_dmasgl_lockbuffer_and_build(wc, i);
	
	if (sglpciaddress == 0)
	{
		printk("unable to lock buffer and build SGL list\n");
		wc->dmainfo[i].pending = 0;
		return -1;
	}

	plx_setreg(wc, offsetdesc, sglpciaddress | (1 << 0));
	regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
	plx_setreg(wc, PCI9054_DMA_COMMAND_STAT ,regvalue | ((1 << 0) << shift));	//enable channel to transfer data

	plx_setreg(wc, PCI9054_DMA_COMMAND_STAT, regvalue | (((1 << 0) | (1 << 1)) << shift)); // start dma

	wc->dmainfo[i].pending = 1;
	wc->dmainfo[i].timeout = 0;

	return 0;
}

static void plx_dmasgl_transfer_complete(struct shcdc *wc, unsigned char dmaindex)
{
	if (wc->dmainfo[dmaindex].pending == 0)
	{
		printk("No pending SGL Dma to complete\n");
		return; 
	}

	wc->dmainfo[dmaindex].pending = 0;
	wc->dmainfo[dmaindex].dmadatacount = 0;
	wc->dmainfo[dmaindex].numpages = 0;

	if (wc->dmainfo[0].pending == 0 && wc->dmainfo[1].pending == 0)
	{
		shcdc_enable_cpld_interrupts(wc);
		shcdc_enable_local_interrupts(wc);
	}

	return;
}

static int shcdc_stop_dma(struct shcdc *wc)
{
	plx_dmasgl_channel_close(wc, DMA_CHANNEL0, 1);
	plx_dmasgl_channel_close(wc, DMA_CHANNEL1, 1);
	return 0;
}



static int shcdc_hostdmawrite(struct shcdc *wc, pDMA_PARAMS pDmaParams)
{
	unsigned short reg, i, Count;
	//unsigned short *pusWriteDataBuf;
	u32 reg32;

	struct dma_transfer_element el;

	reg32 = 0x00000001;
	plx_setreg(wc, 0x4, reg32);
	
	if(wc->BF527Init)
		Count = 8000;
	else
		Count = 3;

	/****************Test for Host DMA write*******************/

		//read HOST_STATUS register
		// 15  9  8  7  6  5  4  3  2  1  0---Reset=0x000c
		//	|  |  |  |  |  |  |  |  |  |  |---DMA ready
		//  |  |  |  |  |  |  |  |  |  |------FIFO full
		//  |  |  |  |  |  |  |  |  |---------FIFO empty
		//  |  |  |  |  |  |  |  |------------DMA complete
		//  |  |  |  |  |  |  |---------------Host handshake
		//  |  |  |  |  |  |------------------HOSTDP timeout
		//  |  |  |  |  |---------------------Host interrrupt Request
		//  |  |  |  |------------------------Allow Config
		//  |  |  |---------------------------DMA direction
		//  |  |------------------------------Bus Timeout
		//  |---------------------------------Host read done

	//poll allow config
	i = 0;
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

		i++;
		if(i>Count)
		{
			//printk("[%x]\n", reg);
			return DISALLOW_CONFIG_ERROR;
		}
	}while((reg & 0x0080) != 0x0080);

	//configure DMA
	//pusWriteDataBuf = (unsigned short *)wc->confbuffer.kernelva;
	//memcpy((void *)pusWriteDataBuf, (void *)pDmaParams, DMA_CONFIG_LENGTH);
	//pciblockdmawrite(wc, pDmaParams->PortAddr, DMA_CONFIG_LENGTH, pusWriteDataBuf);
	
	//configure BFA host dma port
	//1. configure HOST_CONFIGURE
	reg = pDmaParams->Config;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//2. configure Start low addres
	reg = pDmaParams->LowAddr;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//3. configure Start high address
	reg = pDmaParams->HighAddr;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//4. configure X-Counter
	reg = pDmaParams->XCount;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//5. configure X-Modify
	reg = pDmaParams->XModify;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//6. configure Y-Counter
	reg = pDmaParams->YCount;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//7. configure Y-Modify
	reg = pDmaParams->YModify;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//poll DMA ready
	i=0;
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

		i++;
		if(i>Count)
		{
			//printk("[%x]\n", reg);
			return DMA_NOT_READY_ERROR;
		}
	}while((reg & 0x0001)==0x0000);

	//write data
	//pciblockdmawrite(wc, pDmaParams->MemAddr, pDmaParams->XCount<<1, pDmaParams->DmaBusPhysical);
	
	el.localtopcidma = 0;
	el.localaddr = pDmaParams->MemAddr;
	el.u.pciaddrlow = pDmaParams->DmaBusPhysical;
	el.transfercount = pDmaParams->XCount<<1;
	plx_dmasgl_transfer_listadd(wc, DMA_CHANNEL0, &el);

	return 0;

}

static int shcdc_hostdmaread(struct shcdc *wc, pDMA_PARAMS pDmaParams)
{
	unsigned short reg, i, Count;
	//unsigned short *pusReadDataBuf;
	
	struct dma_transfer_element el;

	plx_setreg(wc, 0x4, 0x00000001);
	
	if(wc->BF527Init)
		Count = 8000;
	else
		Count = 3;

	/****************Test for Host DMA write*******************/
	//BFA_PORT_ADDR BFB_PORT_ADDR BFC_PORT_ADDR
		//read HOST_STATUS register
		// 15  9  8  7  6  5  4  3  2  1  0---Reset=0x000c
		//	|  |  |  |  |  |  |  |  |  |  |---DMA ready
		//  |  |  |  |  |  |  |  |  |  |------FIFO full
		//  |  |  |  |  |  |  |  |  |---------FIFO empty
		//  |  |  |  |  |  |  |  |------------DMA complete
		//  |  |  |  |  |  |  |---------------Host handshake
		//  |  |  |  |  |  |------------------HOSTDP timeout
		//  |  |  |  |  |---------------------Host interrrupt Request
		//  |  |  |  |------------------------Allow Config
		//  |  |  |---------------------------DMA direction
		//  |  |------------------------------Bus Timeout
		//  |---------------------------------Host read done

	//poll allow config
	i = 0;
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

		i++;
		if(i>Count)
		{
			//printk("[%x]\n", reg);
			return DISALLOW_CONFIG_ERROR;
		}
	}while((reg & 0x0080) != 0x0080);

	//configure DMA
	//pusReadDataBuf = (unsigned short *)wc->confbuffer.kernelva;
	//memcpy((void *)pusReadDataBuf, (void *)pDmaParams, DMA_CONFIG_LENGTH);
	//pciblockdmawrite(wc, pDmaParams->PortAddr, DMA_CONFIG_LENGTH, wc->confbuffer.busphysical);
	
	//configure BFA host dma port
	//1. configure HOST_CONFIGURE
	reg = pDmaParams->Config;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//2. configure Start low addres
	reg = pDmaParams->LowAddr;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//3. configure Start high address
	reg = pDmaParams->HighAddr;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//4. configure X-Counter
	reg = pDmaParams->XCount;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//5. configure X-Modify
	reg = pDmaParams->XModify;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//6. configure Y-Counter
	reg = pDmaParams->YCount;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//7. configure Y-Modify
	reg = pDmaParams->YModify;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//poll DMA ready
	i=0;
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

		i++;
		if(i>Count)
		{
			//printk("[%x]\n", reg);
			return DMA_NOT_READY_ERROR;
		}
	}while((reg & 0x0001)==0x0000);
	
	//read data
	//pciblockdmaread(wc, pDmaParams->MemAddr, pDmaParams->XCount<<1, pDmaParams->DmaBusPhysical);
	
	el.localtopcidma = 1;
	el.localaddr = pDmaParams->MemAddr;
	el.u.pciaddrlow = pDmaParams->DmaBusPhysical;
	el.transfercount = pDmaParams->XCount<<1;
	plx_dmasgl_transfer_listadd(wc, DMA_CHANNEL1, &el);

	return 0;

}

static int shcdc_hostwrite(struct shcdc *wc, pDMA_PARAMS pDmaParams)
{
	unsigned short reg, Count, i=0;
	unsigned short *pDataBuf;

	pDataBuf = pDmaParams->pData;

	plx_setreg(wc, 0x4, 0x00000001);
	
	if(wc->BF527Init)
		Count = 8000;
	else
		Count = 3;

	/****************Test for Host DMA write*******************/
	//poll allow config
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

		i++;
		if(i>Count)
		{
			printk("shcdc_hostwrite[%x]1\n", reg);
			return DISALLOW_CONFIG_ERROR;
		}
	}while((reg & 0x0080) != 0x0080);

	//printk("[%x]\n", reg);

	//configure BFA host dma port
	//1. configure HOST_CONFIGURE
	reg = pDmaParams->Config;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//2. configure Start low addres
	reg = pDmaParams->LowAddr;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//3. configure Start high address
	reg = pDmaParams->HighAddr;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//4. configure X-Counter
	reg = pDmaParams->XCount;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//5. configure X-Modify
	reg = pDmaParams->XModify;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//6. configure Y-Counter
	reg = pDmaParams->YCount;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//7. configure Y-Modify
	reg = pDmaParams->YModify;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));
	
	//printk("[%4x],[%4x],[%4x],[%4x],[%4x],[%4x],[%4x]\n", pDmaParams->Config, pDmaParams->LowAddr, pDmaParams->HighAddr, pDmaParams->XCount, pDmaParams->XModify, pDmaParams->YCount, pDmaParams->YModify);


	i=0;
	//poll DMA ready
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

		i++;
		if(i>Count)
		{
			printk("shcdc_hostwrite[%x]2\n", reg);
			return DMA_NOT_READY_ERROR;
		}
	}while((reg & 0x0001)==0x0000);

	if(wc->BF527Init)
		shcdc_wait_ms(10);

	//DMA is write data
	for(i=0;i<pDmaParams->XCount;i++)
	{
		reg = *(pDataBuf + i);
		shcdc_setdsp_data(wc, 0, pDmaParams->DataAddr, &reg, sizeof(reg) / sizeof(unsigned short));
		if(wc->BF527Init)
			shcdc_wait_ms(10);
	}

	i=0;
	//poll DMA complete
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));
		
		i++;
		if(i>Count)
		{
			printk("shcdc_hostwrite[%x]3\n", reg);
			return DMA_TIMEOUT_ERROR;
		}
	}while((reg & 0x8008) == 0x0000);

	return 0;

}

static int shcdc_hostread(struct shcdc *wc, pDMA_PARAMS pDmaParams)
{
	unsigned short	*pDataBuf;
	unsigned short reg, Count, i=0;

	pDataBuf = pDmaParams->pData;
	
	plx_setreg(wc, 0x4, 0x00000001);
	
	if(wc->BF527Init)
		Count = 8000;
	else
		Count = 3;

	/****************Test for Host DMA write*******************/
	//poll allow config
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

		i++;
		if(i>Count)
		{
			return DISALLOW_CONFIG_ERROR;
		}
	}while((reg & 0x0080) != 0x0080);

	
	//configure BFA host dma port
	//1. configure HOST_CONFIGURE
	reg = pDmaParams->Config;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//2. configure Start low addres
	reg = pDmaParams->LowAddr;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//3. configure Start high address
	reg = pDmaParams->HighAddr;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//4. configure X-Counter
	reg = pDmaParams->XCount;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//5. configure X-Modify
	reg = pDmaParams->XModify;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//6. configure Y-Counter
	reg = pDmaParams->YCount;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

	//7. configure Y-Modify
	reg = pDmaParams->YModify;
	shcdc_setdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));
	
	//printk("[%4x],[%4x],[%4x],[%4x],[%4x],[%4x],[%4x]\n", pDmaParams->Config, pDmaParams->LowAddr, pDmaParams->HighAddr, pDmaParams->XCount, pDmaParams->XModify, pDmaParams->YCount, pDmaParams->YModify);

	i=0;
	//poll DMA ready
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

		i++;
		if(i>Count)
		{
			return DMA_NOT_READY_ERROR;
		}
	}while((reg & 0x0001)==0x0000);


	for(i=0;i<pDmaParams->XCount;i++)
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->DataAddr, &reg, sizeof(reg) / sizeof(unsigned short));
		*(pDataBuf + i) = reg;
		//shcdc_wait_ms(10);
	}


	i=0;
	//poll DMA complete
	do
	{
		shcdc_getdsp_data(wc, 0, pDmaParams->PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));
		
		i++;
		if(i>Count)
		{
			printk("shcdc_hostread[%x]\n", reg);
			return DMA_TIMEOUT_ERROR;
		}
	}while((reg & 0x8008) == 0x0000);

	return 0;

}

static void shcdc_initqo(struct shcdc *wc)
{
	u32 reg32;
	DMA_PARAMS HostDmaParams;

	HostDmaParams.XModify = 0x2;
	HostDmaParams.YCount = 0x0;
	HostDmaParams.YModify = 0x0;
	
	//BFA
	reg32 = 0;
	HostDmaParams.PortAddr = BFA_PORT_ADDR;
	HostDmaParams.MemAddr = BFA_MEM_ADDR;
	HostDmaParams.DataAddr = BFA_DATA_ADDR;
	
	HostDmaParams.Config = DMA_WRITE_COM;
	HostDmaParams.XCount = 0x2;
	HostDmaParams.pData = (unsigned short *)&reg32;
	HostDmaParams.LowAddr = (unsigned short)QOR_W_ADDR;
	HostDmaParams.HighAddr = (unsigned short)(QOR_W_ADDR >> 16);
	shcdc_hostwrite(wc, &HostDmaParams);
	HostDmaParams.LowAddr = (unsigned short)QOT_W_ADDR;
	HostDmaParams.HighAddr = (unsigned short)(QOT_W_ADDR >> 16);
	shcdc_hostwrite(wc, &HostDmaParams);
	
	//BFB
	reg32 = 0;
	HostDmaParams.PortAddr = BFB_PORT_ADDR;
	HostDmaParams.MemAddr = BFB_MEM_ADDR;
	HostDmaParams.DataAddr = BFB_DATA_ADDR;
	
	HostDmaParams.Config = DMA_WRITE_COM;
	HostDmaParams.XCount = 0x2;
	HostDmaParams.pData = (unsigned short *)&reg32;
	HostDmaParams.LowAddr = (unsigned short)QOR_W_ADDR;
	HostDmaParams.HighAddr = (unsigned short)(QOR_W_ADDR >> 16);
	shcdc_hostwrite(wc, &HostDmaParams);
	HostDmaParams.LowAddr = (unsigned short)QOT_W_ADDR;
	HostDmaParams.HighAddr = (unsigned short)(QOT_W_ADDR >> 16);
	shcdc_hostwrite(wc, &HostDmaParams);
	
	//BFC
	reg32 = 0;
	HostDmaParams.PortAddr = BFC_PORT_ADDR;
	HostDmaParams.MemAddr = BFC_MEM_ADDR;
	HostDmaParams.DataAddr = BFC_DATA_ADDR;
	
	HostDmaParams.Config = DMA_WRITE_COM;
	HostDmaParams.XCount = 0x2;
	HostDmaParams.pData = (unsigned short *)&reg32;
	HostDmaParams.LowAddr = (unsigned short)QOR_W_ADDR;
	HostDmaParams.HighAddr = (unsigned short)(QOR_W_ADDR >> 16);
	shcdc_hostwrite(wc, &HostDmaParams);
	HostDmaParams.LowAddr = (unsigned short)QOT_W_ADDR;
	HostDmaParams.HighAddr = (unsigned short)(QOT_W_ADDR >> 16);
	shcdc_hostwrite(wc, &HostDmaParams);
}

static int shcdc_receive_data(struct shcdc *wc, int ndsp)
{
	int result = 0;
	u32 tmp = 0;
	unsigned short length;
	DMA_PARAMS pDmaParams;
	unsigned char *tmpchunk;
	
	switch(ndsp)
	{

		case 0:
		{
			pDmaParams.DataAddr = BFA_DATA_ADDR;
			pDmaParams.PortAddr = BFA_PORT_ADDR;
			pDmaParams.MemAddr = BFA_MEM_ADDR;
			pDmaParams.DmaBusPhysical = wc->dmainfo[1].dspbuffer.busphysical;
			tmpchunk = wc->readchunk[0];
			break;
		}
		case 1:
		{
			pDmaParams.DataAddr = BFB_DATA_ADDR;
			pDmaParams.PortAddr = BFB_PORT_ADDR;
			pDmaParams.MemAddr = BFB_MEM_ADDR;
			pDmaParams.DmaBusPhysical = wc->dmainfo[1].dspbuffer.busphysical + MAX_DMA_RW_SIZE;
			tmpchunk = wc->readchunk[1];
			break;
		}
		case 2:
		{
			pDmaParams.DataAddr = BFC_DATA_ADDR;
			pDmaParams.PortAddr = BFC_PORT_ADDR;
			pDmaParams.MemAddr = BFC_MEM_ADDR;
			pDmaParams.DmaBusPhysical = wc->dmainfo[1].dspbuffer.busphysical + MAX_DMA_RW_SIZE * 2;
			tmpchunk = wc->readchunk[2];
			break;
		}
		default:	return BLACKFIN_SELECT_ERROR;
	}
	
	pDmaParams.XModify = 0x2;
	pDmaParams.YCount = 0x0;
	pDmaParams.YModify = 0x0;

	
	pDmaParams.Config = DMA_READ_COM;
	pDmaParams.LowAddr = (unsigned short)QOT_W_ADDR;
	pDmaParams.HighAddr = (unsigned short)(QOT_W_ADDR >> 16);
	pDmaParams.XCount = 0x2;
	pDmaParams.pData = (unsigned short *)&tmp;
	
	result = shcdc_hostread(wc, &pDmaParams);
	if(result)
		return result;

	wc->pQ0Tw[ndsp] = (unsigned short)tmp;

	if(wc->pQ0Tr[ndsp] == wc->pQ0Tw[ndsp])
	{
		return 1;
	}

	//if(debug)
	//	printk("shcdc_receive_data \t[%d], [%d]\n", wc->pQ0Tr[ndsp], wc->pQ0Tw[ndsp]);
	
	if(wc->pQ0Tw[ndsp] < wc->pQ0Tr[ndsp])
		length = QORT_LEN - wc->pQ0Tr[ndsp] + 41;
	else
		length = wc->pQ0Tw[ndsp] - wc->pQ0Tr[ndsp];
	

	pDmaParams.Config = DMA_READ_COM;
	pDmaParams.LowAddr = (unsigned short)(Q0T_BASE_ADDR + wc->pQ0Tr[ndsp] * 2);
	pDmaParams.HighAddr = (unsigned short)((Q0T_BASE_ADDR + wc->pQ0Tr[ndsp] * 2)>>16);
	pDmaParams.XCount = length;
	
	result = shcdc_hostdmaread(wc, &pDmaParams);
	
	wc->EnableReadDsp[ndsp] = 1;

	return result;
}

static int shcdc_transmit_data(struct shcdc *wc, int ndsp)//maybe changed ......
{
	int result;
	u32 tmp = 0xffffffff;
	unsigned short code_len = wc->writechunk_offset[ndsp] >> 1;
	unsigned char *tmpchunk;
	
	DMA_PARAMS pDmaParams;
	
	switch(ndsp)
	{

		case 0:
		{
			pDmaParams.DataAddr = BFA_DATA_ADDR;
			pDmaParams.PortAddr = BFA_PORT_ADDR;
			pDmaParams.MemAddr = BFA_MEM_ADDR;
			pDmaParams.DmaBusPhysical = wc->dmainfo[0].dspbuffer.busphysical;
			tmpchunk = wc->writechunk[0];
			break;
		}
		case 1:
		{
			pDmaParams.DataAddr = BFB_DATA_ADDR;
			pDmaParams.PortAddr = BFB_PORT_ADDR;
			pDmaParams.MemAddr = BFB_MEM_ADDR;
			pDmaParams.DmaBusPhysical = wc->dmainfo[0].dspbuffer.busphysical + MAX_DMA_RW_SIZE;
			tmpchunk = wc->writechunk[1];
			break;
		}
		case 2:
		{
			pDmaParams.DataAddr = BFC_DATA_ADDR;
			pDmaParams.PortAddr = BFC_PORT_ADDR;
			pDmaParams.MemAddr = BFC_MEM_ADDR;
			pDmaParams.DmaBusPhysical = wc->dmainfo[0].dspbuffer.busphysical + MAX_DMA_RW_SIZE * 2;
			tmpchunk = wc->writechunk[2];
			break;
		}
		default:	return BLACKFIN_SELECT_ERROR;
	}
	
	result = 0;
	
	pDmaParams.XModify = 0x2;
	pDmaParams.YCount = 0x0;
	pDmaParams.YModify = 0x0;
	
	pDmaParams.Config = DMA_WRITE_COM;
	pDmaParams.LowAddr = (unsigned short)AU_FLAG_ADDR;
	pDmaParams.HighAddr = (unsigned short)(AU_FLAG_ADDR >> 16);
	pDmaParams.XCount = PER_CHANNEL;
	pDmaParams.pData = wc->dstfmt + ndsp*PER_CHANNEL;
	
	result = shcdc_hostwrite(wc, &pDmaParams);
	if(result)
		return result;
	
	
	pDmaParams.Config = DMA_READ_COM;
	pDmaParams.LowAddr = (unsigned short)QOR_W_ADDR;
	pDmaParams.HighAddr = (unsigned short)(QOR_W_ADDR >> 16);
	pDmaParams.XCount = 0x2;
	pDmaParams.pData = (unsigned short *)&tmp;
	
	result = shcdc_hostread(wc, &pDmaParams);
	if(result)
		return result;

	wc->pQ0Rr[ndsp] = (unsigned short)(tmp >> 16);
	
	pDmaParams.Config = DMA_WRITE_COM;
	pDmaParams.LowAddr = (unsigned short)(Q0R_BASE_ADDR + wc->pQ0Rw[ndsp] * 2);
	pDmaParams.HighAddr = (unsigned short)((Q0R_BASE_ADDR + wc->pQ0Rw[ndsp] * 2)>>16);

	pDmaParams.XCount = code_len;
	
#if 0
	if(code_len%41 != 0)
	{
		printk("tmpwritechunk_offset\t%d, %d, %d\n", wc->tmpwritechunk_offset[0], wc->tmpwritechunk_offset[1], wc->tmpwritechunk_offset[2]);
		printk("Stmpwritechunk_offset\t%d, %d, %d\n", wc->Stmpwritechunk_offset[0], wc->Stmpwritechunk_offset[1], wc->Stmpwritechunk_offset[2]);
		
		printk("pQ0Rw\t%d, %d, %d\n", wc->pQ0Rw[0], wc->pQ0Rw[1], wc->pQ0Rw[2]);
		printk("pQ0Rr\t%d, %d, %d\n", wc->pQ0Rr[0], wc->pQ0Rr[1], wc->pQ0Rr[2]);
		printk("pQ0Tw\t%d, %d, %d\n", wc->pQ0Tw[0], wc->pQ0Tw[1], wc->pQ0Tw[2]);
		printk("pQ0Tr\t%d, %d, %d\n", wc->pQ0Tr[0], wc->pQ0Tr[1], wc->pQ0Tr[2]);
	}
#endif
	
	result = shcdc_hostdmawrite(wc, &pDmaParams);
	if(result)
		return result;
	
	if((wc->pQ0Rw[ndsp] + code_len) < QORT_LEN)
		wc->pQ0Rw[ndsp] = wc->pQ0Rw[ndsp] + code_len;
	else
		wc->pQ0Rw[ndsp] = 0;
	
	CountTRI += code_len*2;
	
	wc->EnableWriteDsp[ndsp] = 1;
	
	return 0;
}

static int initbf527(struct shcdc *wc)
{
	struct file *fp = NULL;
	unsigned char * pusWriteDataBuf;
	u32 reg32;
	
	//'0'--unmask, '1'--mask;
	//bit0--BF527A, bit1--BF527B, bit2--BF527C
	unsigned short	MaskNum[3]={1,1,1};
	unsigned short	NumSel, i, usResult, Wdata, reg;
	unsigned long filelen, fileoffset;
	size_t read_len;
	
	DMA_PARAMS HostDmaParams;
	HEAD_INF HeadInf, *pHeadInf;
	
	pHeadInf = &HeadInf;

	//default
	HostDmaParams.XModify = 0x2;
	HostDmaParams.YCount = 0x0;
	HostDmaParams.YModify = 0x0;


	printk("\n************* Init Blackfin *************\n");

	plx_setreg(wc, 0x4, 0x00000001);

	//reset BF527
	reg = 0x0;
	shcdc_setdsp_data(wc, 0, BF_RESET_ADDR, &reg, sizeof(reg) / sizeof(unsigned short));
	
	shcdc_wait_ms(10000);
	
	reg = 0x1;
	shcdc_setdsp_data(wc, 0, BF_RESET_ADDR, &reg, sizeof(reg) / sizeof(unsigned short));
	
	shcdc_wait_ms(10000);

	for (NumSel=0; NumSel<wc->CurDspNum; NumSel++)
	{
		if (MaskNum[NumSel])
		{
			if (!(fp = filp_open(BINFILE_DIR, O_RDONLY, S_IRUSR)) || !IS_ERR(fp))
			{
				filelen = fp->f_dentry->d_inode->i_size;
				if (filelen < 1)
				{
					printk("file offset = NULL %s at line %d of %s\n",
					       __FILE__, __LINE__, __PRETTY_FUNCTION__);
					filp_close(fp , NULL);
					return -1;
				}
			}
			else
			{
				printk("Can't open file cdcx522apci.bin\n");
				return -1;
			}
			
			fileoffset = 0;
			
			switch(NumSel)
			{

				case 0:
				{
					HostDmaParams.Config = DMA_WRITE_COM;
					HostDmaParams.DataAddr = BFA_DATA_ADDR;
					HostDmaParams.PortAddr = BFA_PORT_ADDR;
					HostDmaParams.MemAddr = BFA_MEM_ADDR;
					break;
				}
				case 1:
				{
					HostDmaParams.Config = DMA_WRITE_COM;
					HostDmaParams.DataAddr = BFB_DATA_ADDR;
					HostDmaParams.PortAddr = BFB_PORT_ADDR;
					HostDmaParams.MemAddr = BFB_MEM_ADDR;
					break;
				}
				case 2:
				{
					HostDmaParams.Config = DMA_WRITE_COM;
					HostDmaParams.DataAddr = BFC_DATA_ADDR;
					HostDmaParams.PortAddr = BFC_PORT_ADDR;
					HostDmaParams.MemAddr = BFC_MEM_ADDR;
					break;
				}
				default:	return BLACKFIN_SELECT_ERROR;
			}//switch

			while(fileoffset != filelen)
			{
				read_len = kernel_read(fp, fileoffset, (char *)pHeadInf, sizeof(HEAD_INF));
				if(read_len != sizeof(HEAD_INF))
				{
					printk("Read File Length Error %s at line %d of %s\n",
					       __FILE__, __LINE__, __PRETTY_FUNCTION__);
					filp_close(fp , NULL);
					return -1;
				}
				
				fileoffset += read_len;

				//judge block code
				if(pHeadInf->ulBlockCode & 0x00001000)
				{   //block is ignore
					continue;
				}
				else if(pHeadInf->ulBlockCode & 0x00000100)
				{
				// block is fill
					if(pHeadInf->ulByteCount > 0x4000)
					{
						pusWriteDataBuf = (unsigned char *)kmalloc(0x4000, GFP_KERNEL);
						HostDmaParams.XCount = 0x2000;
						printk("No More than 128k can get use kmalloc[1]=>%x, %x, %x\n", pHeadInf->ulTargetAddr, pHeadInf->ulArgument, pHeadInf->ulByteCount);
					}
					else
					{
						pusWriteDataBuf = (unsigned char *)kmalloc(pHeadInf->ulByteCount, GFP_KERNEL);
						HostDmaParams.XCount = (unsigned short)(pHeadInf->ulByteCount >> 1);
					}
					
					HostDmaParams.LowAddr = (unsigned short)pHeadInf->ulTargetAddr;
					HostDmaParams.HighAddr = (unsigned short)(pHeadInf->ulTargetAddr>>16);
					
					for(i=0;i<HostDmaParams.XCount/2;i++)
						*((u32 *)pusWriteDataBuf + i) = pHeadInf->ulArgument;
					
					HostDmaParams.pData = (unsigned short *)pusWriteDataBuf;
					usResult = shcdc_hostwrite(wc, &HostDmaParams);
					
					kfree(pusWriteDataBuf);
					
					if(usResult)
					{
						printk("Host DMA error, error code = %x, %s at line %d of %s\n", usResult,
								__FILE__, __LINE__, __PRETTY_FUNCTION__);
						filp_close(fp , NULL);
						return usResult;
					}

				}
				else
				{//block is valid
					if (pHeadInf->ulByteCount)
					{
						if(pHeadInf->ulByteCount > 0x20000)	//128k
						{
							pusWriteDataBuf = (unsigned char *)kmalloc(0x20000, GFP_KERNEL);
							printk("No More than 128k can get use kmalloc[2]=>%x, %x, %x\n", pHeadInf->ulTargetAddr, pHeadInf->ulArgument, pHeadInf->ulByteCount);
							read_len = kernel_read(fp, fileoffset, (void *)pusWriteDataBuf, 0x20000);
						}
						else
						{
							pusWriteDataBuf = (unsigned char *)kmalloc(pHeadInf->ulByteCount, GFP_KERNEL);
							read_len = kernel_read(fp, fileoffset, (void *)pusWriteDataBuf, pHeadInf->ulByteCount);
						}

						if(pHeadInf->ulByteCount > 0x20000)
						{
							if(read_len != 0x20000)
							{
								printk("Read File Length Error, %s at line %d of %s\n",
										__FILE__, __LINE__, __PRETTY_FUNCTION__);
								filp_close(fp , NULL);
								kfree(pusWriteDataBuf);
								return -1;
							}
						}
						else if(read_len != pHeadInf->ulByteCount)
						{
							printk("Read File Length Error, %s at line %d of %s\n",
									__FILE__, __LINE__, __PRETTY_FUNCTION__);
							filp_close(fp , NULL);
							kfree(pusWriteDataBuf);
							return -1;
						}
						
						fileoffset += pHeadInf->ulByteCount;

						HostDmaParams.LowAddr = (unsigned short)pHeadInf->ulTargetAddr;
						HostDmaParams.HighAddr = (unsigned short)(pHeadInf->ulTargetAddr>>16);
						HostDmaParams.XCount = read_len>>1;
						HostDmaParams.pData = (unsigned short *)pusWriteDataBuf;
						usResult = shcdc_hostwrite(wc, &HostDmaParams);
						
						kfree(pusWriteDataBuf);
						if(usResult)
						{
							printk("Host DMA error, error code = %x, %s at line %d of %s\n", usResult,
									__FILE__, __LINE__, __PRETTY_FUNCTION__);
							filp_close(fp , NULL);
							return usResult;
						}
					}//end if count
					
					if(pHeadInf->ulBlockCode & 0x00008800)
					{
						////block is final or init,send HIRQ command
						reg = HOST_IRQ_COM;
						shcdc_setdsp_data(wc, 0, HostDmaParams.PortAddr, &reg, sizeof(reg) / sizeof(unsigned short));

						//wait for init or execute application
						i=0;
						do{
							shcdc_wait_ms(10);
							i++;
							if(i>8000)
							{
								filp_close(fp , NULL);
								return -1;
							}

							shcdc_getdsp_data(wc, 0, BF_HWAIT_ADDR, &reg, sizeof(reg) / sizeof(unsigned short));
							
							//if(i % 100 == 0)
								//printk("\t[%d]\n", reg);

						}while(((reg>>NumSel) & 0x0001) == 0x0001);
					}//end if
				}//end if
			}//end while
			
			filp_close(fp , NULL);
			shcdc_wait_ms(50);
			
			//read buf inf mapped in sdram address
			HostDmaParams.LowAddr = 0x10;
			HostDmaParams.HighAddr = BUF_INF_HIGH;
			HostDmaParams.XCount = 0x1;
			Wdata = 0x6666;
			HostDmaParams.pData = &Wdata;
			HostDmaParams.Config = DMA_WRITE_COM;
			usResult = shcdc_hostwrite(wc, &HostDmaParams);
			if(usResult)
			{
				printk("Host DMA Write error, error code = %x\n",usResult);
				return usResult;
			}
			
			HostDmaParams.pData = (unsigned short *)&reg32;
			HostDmaParams.Config = DMA_READ_COM;
			HostDmaParams.XCount = 0x2;
			usResult = shcdc_hostread(wc, &HostDmaParams);
			if(usResult)
			{
				printk("Host DMA Read error, error code = %x\n",usResult);
				return usResult;
			}

			if((unsigned short)reg32 != Wdata)
			{
				printk("No.%d BF527 is not running\n",NumSel);
				return -1;
			}
			printk("No.%d BF527 is running\n",NumSel);
		}//if
	}//for

	shcdc_initqo(wc);
	
	printk("Initialize Blackfin complete\n");
	return 0;
}

static int shcdc_plx_init(struct shcdc *wc)
{//init plx the chip type is 9054
	u8 mu_enabled;
	u8 eeprompresent;
	u32 regvalue;
	u32 reginterrupt;
	u32 reghotswap;
	u32 regpowermgmnt;

	wc->dsp_addr = 0x1;
	plx_setreg(wc, 0x4, wc->dsp_addr);

	if (plx_chip_type_get(wc) != 0x9054)
	{
		return -1;
	}

	pci_getconfig_dword(wc, 0x4, &regvalue);

	if (regvalue & (0xf8 << 24))
	{
		pci_setconfig_dword(wc, 0x4, regvalue);
	}

	regvalue = plx_getreg(wc, PCI9054_FIFO_CTRL_STAT);
	mu_enabled = (unsigned char)(regvalue & (1 << 0));
	regvalue = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);

	regvalue &= ~((1 << 30) | (1 << 29));
	eeprompresent = (unsigned char)((regvalue >> 28) & (1 << 0));

	pci_getconfig_dword(wc, 0x3c, &reginterrupt);

	if (eeprompresent)
	{
		pci_getconfig_dword(wc, 0x48, &reghotswap);
		pci_getconfig_dword(wc, 0x44, &regpowermgmnt);
	}


	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue | (1 << 30));
	shcdc_wait_ms(1000);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);

	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue | (1 << 29));
	shcdc_wait_ms(100);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);

	if (mu_enabled)
	{
		regvalue = plx_getreg(wc, PCI9054_FIFO_CTRL_STAT);
		plx_setreg(wc, PCI9054_FIFO_CTRL_STAT, regvalue | (1 << 0));
	}

	pci_setconfig_dword(wc, 0x3c, reginterrupt);

	if (eeprompresent)
	{
		reghotswap &= ~((1 << 23) | (1 << 22) | (1 << 17));
		pci_setconfig_dword(wc, 0x48, reghotswap);
		regpowermgmnt &= ~(1 << 15);
		pci_getconfig_dword(wc, 0x44, &regpowermgmnt);
	}
	

	return 0;
}

static int shcdc_init_dma(struct shcdc *wc)
{
	int i;
	struct dma_channel_desc desc;

	for (i = 0; i < 2; i++)
	{
		wc->dmainfo[i].sglbuffer.size = 4096;
		plx_dmabuf_alloc(wc, &wc->dmainfo[i].sglbuffer);
	}

	memset(&desc, 0x0, sizeof(struct dma_channel_desc));
	desc.iopbuswidth = 1;
	desc.waitstates = 0;
	desc.enablereadyinput = 1;
	desc.enablebterminput = 0;
	desc.enableiopburst = 0;
	desc.holdiopaddrconst = 0;
	desc.demandmode = 0;
	desc.enablewriteinvalidmode = 0;
	desc.enabledmaeoppin = 0;
	desc.dmastoptransfermode = 1;
	desc.tholdforpciwrites = 15;
	desc.tholdforpcireads = 15;
	if (plx_dmasgl_channnel_open(wc,DMA_CHANNEL0, &desc, wc))
	{
		return -1;
	}
	if (plx_dmasgl_channnel_open(wc,DMA_CHANNEL1, &desc, wc))
	{
		return -1;
	}
	return 0;
}

static void shcdc_init_readandwrite_chunk(struct shcdc *wc)
{
	int i;

	for (i = 0; i < 2; i++)
	{
		wc->dmainfo[i].dspbuffer.size = MAX_DMA_RW_SIZE * 3;
		plx_dmabuf_alloc(wc, &wc->dmainfo[i].dspbuffer);
	}

	wc->writechunk[0] = wc->dmainfo[0].dspbuffer.kernelva;
	wc->writechunk[1] = wc->dmainfo[0].dspbuffer.kernelva + MAX_DMA_RW_SIZE;
	wc->writechunk[2] = wc->dmainfo[0].dspbuffer.kernelva + MAX_DMA_RW_SIZE * 2;
	wc->readchunk[0] = wc->dmainfo[1].dspbuffer.kernelva;
	wc->readchunk[1] = wc->dmainfo[1].dspbuffer.kernelva + MAX_DMA_RW_SIZE;
	wc->readchunk[2] = wc->dmainfo[1].dspbuffer.kernelva + MAX_DMA_RW_SIZE * 2;

	wc->tmpreadchunk[0] = kmalloc(MAX_DMA_RW_SIZE, GFP_KERNEL);
	wc->tmpreadchunk[1] = kmalloc(MAX_DMA_RW_SIZE, GFP_KERNEL);
	wc->tmpreadchunk[2] = kmalloc(MAX_DMA_RW_SIZE, GFP_KERNEL);

	wc->tmpwritechunk[0] = kmalloc(MAX_DMA_RW_SIZE, GFP_KERNEL);
	wc->tmpwritechunk[1] = kmalloc(MAX_DMA_RW_SIZE, GFP_KERNEL);
	wc->tmpwritechunk[2] = kmalloc(MAX_DMA_RW_SIZE, GFP_KERNEL);

}

static void shcdc_release(struct shcdc *wc)
{
	if (wc->freeregion_plx)
		release_mem_region(wc->plx_region, wc->plx_len);
	if (wc->freeregion_dsp)
		release_mem_region(wc->dsp_region, wc->dsp_len);
	if (wc->plx)
		iounmap(wc->plx);
	if (wc->dsp)
		iounmap(wc->dsp);
	if (wc) 
		vfree(wc);
	printk("Freed a CDC-3522a\n");
}

static int shcdc_hardware_init_1(struct shcdc *wc)
{
	unsigned short val, val1, val2, val3;

	if (shcdc_plx_init(wc))
	{
		printk("shcdc: plx chip type is not 9054\n");
		return -1;
	}

	/*shcdc_card_reset(wc);*/

	shcdc_getdsp_data(wc, 0, 0x46, &val, sizeof(val) / sizeof(unsigned short));
	wc->order = val & 0xf;
	
	shcdc_getdsp_data(wc, 0, 0x40, &val1, sizeof(val) / sizeof(unsigned short));
	shcdc_getdsp_data(wc, 0, 0x42, &val2, sizeof(val) / sizeof(unsigned short));
	shcdc_getdsp_data(wc, 0, 0x44, &val3, sizeof(val) / sizeof(unsigned short));
	printk("%x, %x, %x\n", val1, val2, val3);
	
	return 0;
}

static int shcdc_hardware_init_2(struct shcdc *wc)
{
	shcdc_disable_pci_interrupts(wc);
	shcdc_disable_local_interrupts(wc);

	if (initbf527(wc))
	{
		printk("shcdc: initblackfin fail\n");
		return -1;
	}

	return 0;
}

static void cdc_chunkque_run(struct shcdc *wc) //maybe changed ...........
{
	int result, length, maxwritelength;
	
	int ndspw = 0;
	int ndspr = 0;

	if (wc->dmainfo[0].pending == 0 && ((NDsp+1)%wc->CurChNum == 0 || DisalbeAllCh))
	{
		//wc->dmainfo[0].pending	= 1;
		
		for(ndspw=0; ndspw<wc->CurDspNum; ++ndspw)
		{
			if(wc->tmpwritechunk_offset[ndspw] == wc->Stmpwritechunk_offset[ndspw])
				continue;
			
			if(wc->tmpwritechunk_offset[ndspw] != wc->Stmpwritechunk_offset[ndspw] && !wc->dead)
			{
				//down(&wc->tmpwritesem[ndspw]);
				
				maxwritelength = (MAX_DMA_RW_SIZE + wc->tmpwritechunk_offset[ndspw] - wc->Stmpwritechunk_offset[ndspw])%MAX_DMA_RW_SIZE;

				if(maxwritelength <= (QORT_LEN-wc->pQ0Rw[ndspw])*2)
				{
					if(wc->pQ0Rw[ndspw] < wc->pQ0Rr[ndspw] && maxwritelength > (wc->pQ0Rr[ndspw]-wc->pQ0Rw[ndspw])*2)
					{//很少出现
						wc->writechunk_offset[ndspw] = (wc->pQ0Rr[ndspw]-wc->pQ0Rw[ndspw])*2;
						
						if(wc->Stmpwritechunk_offset[ndspw]+wc->writechunk_offset[ndspw]<=MAX_DMA_RW_SIZE)
						{PrintDebugMsg(wc, ndspw, 1);
							memcpy(wc->writechunk[ndspw], wc->tmpwritechunk[ndspw]+wc->Stmpwritechunk_offset[ndspw], wc->writechunk_offset[ndspw]);
						}
						else
						{PrintDebugMsg(wc,ndspw, 2);
							memcpy(wc->writechunk[ndspw], wc->tmpwritechunk[ndspw]+wc->Stmpwritechunk_offset[ndspw], MAX_DMA_RW_SIZE-wc->Stmpwritechunk_offset[ndspw]);
							
							memcpy(wc->writechunk[ndspw]+MAX_DMA_RW_SIZE-wc->Stmpwritechunk_offset[ndspw], wc->tmpwritechunk[ndspw], wc->writechunk_offset[ndspw]+wc->Stmpwritechunk_offset[ndspw]-MAX_DMA_RW_SIZE);
						}
						
						wc->Stmpwritechunk_offset[ndspw] = (wc->Stmpwritechunk_offset[ndspw]+wc->writechunk_offset[ndspw])%MAX_DMA_RW_SIZE;
					}
					else
					{
						wc->writechunk_offset[ndspw] = maxwritelength;
						
						if(wc->Stmpwritechunk_offset[ndspw]+wc->writechunk_offset[ndspw]<=MAX_DMA_RW_SIZE)
						{//PrintDebugMsg(wc, ndspw, 3);
							memcpy(wc->writechunk[ndspw], wc->tmpwritechunk[ndspw]+wc->Stmpwritechunk_offset[ndspw], wc->writechunk_offset[ndspw]);
						}
						else
						{//PrintDebugMsg(wc, ndspw, 4);
							memcpy(wc->writechunk[ndspw], wc->tmpwritechunk[ndspw]+wc->Stmpwritechunk_offset[ndspw], MAX_DMA_RW_SIZE-wc->Stmpwritechunk_offset[ndspw]);
							
							memcpy(wc->writechunk[ndspw]+MAX_DMA_RW_SIZE-wc->Stmpwritechunk_offset[ndspw], wc->tmpwritechunk[ndspw], wc->writechunk_offset[ndspw]+wc->Stmpwritechunk_offset[ndspw]-MAX_DMA_RW_SIZE);
						}
						
						wc->Stmpwritechunk_offset[ndspw] = (wc->Stmpwritechunk_offset[ndspw]+wc->writechunk_offset[ndspw])%MAX_DMA_RW_SIZE;
					}
				}
				else
				{
					length = 0;
					while(length < (QORT_LEN - wc->pQ0Rw[ndspw]) * 2)
					{
						if(*(wc->tmpwritechunk[ndspw]+(wc->Stmpwritechunk_offset[ndspw]+length + 1)%MAX_DMA_RW_SIZE) == CDC_FORMAT_ULAW || *(wc->tmpwritechunk[ndspw]+(wc->Stmpwritechunk_offset[ndspw]+length + 1)%MAX_DMA_RW_SIZE) == CDC_FORMAT_ALAW)
						{
							if(wc->pQ0Rw[ndspw]<wc->pQ0Rr[ndspw] && length+AULAW_FRAME_LEN+2>(wc->pQ0Rr[ndspw]-wc->pQ0Rw[ndspw])*2)
								break;
							length += AULAW_FRAME_LEN + 2;
						}
						else if(*(wc->tmpwritechunk[ndspw]+(wc->Stmpwritechunk_offset[ndspw]+length + 1)%MAX_DMA_RW_SIZE) == CDC_FORMAT_G729A)
						{
							if(wc->pQ0Rw[ndspw]<wc->pQ0Rr[ndspw] && length+G729_FRAME_LEN+2>(wc->pQ0Rr[ndspw]-wc->pQ0Rw[ndspw])*2)
								break;
							length += G729_FRAME_LEN + 2;
						}
						else
						{
							if(debug)
								printk("shcdc_transmit_data ERROR TYPE\n");
							length = (QORT_LEN - wc->pQ0Rw[ndspw]) * 2;
						}
					}
					wc->writechunk_offset[ndspw] = length;
					
					if(wc->Stmpwritechunk_offset[ndspw]+wc->writechunk_offset[ndspw]<=MAX_DMA_RW_SIZE)
					{//PrintDebugMsg(wc, ndspw, 5);
						memcpy(wc->writechunk[ndspw], wc->tmpwritechunk[ndspw]+wc->Stmpwritechunk_offset[ndspw], wc->writechunk_offset[ndspw]);
					}
					else
					{//PrintDebugMsg(wc, ndspw, 6);
						memcpy(wc->writechunk[ndspw], wc->tmpwritechunk[ndspw]+wc->Stmpwritechunk_offset[ndspw], MAX_DMA_RW_SIZE-wc->Stmpwritechunk_offset[ndspw]);
							
						memcpy(wc->writechunk[ndspw]+MAX_DMA_RW_SIZE-wc->Stmpwritechunk_offset[ndspw], wc->tmpwritechunk[ndspw], wc->writechunk_offset[ndspw]+wc->Stmpwritechunk_offset[ndspw]-MAX_DMA_RW_SIZE);
					}
						
					wc->Stmpwritechunk_offset[ndspw] = (wc->Stmpwritechunk_offset[ndspw]+wc->writechunk_offset[ndspw])%MAX_DMA_RW_SIZE;
				}

				//up(&wc->tmpwritesem[ndspw]);
				
				result = shcdc_transmit_data(wc, ndspw);
	
				if(result && debug)
				{
					printk("In shcdc_transmit_data Error Code = %x\n", result);
					//wc->dmainfo[0].pending	= 0;
				}
			}
			//else
				//wc->dmainfo[0].pending	= 0;
		}
	}
	
	if (wc->dmainfo[1].pending == 0)
	{
		for(ndspr=0; ndspr<wc->CurDspNum; ++ndspr)
		{
			//wc->dmainfo[1].pending	= 1;
	
			if(!wc->dead)
			{
				result = shcdc_receive_data(wc, ndspr);
	
				/*if(result)
				{
					if(result != 1 && debug)
						printk("In shcdc_receive_data Error Code = %x\n", result);
					//wc->dmainfo[1].pending	= 0;
				}*/
			}
			//else
				//wc->dmainfo[1].pending	= 0;
		}
	}

}




static void do_interrupts(void *p)
{
	u32 regvalue;
	unsigned long intsrc;
	struct shcdc *wc = (struct shcdc *)p;
	int length, tmplength, ndsp, maxreadlength;
	
//	DMA_PARAMS pDmaParams;

	intsrc = wc->interruptsrc;

	if (intsrc & INTR_TYPE_LOCAL_1)
	{
		cdc_chunkque_run(wc);
		
		if (wc->dmainfo[0].dmadatacount && wc->dmainfo[0].pending == 0)
		{
			plx_dmasgl_transfer(wc, DMA_CHANNEL0);
		}
		
		if (wc->dmainfo[1].dmadatacount && wc->dmainfo[1].pending == 0)
		{
			plx_dmasgl_transfer(wc, DMA_CHANNEL1);
		}
		
		if (wc->dmainfo[0].pending == 0 && wc->dmainfo[1].pending == 0)
		{
			shcdc_enable_cpld_interrupts(wc);
			shcdc_enable_local_interrupts(wc);
		}
		
	}

	if (intsrc & INTR_TYPE_DOORBELL)
	{
		regvalue = plx_getreg(wc, 0x64);
		plx_setreg(wc, 0x64, regvalue);
	}

	if (intsrc & INTR_TYPE_PCI_ABORT)
	{
		pci_getconfig_dword(wc, 0x4, &regvalue);
		pci_setconfig_dword(wc, 0x4, regvalue);
	}

	if (intsrc & INTR_TYPE_DMA_0)
	{
		regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
		plx_setreg(wc, PCI9054_DMA_COMMAND_STAT, regvalue | (1 << 3));
		regvalue = plx_getreg(wc, PCI9054_DMA0_MODE);
		if (regvalue & (1 << 9))
		{
			DMA_PARAMS pDmaParams;
			for(ndsp=0; ndsp<wc->CurDspNum; ndsp++)
			{
				if(wc->EnableWriteDsp[ndsp] == 0)
					continue;
				wc->EnableWriteDsp[ndsp] = 0;
				switch(ndsp)
				{
		
					case 0:
					{
						pDmaParams.DataAddr = BFA_DATA_ADDR;
						pDmaParams.PortAddr = BFA_PORT_ADDR;
						pDmaParams.MemAddr = BFA_MEM_ADDR;
						pDmaParams.DmaBusPhysical = wc->dmainfo[0].dspbuffer.busphysical;
						break;
					}
					case 1:
					{
						pDmaParams.DataAddr = BFB_DATA_ADDR;
						pDmaParams.PortAddr = BFB_PORT_ADDR;
						pDmaParams.MemAddr = BFB_MEM_ADDR;
						pDmaParams.DmaBusPhysical = wc->dmainfo[0].dspbuffer.busphysical + MAX_DMA_RW_SIZE;
						break;
					}
					case 2:
					{
						pDmaParams.DataAddr = BFC_DATA_ADDR;
						pDmaParams.PortAddr = BFC_PORT_ADDR;
						pDmaParams.MemAddr = BFC_MEM_ADDR;
						pDmaParams.DmaBusPhysical = wc->dmainfo[0].dspbuffer.busphysical + MAX_DMA_RW_SIZE * 2;
						break;
					}
				}
		
				pDmaParams.XModify = 0x2;
				pDmaParams.YCount = 0x0;
				pDmaParams.YModify = 0x0;
				pDmaParams.Config = DMA_WRITE_COM;
				pDmaParams.LowAddr = (unsigned short)QOR_W_ADDR;
				pDmaParams.HighAddr = (unsigned short)(QOR_W_ADDR >> 16);
				pDmaParams.XCount = 1;
				pDmaParams.pData = (unsigned short *)&(wc->pQ0Rw[ndsp]);
				shcdc_hostwrite(wc, &pDmaParams);
			}
			
			plx_dmasgl_transfer_complete(wc, 0);
		}
	}

	if (intsrc & INTR_TYPE_DMA_1)
	{
		regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
		plx_setreg(wc, PCI9054_DMA_COMMAND_STAT, regvalue | (1 << 11));
		regvalue = plx_getreg(wc, PCI9054_DMA1_MODE);
		if (regvalue & (1 << 9))
		{
			for(ndsp=0; ndsp<wc->CurDspNum; ndsp++)
			{
				if(wc->EnableReadDsp[ndsp] == 0)
					continue;
				wc->EnableReadDsp[ndsp] = 0;
				
				maxreadlength = (MAX_DMA_RW_SIZE + wc->Stmpreadchunk_offset[ndsp] - wc->tmpreadchunk_offset[ndsp])%MAX_DMA_RW_SIZE;
				
				if(wc->pQ0Tw[ndsp] < wc->pQ0Tr[ndsp])
				{
					length = 0;
					while(length < (QORT_LEN - wc->pQ0Tr[ndsp]) * 2)
					{
						if(wc->readchunk[ndsp][length + 1] == CDC_FORMAT_ULAW || wc->readchunk[ndsp][length + 1] == CDC_FORMAT_ALAW)
						{
							if(maxreadlength>0 && (length+AULAW_FRAME_LEN+2)>=maxreadlength)
								break;
							length += AULAW_FRAME_LEN + 2;
						}
						else if(wc->readchunk[ndsp][length + 1] == CDC_FORMAT_G729A)
						{
							if(maxreadlength>0 && (length+G729_FRAME_LEN+2)>=maxreadlength)
								break;
							length += G729_FRAME_LEN + 2;
						}
						else
						{
							if(debug)
								printk("shcdc_receive_data ERROR TYPE[%x], %d, %d, %d\n", wc->readchunk[ndsp][length + 1], ndsp, wc->pQ0Tr[ndsp], length);
							length = 0;
							break;
						}
					}
					wc->pQ0Tr[ndsp] = 0;
					wc->readchunk_offset[ndsp] = length;
				}
				else
				{
					wc->readchunk_offset[ndsp]= (wc->pQ0Tw[ndsp] - wc->pQ0Tr[ndsp])*2;
					wc->pQ0Tr[ndsp] = wc->pQ0Tw[ndsp];
					
					if(maxreadlength>0 && wc->readchunk_offset[ndsp]>=maxreadlength)
					{
						length = 0;
						while(length < wc->readchunk_offset[ndsp])
						{
							if(wc->readchunk[ndsp][length + 1] == CDC_FORMAT_ULAW || wc->readchunk[ndsp][length + 1] == CDC_FORMAT_ALAW)
							{
								if(maxreadlength>0 && (length+AULAW_FRAME_LEN+2)>=maxreadlength)
									break;
								length += AULAW_FRAME_LEN + 2;
							}
							else if(wc->readchunk[ndsp][length + 1] == CDC_FORMAT_G729A)
							{
								if(maxreadlength>0 && (length+G729_FRAME_LEN+2)>=maxreadlength)
									break;
								length += G729_FRAME_LEN + 2;
							}
							else
							{
								length = wc->readchunk_offset[ndsp];
							}
						}
						wc->readchunk_offset[ndsp] = length;
					}
				}
				
				if(wc->readchunk_offset[ndsp] == 0)
					continue;
				
				CountREC += wc->readchunk_offset[ndsp];
	
				tmplength = 0;
				if(wc->tmpreadchunk_offset[ndsp]+wc->readchunk_offset[ndsp] > MAX_DMA_RW_SIZE)
				{
					tmplength = MAX_DMA_RW_SIZE-wc->tmpreadchunk_offset[ndsp];
					memcpy(wc->tmpreadchunk[ndsp]+wc->tmpreadchunk_offset[ndsp], wc->readchunk[ndsp], tmplength);
				
					//down(&wc->tmpreadsem[ndsp]);
					wc->tmpreadchunk_offset[ndsp] = 0;
					//up(&wc->tmpreadsem[ndsp]);
					
					wc->readchunk_offset[ndsp] = wc->readchunk_offset[ndsp] - tmplength;
				}
				memcpy(wc->tmpreadchunk[ndsp]+wc->tmpreadchunk_offset[ndsp], wc->readchunk[ndsp]+tmplength, wc->readchunk_offset[ndsp]);
				
				//down(&wc->tmpreadsem[ndsp]);
				wc->tmpreadchunk_offset[ndsp] = (wc->tmpreadchunk_offset[ndsp] + wc->readchunk_offset[ndsp])%MAX_DMA_RW_SIZE;
				//up(&wc->tmpreadsem[ndsp]);
				
				wc->readchunk_offset[ndsp] = 0;
			}
			
			plx_dmasgl_transfer_complete(wc, 1);

		}
	}

	if (intsrc & INTR_TYPE_OUTBOUND_POST)
	{
		plx_setreg(wc, PCI9054_OUTPOST_INT_MASK, (1 << 3));
	}

	shcdc_enable_pci_interrupts(wc);
}

DAHDI_IRQ_HANDLER(shcdc_interrupt)
{
	struct shcdc *wc = dev_id;
	unsigned long interruptreg, interruptsrc, regvalue;

	interruptreg = plx_getreg(wc, PCI9054_INT_CTRL_STAT);
	
	IntCount1++;
		
	if (interruptreg == 0xffffffff)
	{
		return IRQ_NONE;
	}

	if ((interruptreg & (1 << 8)) == 0)
	{
		return IRQ_NONE;
	}

	interruptsrc = INTR_TYPE_NONE;

	if ((interruptreg & (1 << 13)) && (interruptreg & (1 << 9)))
	{
		interruptsrc |= INTR_TYPE_DOORBELL;
	}

	if ((interruptreg & (1 << 14)) && (interruptreg & (1 << 10)))
	{
		interruptsrc |= INTR_TYPE_PCI_ABORT;
	}

	if ((interruptreg & (1 << 15)) && (interruptreg & (1 << 11)))
	{
		IntCount++;
		interruptsrc |= INTR_TYPE_LOCAL_1;
		shcdc_disable_cpld_interrupts(wc);
		shcdc_disable_local_interrupts(wc);
	}

	if ((interruptreg & (1 << 21)) && (interruptreg & (1 << 18)))
	{
		regvalue = plx_getreg(wc, PCI9054_DMA0_MODE);
		if (regvalue & (1 << 17))
			interruptsrc |= INTR_TYPE_DMA_0;
	}

	if ((interruptreg & (1 << 22)) && (interruptreg & (1 << 19)))
	{
		regvalue = plx_getreg(wc, PCI9054_DMA1_MODE);
		if (regvalue & (1 << 17)) 
			interruptsrc |= INTR_TYPE_DMA_1;
	}

	regvalue = plx_getreg(wc, PCI9054_OUTPOST_INT_STAT);
	if (regvalue & (1 << 3))
	{
		regvalue = plx_getreg(wc, PCI9054_OUTPOST_INT_MASK);
		if ((regvalue & (1 << 3)) == 0)
			interruptsrc |= INTR_TYPE_OUTBOUND_POST;
	}


	if (interruptsrc == INTR_TYPE_NONE)
	{
		return IRQ_NONE;
	}

	wc->interruptsrc = interruptsrc;

	shcdc_disable_pci_interrupts(wc);

	if (wc->irq_enable)
		do_interrupts((void *)wc);

	return IRQ_RETVAL(1);
}

static void cdc_init_state(struct cdc_state *state_ptr, int encoder, unsigned int channel, struct shcdc *wc)
{
	state_ptr->encoder = encoder;
	state_ptr->wc = wc;

	state_ptr->packets_sent = 0;
	state_ptr->packets_received = 0;
	state_ptr->last_cdc_seqno = 0;
	state_ptr->cdc_seqno_rcv = 0;

	state_ptr->chan_in_num = 999;
	state_ptr->chan_out_num = 999;

	state_ptr->ssrc = 0x78;
	
	if (encoder == 1)
	{
		state_ptr->timeslot_in_num = channel * 2;
		state_ptr->timeslot_out_num = channel * 2 + 1;
	} else {
		state_ptr->timeslot_in_num = channel * 2 + 1;
		state_ptr->timeslot_out_num = channel * 2;
	}
}

//static int cdc_operation(struct dahdi_transcoder_channel *ztc, int op);//changed...........


static void shfxm_SerialNo(struct shcdc *wc)
{
	s8 BitPos, count;
	u32 regvalue, EepromCommand, tmpreg;
	
	regvalue = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
	
	regvalue &= ~(0xF << 24);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);
	regvalue |= (1 << 25);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);
	
	EepromCommand = (0x0180<<2) | (0x1f0/2);
	for (BitPos = 10; BitPos >= 0; BitPos--)
	{
		if (EepromCommand & (1 << BitPos))
			plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue | (1 << 26));
		else
			plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);
		
		tmpreg = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, tmpreg | (1 << 24));
		for (count=0; count<20; count++)
			tmpreg = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, tmpreg & ~(1 << 24));
	}
	
	regvalue = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
	regvalue |= (1 << 31);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue | (1 << 26));

	for (BitPos = 0; BitPos < 32; BitPos++)
	{
		tmpreg = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, tmpreg | (1 << 24));
		for (count=0; count<20; count++)
			tmpreg = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, tmpreg & ~(1 << 24));

		for (count=0; count <20; count++)
			regvalue = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);

		if (regvalue & (1 << 27))
			wc->serialno = (wc->serialno << 1) | 1;
		else
			wc->serialno = (wc->serialno << 1);
	}
	
	regvalue &= ~(1 << 31);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue & ~(0xF << 24));
	
	wc->serialno = (wc->serialno >> 16 | wc->serialno << 16);
}

static ssize_t 
cdc_read(struct file *file, char __user *frame, size_t count, loff_t *ppos)
{
	unsigned int tmp_readchunk_offset, tmp_readchunk_end,i;
	unsigned char ztc_ndx, rcodec;
	struct dahdi_transcoder_channel *dtc = file->private_data;
	struct  cdc_state *cpvt = dtc->pvt;
	struct shcdc *wc = cpvt->wc;
	unsigned char *chars = NULL;
	int ibytes = 0;
	int ndsp;
	struct cdc_state *st = NULL;

	chars =	kmalloc(count, GFP_KERNEL);
	
	for(ndsp=0; ndsp<wc->CurDspNum; ++ndsp)
	{
		down(&wc->tmpreadsem[ndsp]);
		tmp_readchunk_end = wc->tmpreadchunk_offset[ndsp];
		
		tmp_readchunk_offset = wc->Stmpreadchunk_offset[ndsp];
		while(tmp_readchunk_offset != tmp_readchunk_end)
		{
			ztc_ndx = wc->tmpreadchunk[ndsp][tmp_readchunk_offset] + (char)(ndsp*50);
			rcodec = wc->tmpreadchunk[ndsp][(tmp_readchunk_offset + 1)%MAX_DMA_RW_SIZE];
	
			if (ztc_ndx >= wc->numchannels)
			{
				if (debug)
					printk("shcdc error: Invalid channel number received (ztc_ndx = %d) (numchannels = %d)\n", ztc_ndx, wc->numchannels);
				rcodec = CDC_FORMAT_UNDEF;
			}
	
			if ((rcodec == CDC_FORMAT_ULAW) || (rcodec == CDC_FORMAT_ALAW))	/* ulaw or alaw (decoders) */
			{
				dtc = &(wc->udecode->channels[ztc_ndx]);
			
				st = dtc->pvt;
			}
			
			if (rcodec == CDC_FORMAT_G729A)		/*g.729 (encoders) */
			{
				dtc = &(wc->uencode->channels[ztc_ndx]);
			
				st = dtc->pvt;
			}
	
			if (rcodec == CDC_FORMAT_ULAW || rcodec == CDC_FORMAT_ALAW)	/* ulaw or alaw*/
			{
			
					for (i = 0; i < AULAW_FRAME_LEN; i++)
					{
						chars[i] = wc->tmpreadchunk[ndsp][(tmp_readchunk_offset + i + 2)%MAX_DMA_RW_SIZE];
						ibytes++;
					}
	
					tmp_readchunk_offset = (tmp_readchunk_offset + AULAW_FRAME_LEN + 2)%MAX_DMA_RW_SIZE;
	
		
				
			}
			else if (rcodec == CDC_FORMAT_G729A)	/* G.729a */
			{
		
					for (i = 0; i < G729_FRAME_LEN; i+=2)
					{
						chars[i+1] = wc->tmpreadchunk[ndsp][(tmp_readchunk_offset + i + 2)%MAX_DMA_RW_SIZE];
						chars[i] = wc->tmpreadchunk[ndsp][(tmp_readchunk_offset + i + 3)%MAX_DMA_RW_SIZE];
						ibytes+=2;
					}
	
				
					tmp_readchunk_offset = (tmp_readchunk_offset + G729_FRAME_LEN + 2)%MAX_DMA_RW_SIZE;
			}
			else
			{
				if (debug)
					PrintDebugMsg(wc, ndsp, 93);
				tmp_readchunk_offset = tmp_readchunk_end;
			}
		}

		wc->Stmpreadchunk_offset[ndsp]	= tmp_readchunk_end;
		up(&wc->tmpreadsem[ndsp]);
	}
	if (unlikely(copy_to_user(frame, chars, ibytes))) 
	{
		kfree(chars);
	       	return -EFAULT;
	}
	kfree(chars);

	return 0;
}

static ssize_t 
cdc_write(struct file *file, const char __user *frame, size_t count, loff_t *ppos)
{
	
	struct dahdi_transcoder_channel *dtc = file->private_data;
	struct cdc_state *cpvt = dtc->pvt;
	struct shcdc *wc = cpvt->wc;
  	unsigned char *chars = NULL;
	int inbytes = 0,ndx,ndsp,NDsp,i;

	if ( (((dtc->srcfmt == DAHDI_FORMAT_ULAW) || (dtc->srcfmt == DAHDI_FORMAT_ALAW)) && ((dtc->dstfmt == DAHDI_FORMAT_G729A  ))) || ((dtc->srcfmt == DAHDI_FORMAT_G729A) ))
			{
				ndx = cpvt->timeslot_in_num/2;
				if(ndx >= wc->numchannels)
					return 0;
			
				NDsp = ndx;
				ndsp = ndx / 50;
				ndx = ndx % 50;
				
					do
				{
					chars =	kmalloc(count, GFP_KERNEL);
					if (copy_from_user(chars, frame, count))
						{
							kfree(chars);
							return -EFAULT;
						}
					
					if ((dtc->srcfmt == DAHDI_FORMAT_ULAW) || (dtc->srcfmt == DAHDI_FORMAT_ALAW))
					{
						inbytes = AULAW_FRAME_LEN;
					}
					else if (dtc->srcfmt == DAHDI_FORMAT_G729A)
					{
						inbytes = G729_FRAME_LEN;
					}
					
					down(&wc->tmpwritesem[ndsp]);
					
					if((wc->Stmpwritechunk_offset[ndsp]<wc->tmpwritechunk_offset[ndsp] && (wc->tmpwritechunk_offset[ndsp]+inbytes-MAX_DMA_RW_SIZE)>=wc->Stmpwritechunk_offset[ndsp]) || (wc->Stmpwritechunk_offset[ndsp]>wc->tmpwritechunk_offset[ndsp] && (wc->tmpwritechunk_offset[ndsp]+inbytes)>=wc->Stmpwritechunk_offset[ndsp]))
					{
					//	printk("The Channel %d throw packet====[%d], [%d], [%d]\n", ndsp*50+ndx, wc->Stmpwritechunk_offset[ndsp], wc->tmpwritechunk_offset[ndsp], inbytes);
						
						up(&wc->tmpwritesem[ndsp]);
						dahdi_transcoder_alert(dtc);
						break;
					}
					
					if ((dtc->srcfmt == DAHDI_FORMAT_ULAW) || (dtc->srcfmt == DAHDI_FORMAT_ALAW))
					{
						if(dtc->srcfmt == DAHDI_FORMAT_ULAW)
							wc->tmpwritechunk[ndsp][(wc->tmpwritechunk_offset[ndsp] + 1)%MAX_DMA_RW_SIZE] = CDC_FORMAT_ULAW;
						else
							wc->tmpwritechunk[ndsp][(wc->tmpwritechunk_offset[ndsp] + 1)%MAX_DMA_RW_SIZE] = CDC_FORMAT_ALAW;
					}
					else if (dtc->srcfmt == DAHDI_FORMAT_G729A)
					{
						wc->tmpwritechunk[ndsp][(wc->tmpwritechunk_offset[ndsp] + 1)%MAX_DMA_RW_SIZE] = CDC_FORMAT_G729A;
						
						if(dtc->dstfmt == DAHDI_FORMAT_ULAW)
							wc->dstfmt[ndx + ndsp*50] = 0;
						else
							wc->dstfmt[ndx + ndsp*50] = 1;
					}
					
					wc->tmpwritechunk[ndsp][wc->tmpwritechunk_offset[ndsp]%MAX_DMA_RW_SIZE] = (char)(ndx & 0x000000ff);

					

					if(inbytes == AULAW_FRAME_LEN)
					{
						for (i = 0; i < inbytes; i++)
							wc->tmpwritechunk[ndsp][(wc->tmpwritechunk_offset[ndsp] + i + 2)%MAX_DMA_RW_SIZE]= chars[i];
					}
					else if(inbytes == G729_FRAME_LEN)
					{
						for (i = 0; i < G729_FRAME_LEN; i+=2)
						{
							wc->tmpwritechunk[ndsp][(wc->tmpwritechunk_offset[ndsp] + i + 2)%MAX_DMA_RW_SIZE]= chars[i+1];
							wc->tmpwritechunk[ndsp][(wc->tmpwritechunk_offset[ndsp] + i + 3)%MAX_DMA_RW_SIZE]= chars[i];
						}
					}
					else
						printk("Read ERROR PACKET\n");

					wc->tmpwritechunk_offset[ndsp] = (wc->tmpwritechunk_offset[ndsp] + inbytes + 2)%MAX_DMA_RW_SIZE;
					
					up(&wc->tmpwritesem[ndsp]);

					cpvt->packets_sent++;

					

				} while( (((dtc->srcfmt == DAHDI_FORMAT_ULAW) || (dtc->srcfmt == DAHDI_FORMAT_ALAW)) && ((dtc->dstfmt == DAHDI_FORMAT_G729A ))) || ((dtc->srcfmt == DAHDI_FORMAT_G729A) ));
	
			}
			else
			{
				dahdi_transcoder_alert(dtc);
			}
			kfree(chars);

	return 0;
}

static int
cdc_release(struct dahdi_transcoder_channel *dtc)
{

			struct dahdi_transcoder_channel *compl_dtc;
			struct cdc_state *st = dtc->pvt, *compl_st;
			struct shcdc *wc = ((struct cdc_state*)(dtc->pvt))->wc;
			unsigned int ndx;
	
			down(&wc->chansemclose);
			ndx = st->timeslot_in_num/2;

			if (st->encoder == 1)
				compl_dtc = &(wc->udecode->channels[ndx]);
			else
				compl_dtc = &(wc->uencode->channels[ndx]);

			/* If the channel complement (other half of the encoder/decoder pair) is not being used... */
			if ((compl_dtc->flags & DAHDI_TC_FLAG_BUSY) == 0)
			{
				/* Mark this channel as not built */
				dahdi_tc_clear_built(dtc);
				dtc->built_fmts = 0;
				st->chan_in_num = 999;
				st->chan_out_num = 999;
			
				/* Mark the channel complement as not built */
				dahdi_tc_clear_built(compl_dtc);
				compl_dtc->built_fmts = 0;
				compl_st = compl_dtc->pvt;
				compl_st->chan_in_num = 999;
				compl_st->chan_out_num = 999;
			}
			up(&wc->chansemclose);		
			return 0;
}

static int
cdc_allocate(struct dahdi_transcoder_channel *dtc)
{

	struct dahdi_transcoder_channel *compl_dtc;
	struct cdc_state *st = dtc->pvt, *compl_st;
	struct shcdc *wc = ((struct cdc_state*)(dtc->pvt))->wc;
	unsigned int ndx;
		
		down(&wc->chansem);
			if (dahdi_tc_is_built(dtc))
			{
				/* Mark this channel as built */
				dahdi_tc_set_built(dtc);
				dtc->built_fmts = dtc->dstfmt | dtc->srcfmt;

				/* Mark the channel complement (other half of encoder/decoder pair) as built */
				ndx = st->timeslot_in_num/2;
				if (st->encoder == 1)
					compl_dtc = &(wc->udecode->channels[ndx]);
				else
					compl_dtc = &(wc->uencode->channels[ndx]);
				dahdi_tc_set_built(compl_dtc);
				compl_dtc->built_fmts = dtc->dstfmt | dtc->srcfmt;
				compl_st = compl_dtc->pvt;
				compl_st->chan_in_num = st->chan_out_num;
				compl_st->chan_out_num = st->chan_in_num;
			}
			up(&wc->chansem);
			
			return 0;
}

static int __devinit shcdc_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int res;
	struct shcdc *wc;
	struct shcdc_desc *d = (struct shcdc_desc *)ent->driver_data;
	int x, i;
	static int initd_ifaces=0;

	char tmpname[30];
	
	if (!initd_ifaces)
	{
		memset((void *)ifaces,0,(sizeof(struct shcdc *))*WC_MAX_IFACES);
		initd_ifaces=1;
	}
	for (x=0;x<WC_MAX_IFACES;x++)
		if (!ifaces[x]) break;
	if (x >= WC_MAX_IFACES)
	{
		printk("shcdc: Too many interfaces\n");
		return -EIO;
	}

	if (pci_enable_device(pdev))
	{
		res = -EIO;
	}
	else
	{
		wc = vmalloc(sizeof(struct shcdc));		//kmalloc
		if (wc)
		{
			ifaces[x] = wc;
			memset(wc, 0, sizeof(struct shcdc));
			wc->BF527Init = 1;
			
			spin_lock_init(&wc->reglock);
			sema_init(&wc->chansem, 1);
			sema_init(&wc->chansemclose, 1);
			sema_init(&wc->tmpwritesem[0], 1);
			sema_init(&wc->tmpwritesem[1], 1);
			sema_init(&wc->tmpwritesem[2], 1);
			sema_init(&wc->tmpreadsem[0], 1);
			sema_init(&wc->tmpreadsem[1], 1);
			sema_init(&wc->tmpreadsem[2], 1);
			
			if(ent->device == 0x00f9)
			{
				wc->CurChNum	= 150;
				wc->CurDspNum	= 3;
			}
			else if(ent->device == 0x00f8)
			{
				wc->CurChNum	= 100;
				wc->CurDspNum	= 2;
			}
			else
			{
				wc->CurChNum	= 50;
				wc->CurDspNum	= 1;
			}
				wc->irq_enable = 0;
			wc->dead = 0;
			
			wc->dev = pdev;
			wc->pos = x;
			wc->variety = d->name;

			wc->rcvflags = 0;
			wc->last_seqno = 999;
			wc->last_rparm1 = 0;
			wc->seq_num = 6;
			wc->timeout = HZ/10 + 1; 	/* 100msec */
			wc->dsp_crashed = 0;
			wc->dumping = 0;
			wc->ztsnd_rtx = 0;
			wc->ztsnd_0010_rtx = 0;

			for(i=0; i<wc->CurDspNum; ++i)
			{
				wc->pQ0Tr[i] = 0;
				wc->pQ0Tw[i] = 0;
				wc->pQ0Rw[i] = 0;
				wc->pQ0Rr[i] = 0;
				wc->EnableReadDsp[i] = 0;
				wc->EnableWriteDsp[i] = 0;
				
				wc->writechunk_offset[i] = 0;
				wc->readchunk_offset[i] = 0;
				wc->tmpwritechunk_offset[i] = 0;
				wc->tmpreadchunk_offset[i] = 0;
				wc->Stmpwritechunk_offset[i] = 0;
				wc->Stmpreadchunk_offset[i] = 0;
			}
				for(i=0; i<wc->CurChNum; ++i)
				wc->dstfmt[i] = 0;

			sprintf(tmpname, "shcdc%d", x);
			wc->plx_region = pci_resource_start(pdev, 0);
			wc->plx_len = pci_resource_len(pdev,0);

			if (request_mem_region(wc->plx_region, wc->plx_len, tmpname))
				wc->freeregion_plx = 1;
			else
			{
				printk("plx region %lx-%lx already in use\n", wc->plx_region, wc->plx_region + wc->plx_len);
				goto err_out_free_wc;
			}
			wc->plx = ioremap(wc->plx_region, wc->plx_len);
	if (!wc->plx)
			{
				printk("Cannot ioremap memory space at 0x%lx\n", wc->plx_region);
				goto err_out_free_wc;
			}

			sprintf(tmpname, "shcdc%d", x);
			wc->dsp_region = pci_resource_start(pdev,2);
			wc->dsp_len = pci_resource_len(pdev,2);
			if (request_mem_region(wc->dsp_region, wc->dsp_len, tmpname))
				wc->freeregion_dsp = 1;
			else
			{
				printk("dsp region %lx-%lx already in use\n", wc->dsp_region, wc->dsp_region + wc->dsp_len);
				goto err_out_free_wc;
			}
			wc->dsp = ioremap(wc->dsp_region, wc->dsp_len);
			if (!wc->dsp)
			{
				printk("Cannot ioremap memory space at 0x%lx\n", wc->dsp_region);
				goto err_out_free_wc;
			}
			if (shcdc_hardware_init_1(wc))
			{
				printk("shcdc: hardware init 1 fail\n");
				goto err_out_free_wc;
			}

			if (shcdc_init_dma(wc))
			{
				printk("shcdc: init dma fail\n");
				goto err_out_free_wc;
			}

			shcdc_init_readandwrite_chunk(wc);

		
			//wc->confbuffer.size = DMA_CONFIG_LENGTH;
			//plx_dmabuf_alloc(wc, &wc->confbuffer);

			init_waitqueue_head(&wc->regq);
			
			/* Setup Encoders and Decoders */
			wc->numchannels = wc->CurChNum;
			uencode = dahdi_transcoder_alloc(wc->numchannels);
			udecode = dahdi_transcoder_alloc(wc->numchannels);
			encoders = vmalloc(sizeof(struct cdc_state) * wc->numchannels);
			decoders = vmalloc(sizeof(struct cdc_state) * wc->numchannels);
			memset(encoders, 0, sizeof(struct cdc_state) * wc->numchannels);
			memset(decoders, 0, sizeof(struct cdc_state) * wc->numchannels);
			if (!uencode || !udecode || !encoders || !decoders)
			{
				if (uencode)
					dahdi_transcoder_free(uencode);
				if (udecode)
					dahdi_transcoder_free(udecode);
				if (encoders)
					vfree(encoders);
				if (decoders)
					vfree(decoders);
				goto err_out_free_wc;
			//	return -ENOMEM;
			}
			sprintf(udecode->name, "cdc Decoder");
			sprintf(uencode->name, "cdc Encoder");

			udecode->srcfmts = uencode->dstfmts = DAHDI_FORMAT_G729A;
			udecode->dstfmts = uencode->srcfmts = DAHDI_FORMAT_ULAW | DAHDI_FORMAT_ALAW;
		
			/* --modify-- */
			
		uencode->allocate = udecode->allocate = cdc_allocate;
		uencode->release = udecode->release = cdc_release;
			
		uencode->fops.owner = udecode->fops.owner = THIS_MODULE;               	   uencode->fops.read = udecode->fops.read = cdc_read;
		uencode->fops.write = udecode->fops.write = cdc_write;
			
			//udecode->operation = uencode->operation = cdc_operation;//changed...........
			
			/* --mod end-- */

				
			for (x=0;x<wc->numchannels;x++)
			{
				cdc_init_state(encoders + x, 1, x, wc);
				encoders[x].encoder = 1;
				decoders[x].encoder = 0;
				cdc_init_state(decoders + x, 0, x, wc);
				uencode->channels[x].pvt = encoders + x;
				udecode->channels[x].pvt = decoders + x;
			}

			wc->uencode = uencode;
			wc->udecode = udecode;
			dahdi_transcoder_register(uencode);
			dahdi_transcoder_register(udecode);
			printk("Zaptel cdc (G.729a) Transcoder support LOADED \n");

			/* Enable bus mastering */
			pci_set_master(pdev);
			/* Keep track of which device we are */
			pci_set_drvdata(pdev, wc);

			sprintf(tmpname, "shcdc%d", x);
			if (request_irq(pdev->irq, shcdc_interrupt, DAHDI_IRQ_SHARED, tmpname, wc))
			{	//ZAP_IRQ_SHARED_DISABLED
				printk("shcdc: Unable to request IRQ %d\n", pdev->irq);
				pci_set_drvdata(pdev, NULL);
				goto err_out_free_wc;
			}
		
			if (shcdc_hardware_init_2(wc)) {
				free_irq(pdev->irq, wc);
				pci_set_drvdata(pdev, NULL);
				printk("shcdc: hardware init 2 fail\n");
				goto err_out_free_wc;
			}

			shfxm_SerialNo(wc);
			printk("CDC: Launching card: %d\n", wc->serialno);

 			wc->irq_enable = 1;
			shcdc_enable_local_interrupts(wc);
			shcdc_enable_pci_interrupts(wc);
			
			wc->BF527Init = 0;

			res = 0;
		}
		else
			res = -ENOMEM;
	}
	
	return res;
	
err_out_free_wc:
	if (wc->freeregion_plx)
		release_mem_region(wc->plx_region, wc->plx_len);
	if (wc->freeregion_dsp)
		release_mem_region(wc->dsp_region, wc->dsp_len);
	if (wc->plx)
		iounmap(wc->plx);
	if (wc->dsp)
		iounmap(wc->dsp);
	if (wc) 
		kfree(wc);
	return -EIO;
}

static void __devexit shcdc_remove_one(struct pci_dev *pdev)
{
	int i;
	struct shcdc *wc = pci_get_drvdata(pdev);
	struct dahdi_transcoder_channel *ztc_en, *ztc_de;
	struct cdc_state *st_en, *st_de;
	
	if (wc)
	{
		if(debug)
		{
			printk("shcdc debug: wc->ztsnd_rtx = %d\n", wc->ztsnd_rtx);
			printk("shcdc debug: wc->ztsnd_0010_rtx = %d\n", wc->ztsnd_0010_rtx);
						
			for(i = 0; i < wc->numchannels; i++)
			{
				ztc_en = &(wc->uencode->channels[i]);
				st_en = ztc_en->pvt;
			
				ztc_de = &(wc->udecode->channels[i]);
				st_de = ztc_de->pvt;

				printk("shcdc debug: en[%d] snt = %d, rcv = %d [%d]\n", i, st_en->packets_sent, st_en->packets_received, st_en->packets_sent - st_en->packets_received);
				printk("shcdc debug: de[%d] snt = %d, rcv = %d [%d]\n", i, st_de->packets_sent, st_de->packets_received, st_de->packets_sent - st_de->packets_received);
			}
		}
		
		wc->dead = 1;

		dahdi_transcoder_unregister(wc->udecode);
		dahdi_transcoder_unregister(wc->uencode);
		dahdi_transcoder_free(wc->uencode);
		dahdi_transcoder_free(wc->udecode);
		vfree(wc->uencode->channels[0].pvt);
		vfree(wc->udecode->channels[0].pvt);
	
		wc->irq_enable = 0;
		/* Stop any DMA */
		kfree(wc->tmpreadchunk[0]);
		kfree(wc->tmpreadchunk[1]);
		kfree(wc->tmpreadchunk[2]);
		kfree(wc->tmpwritechunk[0]);
		kfree(wc->tmpwritechunk[1]);
		kfree(wc->tmpwritechunk[2]);
		//plx_dmabuf_free(wc, &wc->confbuffer);
		shcdc_stop_dma(wc);
		shcdc_wait_ms(5000);

		/* In case hardware is still there */
		shcdc_disable_pci_interrupts(wc);
		shcdc_disable_local_interrupts(wc);


		free_irq(pdev->irq, wc);
		
		ifaces[wc->pos] = NULL;

		shcdc_release(wc);
	}
}

static struct pci_device_id shcdc_pci_tbl[] =
{
	{ 0x10b5, 0x00f7, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (unsigned long) &shcdc_des[2]},
	{ 0x10b5, 0x00f8, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (unsigned long) &shcdc_des[1]},
	{ 0x10b5, 0x00f9, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (unsigned long) &shcdc_des[0]},
	{ 0 }
};

MODULE_DEVICE_TABLE(pci, shcdc_pci_tbl);

static struct pci_driver shcdc_driver =
{
	name:		"shcdc",
	probe:		shcdc_init_one,
	remove:		__devexit_p(shcdc_remove_one),
	suspend:	NULL,
	resume:		NULL,
	id_table:	shcdc_pci_tbl,
};


int shcdc_init(void)
{
	int res;

	res = dahdi_pci_module(&shcdc_driver);
	if (res)
		return -ENODEV;
	return 0;
}

void shcdc_cleanup(void)
{
	pci_unregister_driver(&shcdc_driver);
}



MODULE_DESCRIPTION("Synway CDC-3522A-PCI Driver");
MODULE_AUTHOR("Lin Yu <liny@sanhuid.com>");

#if defined(MODULE_ALIAS)
MODULE_ALIAS("shcdc");
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


module_param(debug, int, S_IRUGO | S_IWUSR);
module_param(CountREC, int, S_IRUGO | S_IWUSR);
module_param(CountTRI, int, S_IRUGO | S_IWUSR);
module_param(IntCount, int, S_IRUGO | S_IWUSR);
module_param(IntCount1, int, S_IRUGO | S_IWUSR);
module_param(DisalbeAllCh, int, S_IRUGO | S_IWUSR);

module_init(shcdc_init);
module_exit(shcdc_cleanup);
