#ifndef SHCDC_H
#define SHCDC_H

#define BINFILE_DIR "/usr/local/ASTBin/cdcx522apci.bin"

#define INTR_TYPE_NONE                  0              // Interrupt identifiers
#define INTR_TYPE_LOCAL_1               (1 << 0)
#define INTR_TYPE_LOCAL_2               (1 << 1)
#define INTR_TYPE_PCI_ABORT             (1 << 2)
#define INTR_TYPE_DOORBELL              (1 << 3)
#define INTR_TYPE_OUTBOUND_POST         (1 << 4)
#define INTR_TYPE_DMA_0                 (1 << 5)
#define INTR_TYPE_DMA_1                 (1 << 6)
#define INTR_TYPE_DMA_2                 (1 << 7)
#define INTR_TYPE_DMA_3                 (1 << 8)
#define INTR_TYPE_SOFTWARE              (1 << 9)

#define PCI9054_DMA0_MODE		0x80
#define PCI9054_DMA1_MODE		0x94
#define PCI9054_OUTPOST_INT_STAT	0x30
#define PCI9054_OUTPOST_INT_MASK	0x34
#define PCI9054_INT_CTRL_STAT		0x68
#define PCI9054_FIFO_CTRL_STAT		0xe8
#define PCI9054_EEPROM_CTRL_STAT	0x6c
#define PCI9054_LOCAL_DMA_ARBIT		0x08
#define PCI9054_DMA_THRESHOLD		0xb0
#define PCI9054_DMA_COMMAND_STAT	0xa8

#define PCI9054_DMA0_PCI_DAC 		0xb4
#define PCI9054_DMA0_DESC_PTR		0x90
#define PCI9054_DMA0_PCIADDR		0x84
#define PCI9054_DMA0_LOCADDR		0x88
#define PCI9054_DMA0_SIZE		0x8c

#define PCI9054_DMA1_PCI_DAC 		0xb8
#define PCI9054_DMA1_DESC_PTR		0xa4
#define PCI9054_DMA1_PCIADDR		0x98
#define PCI9054_DMA1_LOCADDR		0x9c
#define PCI9054_DMA1_SIZE		0xa0

#define	BFA_PORT_ADDR		0x0
#define	BFB_PORT_ADDR		0x10
#define	BFC_PORT_ADDR		0x20
#define	BFA_DATA_ADDR		0x30
#define	BFB_DATA_ADDR		0x32
#define	BFC_DATA_ADDR		0x34
#define	BF_HWAIT_ADDR		0x36
#define	BF_RESET_ADDR		0x38
#define BFA_MEM_ADDR		0x10000
#define BFB_MEM_ADDR		0x20000
#define BFC_MEM_ADDR		0x30000

#define WC_MAX_IFACES		128

#define MAX_CHANNELS		150
#define PER_CHANNEL		50

#define NUM_CARDS		24
#define NUM_CHANNELS		120
#define CHANNEL_BUF_SIZE	188

#define CDC_FORMAT_ULAW		0x00
#define CDC_FORMAT_ALAW		0x01
#define CDC_FORMAT_G729A	0x02
#define CDC_FORMAT_UNDEF	0xFF

#define G729_FRAME_LEN 10
#define AULAW_FRAME_LEN 80


/***********************************************
	defien data address mapped in bf527 memory
************************************************/

#define Q0R_BASE_ADDR	0xFF902000
#define Q0T_BASE_ADDR	0xFF905000


#define QOR_W_ADDR	0xFF803F00
#define QOR_R_ADDR	0xFF803F02
#define QOT_W_ADDR	0xFF803F04
#define QOT_R_ADDR	0xFF803F06
#define AU_FLAG_ADDR	0xFF803F08
#define QORT_LEN	4800


#define MAX_DMA_RW_SIZE	0x10000
#define MAX_DSP_NUMBER	3


#define HOST_IRQ_COM		0x1C
#define DMA_FINISH_COM		0x2C
#define DMA_READ_COM		0x00a9
#define DMA_WRITE_COM		0x00ab
#define DMA_CONFIG_LENGTH	0xE
#define	BUFFER_LENGTH		0x0080
#define	BUF_INF_LOW		0x0000
#define BUF_INF_HIGH		0x0004

#define READ_BUF_LOW		0x0000
#define	READ_BUF_HIGH		0x0008
#define	WRITE_BUF_LOW		0x0100
#define WRITE_BUF_HIGH		0x0008

#define HEAD_LENGTH		(sizeof(HEAD_INF))
#define BUF_INF_LENGTH		(sizeof(BUF_INF))
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*				define error code
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define GENERIC_OK		0x0000
#define GENERIC_ERROR		0xFFFF

#define DMA_WRITE_ERROR		0xFF01
#define DMA_READ_ERROR		0xFF02
#define DMA_WNR_ERROR		0xFF03 //DMA write and read inconstant

//MT90866 ERROR
#define MT90866_WNR_ERROR	0xFF10 //MT90866 write and read inconstant
#define CHANNRL_LINK_ERROR	0xFF11

//DSP ERROR
#define DSP_SELECT_ERROR	0xFF20
#define LOADER_MODE_ERROR	0xFF21
#define DSP_WNR_ERROR		0xFF22 //DSP write and read inconstant
#define BLACKFIN_SELECT_ERROR	0xFF23
//FILE OPERATION ERROR
#define FILE_OPEN_ERROR		0xFF30
#define FILE_CORRUPT_ERROR	0xFF31
#define FILE_POINT_ERROR	0xFF32
#define MEMORY_MALLOC_ERROR	0xFF33
#define FILE_READ_ERROR		0xFF34
#define FILE_BLOCK_SIZE_ERROR	0xFF35
#define FILE_LENGTH_ERROR	0xFF36
#define FILE_CRC_ERROR		0xFF37
#define CHECK_FLASH_ERROR	0xFF40
#define WRITE_FLASH_ERROR	0xFF41
//HOST DMA OPERATION ERROR
#define DISALLOW_CONFIG_ERROR	0xFF50
#define DMA_TIMEOUT_ERROR	0xFF51
#define DMA_NOT_READY_ERROR	0xFF52
#define BF_NOT_RUNNING_ERROR	0xFF53
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*                               TYPES
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef enum  _DMA_CHANNEL
{
	DMA_CHANNEL0,
	DMA_CHANNEL1,
}DMA_CHANNEL;

typedef enum  _DMA_STATE
{
	DMA_STATE_CLOSED,
	DMA_STATE_BLOCK,
	DMA_STATE_SGL,
}DMA_STATE;

typedef enum _DMA_CHANNEL_PRIORITY
{
	CHANNEL_0_HIGHEST,
	CHANNEL_1_HIGHEST,
	CHANNEL_2_HIGHEST,
	CHANNEL_3_HIGHEST,
	ROTATIONAL,
}DMA_CHANNEL_PRIORITY;

struct cdc_state
{
	int encoder;	/* If we're an encoder */
	struct shcdc *wc;

	unsigned int timeslot_in_num;
	unsigned int timeslot_out_num;

	unsigned int chan_in_num;
	unsigned int chan_out_num;
	
	unsigned int packets_sent;
	unsigned int packets_received;

	unsigned int last_cdc_seqno;
	unsigned int cdc_seqno_rcv;

	unsigned char ssrc;
};

struct shcdc_desc
{
	char *name;
	int flags;
};

struct plx_phys_mem_object
{
	struct list_head listentry;
	void *owner;
	void *kernelva;
	u64 cpuphysical;
	u64 busphysical;
	u32 size;
};

struct dma_channel_desc
{
	unsigned int enablereadyinput	:1;
	unsigned int enablebterminput	:1;
	unsigned int enableiopburst	:1;
	unsigned int enablewriteinvalidmode	:1;
	unsigned int enabledmaeoppin	:1;
	unsigned int dmastoptransfermode:1;
	unsigned int holdiopaddrconst	:1;
	unsigned int demandmode		:1;
	unsigned int enabletransferconutclear	:1;
	unsigned int waitstates		:4;
	unsigned int iopbuswidth	:1;
	unsigned int eotendlink		:1;
	unsigned int validstopcontrol	:1;
	unsigned int validmodeenable	:1;
	unsigned int enabledualaddresscycles:1;
	unsigned int tholdforiopwrites	:4;
	unsigned int tholdforiopreads	:4;
	unsigned int tholdforpciwrites	:4;
	unsigned int tholdforpcireads	:4;
	DMA_CHANNEL_PRIORITY dmachannelpriority;
};

struct plx_reg_data {
	struct shcdc *wc;
	unsigned long offset;
	unsigned long bitstoset;
	unsigned long bitstoclear;
};


struct dma_transfer_element
{
	union
	{
		u64 userva;
		u32 pciaddrlow;
	}u;
	u32 pciaddrhigh;
	u32 localaddr;
	u32 transfercount;
	u32 terminalcountintr :1;
	u32 localtopcidma :1;
	
	int Ndsp;
	
};

struct dma_channel_info
{
	DMA_STATE state;
	void *owner;
	unsigned char pending;
	unsigned char localaddrconstant;
	u32 numpages;
	u32 initialoffset;
	u32 buffersize;
	int direction;
	struct page **pagelist;
	struct plx_phys_mem_object sglbuffer;
	struct plx_phys_mem_object dspbuffer;
	struct dma_transfer_element dmadata[MAX_DSP_NUMBER];
	u32	dmadatacount;
	u32	timeout;
};

struct shcdc
{
	struct pci_dev *dev;
	char *variety;
	unsigned int rxints;
	unsigned int txints;
	unsigned int intmask;
	int pos;
	spinlock_t reglock;
	wait_queue_head_t regq;
	int rcvflags;
	int order;
	int irq_enable;
	unsigned long interruptsrc;
	int dead;
	int BF527Init;
	unsigned short dstfmt[MAX_CHANNELS];

	u32 serialno;

	int CurDspNum;
	int CurChNum;
// 	int BeginCh;

	int pQ0Tr[MAX_DSP_NUMBER];
	int pQ0Tw[MAX_DSP_NUMBER];
	int pQ0Rw[MAX_DSP_NUMBER];
	int pQ0Rr[MAX_DSP_NUMBER];
	int EnableWriteDsp[MAX_DSP_NUMBER];
	int EnableReadDsp[MAX_DSP_NUMBER];
	
	struct semaphore chansem;
	struct semaphore chansemclose;
	struct semaphore tmpwritesem[3];
	struct semaphore tmpreadsem[3];

	unsigned int last_seqno;
	unsigned int last_rseqno;
	unsigned int last_rparm1;
	unsigned int seq_num;
	long timeout;

	unsigned int dsp_crashed;
	unsigned int dumping;

	unsigned int ztsnd_rtx;
	unsigned int ztsnd_0010_rtx;

	unsigned char numchannels;

	int freeregion_dsp;
	int freeregion_plx;
	unsigned long plx_region;
	unsigned long plx_len;
	unsigned long dsp_region;
	unsigned long dsp_len;
	unsigned long dsp_addr;
	volatile void 	*plx;
	volatile void 	*dsp;

	unsigned char *treadchunk;
	unsigned char *twritechunk;
	
	unsigned char *tmpreadchunk[3];
	unsigned char *tmpwritechunk[3];
	int tmpwritechunk_offset[3];
	int tmpreadchunk_offset[3];
	int Stmpwritechunk_offset[3];
	int Stmpreadchunk_offset[3];
	unsigned char *writechunk[3];
	unsigned char *readchunk[3];
	int writechunk_offset[3];
	int readchunk_offset[3];

	struct dma_channel_info dmainfo[2];		//has two DMA channel	0:write 1:read

	struct plx_phys_mem_object confbuffer;
	
	int wqueints;
	struct workqueue_struct *cdc_wq;
	struct work_struct cdc_work;
	//struct work_struct cdc_readchunk;

	struct dahdi_transcoder *uencode;
	struct dahdi_transcoder *udecode;
};


typedef struct _DMA_PARAMS_
{
	unsigned short	Config;
	unsigned short	LowAddr;
	unsigned short	HighAddr;
	unsigned short	XCount;
	unsigned short	XModify;
	unsigned short	YCount;
	unsigned short	YModify;

	unsigned short	DataAddr;
	unsigned short	PortAddr;
	u32		MemAddr;
	unsigned short	*pData;
	u64		DmaBusPhysical;
}DMA_PARAMS, *pDMA_PARAMS;

typedef struct _HEAD_INF_
{
	u32	ulBlockCode;
	u32	ulTargetAddr;
	u32	ulByteCount;
	u32	ulArgument;

}HEAD_INF, *pHEAD_INF;

typedef struct _BUF_INF_
{
	unsigned short	HostReadPoint;
	unsigned short	HostWritePoint;
	unsigned short	SlaveReadPoint;
	unsigned short	SlaveWritePoint;
	u32	InterruptTimes;
	u32	ModifyPointTimes;
}BUF_INF, *pBUF_INF;

#endif
