#ifndef SHFXM8A_H
#define SHFXM8A_H

#define BINFILE_DIR	"/usr/local/ASTBin"

#define INTR_TYPE_NONE                  0              // Interrupt identifiers
#define INTR_TYPE_LOCAL_1               (1 << 0)
#define INTR_TYPE_LOCAL_2               (1 << 1)
#define INTR_TYPE_PCI_ABORT             (1 << 2)
#define INTR_TYPE_DOORBELL              (1 << 3)
#define INTR_TYPE_OUTBOUND_POST         (1 << 4)
#define INTR_TYPE_DMA_0                 (1 << 5)
#define INTR_TYPE_DMA_1                 (1 << 6)
#define INTR_TYPE_DMA_2                 (1 << 7)
#define INTR_TYPE_DMA_3                 (1 << 8)
#define INTR_TYPE_SOFTWARE              (1 << 9)


#define PCI9054_DMA0_MODE		0x80
#define PCI9054_DMA1_MODE		0x94
#define PCI9054_OUTPOST_INT_STAT	0x30
#define PCI9054_OUTPOST_INT_MASK	0x34
#define PCI9054_INT_CTRL_STAT		0x68
#define PCI9054_FIFO_CTRL_STAT		0xe8
#define PCI9054_EEPROM_CTRL_STAT	0x6c
#define PCI9054_LOCAL_DMA_ARBIT		0x08
#define PCI9054_DMA_THRESHOLD		0xb0
#define PCI9054_DMA0_PCI_DAC 		0xb4
#define PCI9054_DMA1_PCI_DAC 		0xb8
#define PCI9054_DMA_COMMAND_STAT	0xa8
#define PCI9054_DMA0_DESC_PTR		0x90
#define PCI9054_DMA1_DESC_PTR		0xa4

#define FLAG_DO_OFFHOOK			(1 << 0)
#define FLAG_DO_ECHOCAN			(1 << 1)
#define FLAG_DO_SENDRING		(1 << 2)

/*Dsp Addr	(Word)*/
//CONF 1
/*
#define SET_CW2				0x1DC8
#define SET_RPCW			0x1DD0
#define SET_M3				0x1DD8
#define BASEADDR_PLAYBUF		0x1E80
#define PLAYBUF_FLAG			0x1F04
#define BASEADDR_RECBUF			0x1F48
#define RECBUF_FLAG			0x1FCC
#define BLOCK_LEN			33
#define BLOCK_NUM			4
*/

/*Dsp Addr	(Word)	ec64*/

#define SET_CW2							0x35C8
#define SET_RPCW						0x35D0
#define SET_M3							0x35D8
#define BASEADDR_PLAYBUF		0x3670
#define PLAYBUF_FLAG			0x36f4
#define PLAYBLOCK_LEN				33
#define PLAYBLOCK_NUM				4
#define BASEADDR_RECBUF			0x3710
#define RECBUF_FLAG					0x3894
#define RECBLOCK_LEN				97
#define RECBLOCK_NUM				4
#define	RECDATA_NUM					3

#define PLAYBLOCK_BUF				4
#define PLAYLIMIT_BUF				3
#define RECBLOCK_BUF				8
#define RECLIMIT_BUF				5

/*
#define SET_CW2				0x35C8
#define SET_RPCW			0x35D0
#define SET_M3				0x35D8
#define BASEADDR_PLAYBUF		0x3680
#define PLAYBUF_FLAG			0x380C
#define BASEADDR_RECBUF			0x3818
#define RECBUF_FLAG			0x39A4
#define BLOCK_LEN			33
#define BLOCK_NUM			12
#define	MAXCACHEARRAY			12*/

#define NUM_CARDS 8
#define MOD_TYPE_FXS	0
#define MOD_TYPE_FXO	1
#define RING_DEBOUNCE		4		/* Ringer Debounce (64 ms) */
#define DEFAULT_BATT_DEBOUNCE	4		/* Battery debounce (64 ms) */
#define POLARITY_DEBOUNCE 	4		/* Polarity debounce (64 ms) */
#define DEFAULT_BATT_THRESH	3		/* Anything under this is "no battery" */
#define WC_MAX_IFACES 128

struct dspbindata {
	char name[256];
	void *dsp_data;
};

typedef struct binheader {
	unsigned short bin_type;
	unsigned short start_addr;
	unsigned short block_num;
	unsigned short data[0];
}binheader;

typedef struct binblock {
	unsigned short block_addr;
	unsigned short block_size;
	unsigned short data[0];
}binblock;

struct plx_reg_data {
	struct shfxm *wc;
	unsigned long offset;
	unsigned long bitstoset;
	unsigned long bitstoclear;
};

typedef enum  _DMA_CHANNEL {
	DMA_CHANNEL0,
	DMA_CHANNEL1,
} DMA_CHANNEL;

typedef enum  _DMA_STATE {
	DMA_STATE_CLOSED,
	DMA_STATE_BLOCK,
	DMA_STATE_SGL,
} DMA_STATE;

typedef enum _DMA_CHANNEL_PRIORITY {
	CHANNEL_0_HIGHEST,
	CHANNEL_1_HIGHEST,
	CHANNEL_2_HIGHEST,
	CHANNEL_3_HIGHEST,
	ROTATIONAL,
} DMA_CHANNEL_PRIORITY;

struct dma_channel_desc {
	unsigned int enablereadyinput	:1;
	unsigned int enablebterminput	:1;
	unsigned int enableiopburst	:1;
	unsigned int enablewriteinvalidmode	:1;
	unsigned int enabledmaeoppin	:1;
	unsigned int dmastoptransfermode:1;
	unsigned int holdiopaddrconst	:1;
	unsigned int demandmode		:1;
	unsigned int enabletransferconutclear	:1;
	unsigned int waitstates		:4;
	unsigned int iopbuswidth	:2;
	unsigned int eotendlink		:1;
	unsigned int validstopcontrol	:1;
	unsigned int validmodeenable	:1;
	unsigned int enabledualaddresscycles:1;
	unsigned int tholdforiopwrites	:4;
	unsigned int tholdforiopreads	:4;
	unsigned int tholdforpciwrites	:4;
	unsigned int tholdforpcireads	:4;
	DMA_CHANNEL_PRIORITY dmachannelpriority;
};

struct plx_phys_mem_object {
	struct list_head listentry;
	void *owner;
	void *kernelva;	
	u64 cpuphysical;
	u64 busphysical;
	u32 size;
};

struct dma_transfer_element {
	union {
		u64 userva;
		u32 pciaddrlow;
	} u;
	u32 pciaddrhigh;
	u32 localaddr;
	u32 transfercount;
	u32 terminalcountintr :1;
	u32 localtopcidma :1;
};

struct dma_channel_info {
	DMA_STATE state;
	void *owner;
	unsigned char localaddrconstant;
	u32 numpages;
	u32 initialoffset;
	u32 buffersize;
	int direction;
	struct page **pagelist;
	struct plx_phys_mem_object sglbuffer;
	struct plx_phys_mem_object dspbuffer;
	struct dma_transfer_element dmadata[600];
	u32	dmadatacount;
	u32	timeout;
};

struct dsp_writechunk_frame {
	unsigned short writeflag;
	unsigned char  writechunkbuf[0];
};
struct dsp_readchunk_frame {
	unsigned short readflag;
	unsigned char  readchunkbuf[0];
};

struct shfxm {
	struct pci_dev *dev;
	char *variety;
	struct dahdi_span span;
	int usecount;
	unsigned int intcount;
	int dead;
	int pos;
	int flags[NUM_CARDS];
	int freeregion_dsp;
	int freeregion_plx;
	int alt;
	int curcard;
	int cardflag;
	int irq_enable;
	u32 serialno;
	int tflags;
	int Ttflags;
	unsigned short v3d2;
	unsigned short v36c5;
	
	unsigned short RECFlag;
	unsigned short TRIFlag;
	int RECFlag_First;
	int TRIFlag_First;
	
	spinlock_t lock;

	union {
		struct {
			int offhook;
			unsigned int flags;
			unsigned char voltage;
			int wasringing;
			int ringdebounce;
			int battdebounce;
			int nobatttimer;
			int battery;
		        int lastpol;
		        int polarity;
		        int polaritydebounce;
		} fxo;
		struct {
			int offhook;
			int oldrxhook;
			int debouncehook;
			int lastrxhook;
			int debounce;
			unsigned int flags;
		} fxs;
	} mod[NUM_CARDS];

	int modtype[NUM_CARDS];

	unsigned long plx_region;
	unsigned long plx_len;

	unsigned long dsp_region;
	unsigned long dsp_len;

	unsigned long dsp_addr;

	unsigned long interruptsrc;
	
	volatile void 	*plx;
	volatile void 	*dsp;

	unsigned short regringhook;
	unsigned short echocan;

	struct dsp_readchunk_frame *readchunk_frame[RECBLOCK_BUF];
	struct dsp_writechunk_frame *writechunk_frame[PLAYBLOCK_BUF];
	unsigned char *readchunk[RECBLOCK_BUF][RECDATA_NUM][NUM_CARDS];
	unsigned char *writechunk[PLAYBLOCK_BUF][NUM_CARDS];

	int readcurpos;
	int readfillpos;
	int rpos;
	int readhasdata[RECBLOCK_BUF];
	int writecurpos;
	int writefillpos;
	int writehasdata[PLAYBLOCK_BUF];
	
	struct dma_channel_info dmainfo[2];
	struct dahdi_chan* _chans[NUM_CARDS];
	struct dahdi_chan  chans[NUM_CARDS];
	
	/*dsp voice debug*/
	int complete;
	int Index;
	int filecount;
	unsigned char * tmpwrite[2][NUM_CARDS];
	unsigned char * tmpwriter[2][RECDATA_NUM][NUM_CARDS];
	wait_queue_head_t regq;
	struct workqueue_struct *shfxm_wq;
	struct work_struct shfxm_work;
	/*EC debug*/
	unsigned short ecval[NUM_CARDS];
	
	int BTime;
	struct timeval stime, etime, qtime;
};



struct shfxm_desc {
	char *name;
	int flags;
};

#endif
