/*
 * TEJ-4a\TEJ-2a\TEJ-1a PCI Driver version 1.0, 2008-10-15
 *
 * Written by Xiong ce <liny@sanhuid.com>
 * Copyright (C) 2008 Synway / Zapata Telephony.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

/*parameter TEJMODE (0(E1) 1(T1) 2(J1))*/
 
#include <linux/kernel.h>
#include <linux/errno.h> 
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#ifdef LINUX26
#include <linux/moduleparam.h>
#endif
#include "dahdi/kernel.h"

#include "tejxapci.h"

#define TYPE_T1 1
#define TYPE_E1 2
#define TYPE_J1 1

#define MAX_CARDS 64

#define FLAG_2PORT (1 << 3)

static u32 timing = 0;
static u32 oldtiming = 0;
static int debug = 0;
static int timingcable;
static int TEJMODE = 0;
static int T1S[] = {3, 7, 11, 15, 19, 23, 27, 31};

static struct shtej_desc shtej[3] = { { "TEJ-4a/PCI Prototype", 0 }, { "TEJ-2a/PCI Prototype", 0 }, { "TEJ-1a/PCI Prototype", 0 }};

static struct shtej *cards[MAX_CARDS];

static void e1_configure(struct shtej *wc, unsigned int ndsp, u32 nSel, int lineconfig);
static void t1_configure(struct shtej *wc, unsigned int ndsp, u32 nSel, int lineconfig);
static void shtej_WriteDS2652X(struct shtej *wc, u32 nSel, unsigned short RegAddr, unsigned short Data);


static void DebugInt(struct shtej *wc)
{
	int tmp;

	if(wc->BTime == 0)
	{
		do_gettimeofday(&wc->stime);
		wc->BTime = 1;
	}
	else
	{
		do_gettimeofday(&wc->etime);
		tmp = (wc->etime.tv_sec - wc->stime.tv_sec)*1000000 + (wc->etime.tv_usec - wc->stime.tv_usec);
		wc->BTime = 0;
		if(tmp > 1200)
			printk("xxx-----=%d\n", tmp);
	}
}

int kernel_write(int count, unsigned char * addr, char * filename)
{
	struct file *file;
	mm_segment_t old_fs;
	loff_t pos = 0;
	int  result = -ENOSYS;

	file   =   filp_open(filename,   O_RDWR   |   O_APPEND   |   O_CREAT, 0644);
	if   (IS_ERR(file))
	{
		return   0;
	}

	if (!file->f_op->write)
	{
		return result;
	}
	old_fs = get_fs();
	set_fs(get_ds());
	result = file->f_op->write(file, addr, count, &pos);
	set_fs(old_fs);
	
	filp_close(file, NULL);

	return result;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
static void wque_run(struct work_struct *work)
{
	struct shtej *wc = container_of(work, struct shtej, shtej_work);
#else
static void wque_run(void *work_data)
{
	struct shtej *wc = (struct shtej *)work_data;
#endif
	char tmpname[50];
	int t;
	int r;
	int i;
	
	do_gettimeofday(&wc->qtime);
	t = (wc->Index+1)%2;

	for(i=0; i<31; i++)
	{
			sprintf(tmpname, "/mnt/c/RecEc%d_%d_%05d.pcm", wc->num, i, (int)wc->qtime.tv_sec);
			r = kernel_write(0x20000, wc->tmpwriterec[t][i], tmpname);
	}
	
	for(i=0; i<31; i++)
	{
			sprintf(tmpname, "/mnt/b/T%d_%d_%05d.pcm", wc->num, i, (int)wc->qtime.tv_sec);
			r = kernel_write(0x20000, wc->tmpwrite[t][i], tmpname);
	}
	
	wc->complete = 1;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
static void wque_rec_span(struct work_struct *work)
{
	struct shtej *wc = container_of(work, struct shtej, rec_span_work);
#else
static void wque_rec_span(void *work_data)
{
	struct shtej *wc = (struct shtej *)work_data;
#endif
	int i;
	for (i = 0; i < wc->numspans; i++)
	{
		if (wc->tspans[i]->span.flags & DAHDI_FLAG_RUNNING)
			dahdi_receive(&wc->tspans[i]->span);
	}
}


static inline int pci_getconfig_dword(struct shtej *wc, int offset, u32 *val)
{
	return pci_read_config_dword(wc->dev, offset, val);
}

static inline int pci_setconfig_dword(struct shtej *wc, int offset, u32 val)
{
	return pci_write_config_dword(wc->dev, offset, val);
}

static inline unsigned long plx_getreg(struct shtej *wc, unsigned long offset)
{
	return	readl(wc->plx + offset);
}

static inline void plx_setreg(struct shtej *wc, unsigned long offset, unsigned long val)
{
	writel(val, wc->plx + offset);
} 

static unsigned short dsp_getreg_word(struct shtej *wc, int offset)
{
	return readw(wc->dsp + offset);
}

static void dsp_setreg_word(struct shtej *wc, int offset, unsigned short val)
{
	writew(val, wc->dsp + offset);
}

static int shtej_setdsp_data(struct shtej *wc, int dspsel, unsigned short addr, unsigned short *data, unsigned long len)
{
	unsigned short highaddr;
	unsigned short lowaddr;
	unsigned long i;

	if (addr % 2 == 1 && dspsel > 0)
	{
		return -1;
	}

	for (i = 0; i < len; i++, addr += 2)
	{
		highaddr = addr & 0xff00;
		lowaddr = addr & 0xff;
		if (wc->dsp_addr != (0x10000 * dspsel + highaddr + 0x1))
		{
			wc->dsp_addr = 0x10000 * dspsel + highaddr + 0x1;
			plx_setreg(wc, 0x4, wc->dsp_addr);
		}
		dsp_setreg_word(wc, lowaddr, *(data + i));
	}
	return 0;

}

static int shtej_getdsp_data(struct shtej *wc, int dspsel, unsigned short addr, unsigned short *data, unsigned long len)
{
	unsigned short highaddr;
	unsigned short lowaddr;
	unsigned long i;

	if (addr % 2 == 1 && dspsel > 0)
	{
		return -1;
	}

	for (i = 0; i < len; i++, addr += 2)
	{
		highaddr = addr & 0xff00;
		lowaddr = addr & 0xff;
		if (wc->dsp_addr != (0x10000 * dspsel + highaddr + 0x1))
		{
			wc->dsp_addr = 0x10000 * dspsel + highaddr + 0x1;
			plx_setreg(wc, 0x4, wc->dsp_addr);
		}
		*(data + i) = dsp_getreg_word(wc, lowaddr);
	}

	return 0;
}

static void plx_sync_reg_modify(struct plx_reg_data *reg)
{
	unsigned long regval;

	regval = plx_getreg(reg->wc, reg->offset);
	regval |= reg->bitstoset;
	regval &= ~(reg->bitstoclear);
	plx_setreg(reg->wc, reg->offset, regval);

}

static void shtej_enable_cpld_interrupts(struct shtej *wc)
{
	unsigned short val = 1;

	shtej_setdsp_data(wc, 0, 0x2c, &val, sizeof(val) / sizeof(unsigned short));
}

static void shtej_disable_cpld_interrupts(struct shtej *wc)
{
	unsigned short val = 0;

	shtej_setdsp_data(wc, 0, 0x2c, &val, sizeof(val) / sizeof(unsigned short));
}

static void shtej_enable_pci_interrupts(struct shtej *wc)
{
	struct plx_reg_data regdata;
	unsigned long regval;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoset = (1 << 8);
	regdata.bitstoclear = 0;

	plx_sync_reg_modify(&regdata);

	regval = plx_getreg(regdata.wc, regdata.offset);

}

static void shtej_disable_pci_interrupts(struct shtej *wc)
{
	struct plx_reg_data regdata;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoset = 0;
	regdata.bitstoclear = (1 << 8);

	plx_sync_reg_modify(&regdata);

}

static void shtej_enable_local_interrupts(struct shtej *wc)
{
	struct plx_reg_data regdata;
	unsigned long regval;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoset = (1 << 11);
	regdata.bitstoclear = 0;

	plx_sync_reg_modify(&regdata);

	regval = plx_getreg(regdata.wc, regdata.offset);

}

static void shtej_disable_local_interrupts(struct shtej *wc)
{
	struct plx_reg_data regdata;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoset = 0;
	regdata.bitstoclear = (1 << 11);

	plx_sync_reg_modify(&regdata);

}

static inline void shtej_wait_ms(long ms)
{
	udelay(ms);
}

static void shtej_SerialNo(struct shtej *wc)
{
	s8 BitPos, count;
	u32 regvalue, EepromCommand, tmpreg;
	
	regvalue = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
	
	regvalue &= ~(0xF << 24);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);
	regvalue |= (1 << 25);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);
	
	EepromCommand = (0x0180<<2) | (0x1f0/2);
	for (BitPos = 10; BitPos >= 0; BitPos--)
	{
		if (EepromCommand & (1 << BitPos))
			plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue | (1 << 26));
		else
			plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);
		
		tmpreg = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, tmpreg | (1 << 24));
		for (count=0; count<20; count++)
			tmpreg = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, tmpreg & ~(1 << 24));
	}
	
	regvalue = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
	regvalue |= (1 << 31);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue | (1 << 26));

	for (BitPos = 0; BitPos < 32; BitPos++)
	{
		tmpreg = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, tmpreg | (1 << 24));
		for (count=0; count<20; count++)
			tmpreg = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, tmpreg & ~(1 << 24));

		for (count=0; count <20; count++)
			regvalue = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);

		if (regvalue & (1 << 27))
			wc->serialno = (wc->serialno << 1) | 1;
		else
			wc->serialno = (wc->serialno << 1);
	}
	
	regvalue &= ~(1 << 31);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue & ~(0xF << 24));
	
	wc->serialno = (wc->serialno >> 16 | wc->serialno << 16);
}



static int shtej_open(struct dahdi_chan *chan)
{
	try_module_get(THIS_MODULE);
	return 0;
}

static int shtej_close(struct dahdi_chan *chan)
{
	module_put(THIS_MODULE);

	return 0;
}

static void shtej_echoconf(struct shtej *wc, int Ndsp)
{
	unsigned short res;
	unsigned short ulAddr;

#ifndef ECHO_ALWAYS_OPEN
	ulAddr = 0x6d80 * 2;
	if(TEJMODE == 0)
		res = 0x0000;
	else
		res = 0x0000;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));

	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0x0000;
	else
		res = 0x0001;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0x0000;
	else
		res = 0x0000;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));

	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0x0000;
	else
		res = 0x0001;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0x0000;
	else
		res = 0x0000;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));

	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0x0000;
	else
		res = 0x0001;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));

#else

	ulAddr = 0x6d80 * 2;
	if(TEJMODE == 0)
		res = 0x0000;
	else
		res = 0x0000;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));

	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0x0000;
	else
		res = 0x0001;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0xfffe;
	else
		res = 0x7fff;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));

	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0xfffe;
	else
		res = 0xffff;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0xfffe;
	else
		res = 0x7fff;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));

	ulAddr = ulAddr + 2;
	if(TEJMODE == 0)
		res = 0xfffe;
	else
		res = 0xffff;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
#endif

	ulAddr = ulAddr + 2;
	res = 0xffff;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));

	ulAddr = ulAddr + 2;
	res = 0xffff;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
		
	ulAddr = ulAddr + 2;
	res = Ndsp-1;
	shtej_setdsp_data(wc, Ndsp, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
}

static int shtej_startup(struct dahdi_span *span)
{
	int tspan;
	unsigned long flags;
	int alreadyrunning;
	struct shtej_span *p = span->pvt;

	tspan = p->span.offset;
	if (tspan < 0)
	{
		printk("shtej: Span '%d' isn't us?\n", span->spanno);
		return -1;
	}

	spin_lock_irqsave(&p->owner->lock, flags);

	alreadyrunning = span->flags & DAHDI_FLAG_RUNNING;

	/* Force re-evaluation of the timing source */
	if (timingcable)
		p->owner->syncsrc = -1;

	if (p->owner->cardtype == TYPE_E1)
	{
		e1_configure(p->owner, tspan, 0, span->lineconfig);

		if (!alreadyrunning)
			span->flags |= DAHDI_FLAG_RUNNING;

		spin_unlock_irqrestore(&p->owner->lock, flags);

		if (debug)
		{
			if (alreadyrunning)
				printk("shtej: Reconfigured span %d 120 Ohms\n", span->spanno);
			else
				printk("shtej: Startup span %d 120 Ohms\n", span->spanno);
		}
	}
	else
	{
		t1_configure(p->owner, tspan, 0, span->lineconfig);
	
		if (!alreadyrunning)
			span->flags |= DAHDI_FLAG_RUNNING;

		spin_unlock_irqrestore(&p->owner->lock, flags);

		if (debug)
		{
			if (alreadyrunning)
				printk("shtej: Reconfigured span %d LBO: %s\n", span->spanno, dahdi_lboname(span->txlevel));
			else
				printk("shtej: Startup span %d LBO: %s\n", span->spanno, dahdi_lboname(span->txlevel));
		}
	}
	
	if(p->owner->tspans[0]->sync == span->spanno) printk("SPAN %d: Primary Sync Source\n",span->spanno);
	
	if(p->owner->numspans == 2 || p->owner->numspans == 4)
		if (p->owner->tspans[1]->sync == span->spanno) printk("SPAN %d: Secondary Sync Source\n",span->spanno);
	
	if(p->owner->numspans == 4)
	{
		if (p->owner->tspans[2]->sync == span->spanno) printk("SPAN %d: Tertiary Sync Source\n",span->spanno);
		if (p->owner->tspans[3]->sync == span->spanno) printk("SPAN %d: Quaternary Sync Source\n",span->spanno);
	}
	return 0;
}

static int shtej_shutdown(struct dahdi_span *span)
{
	return 0;
}

static int shtej_dorbsbits(struct shtej_span *ts)
{
	int i;
	unsigned short	c;

	if(ts->owner->cardtype == TYPE_E1)
	{
		for(i=0; i<16; ++i)
		{
			if(ts->dorbs[i]	== 1)
			{
				c = ts->txsigs[i];
				shtej_WriteDS2652X(ts->owner, 0, ts->index*0x200+0x140+i, c);	//TS1-TS16
				ts->dorbs[i]	= 0;
			}
		}
	}
	else if (ts->span.lineconfig & DAHDI_CONFIG_ESF)
	{
		for(i=0; i<12; ++i)
		{
			if(ts->dorbs[i]	== 1)
			{
				c = ts->txsigs[i];
				shtej_WriteDS2652X(ts->owner, 0, ts->index*0x200+0x140+i, c);	//TS1-TS16
				ts->dorbs[i]	= 0;
			}
		}
	} 
	
	return 0;
}

static int shtej_rbsbits(struct dahdi_chan *chan, int bits)
{
	struct shtej_span *ts = chan->pvt;
	unsigned short	m, c;
	int	n, b;
	
	//printk("shtej_rbsbits----[%s(%d), %d]\n", chan->name, chan->chanpos, bits);

	if(ts->owner->cardtype == TYPE_E1)
	{
		if (chan->chanpos == 16)
			return 0;
		n = chan->chanpos - 1;
		if (chan->chanpos > 16) n--;
		b = (n % 15) + 1;
		c = ts->txsigs[b];
		m = (n / 15) * 4;		// nibble selector
		c &= (15 << m);			// keep the other nibble
		c |= (bits & 15) << (4 - m);	// put our new nibble here
		ts->txsigs[b] = c;
		ts->dorbs[b]	= 1;
	}
	else if (ts->span.lineconfig & DAHDI_CONFIG_ESF)
	{
		n = chan->chanpos - 1;
		b = (n/2);
		c = ts->txsigs[b];
		m = ((n % 2) << 2); /* nibble selector */
		c &= (0xf << m); /* keep the other nibble */
		c |= (bits & 0xf) << (4 - m); /* put our new nibble here */
		ts->txsigs[b] = c;
		ts->dorbs[b]	= 1;
	} 
	
	return 0;
}

static int shtej_maint(struct dahdi_span *span, int cmd)
{
	return 0;
}

static int shtej_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
	int	n, b;
	struct tej_regs regs;
	struct shtej_span *ts = chan->pvt;
	
	switch(cmd)
	{
		case TEJ_SET_RBSREGS:
			if(ts->owner->cardtype == TYPE_E1)
			{
				if (chan->chanpos == 16)
					return 0;
				n = chan->chanpos - 1;
				if (chan->chanpos > 16) n--;
				b = (n % 15) + 1;
				if(copy_from_user(&regs, (struct tej_regs *) data, sizeof(regs)))
					return -EFAULT;
				ts->txsigs[b] = regs.reg;
				ts->dorbs[b] = 1;
			}
			break;
			
		default:
			return -ENOTTY;
	}
	
	return 0;
}

static int shtej_doecho(struct shtej_span *ts)
{
	unsigned short res;
	unsigned short ulAddr;
	
	if(ts->doecho == 1)
	{
		ulAddr = 0x6d80 * 2;
		res = 0x0000;
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
		ulAddr = ulAddr + 2;
		if(TEJMODE == 0)
			res = 0x0000;
		else
			res = 0x0001;
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
		
		ulAddr = ulAddr + 2;
		res = (unsigned short)((ts->echocan>>16) & 0xffff);
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
		ulAddr = ulAddr + 2;
		res = (unsigned short)(ts->echocan & 0xffff);
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
		ulAddr = ulAddr + 2;
		res = (unsigned short)((ts->echocan>>16) & 0xffff);
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
		ulAddr = ulAddr + 2;
		res = (unsigned short)(ts->echocan & 0xffff);
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
		ulAddr = ulAddr + 2;
		res = 0xffff;
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	
		ulAddr = ulAddr + 2;
		res = 0xffff;
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
			
		ulAddr = ulAddr + 2;
		res = ts->index;
		shtej_setdsp_data(ts->owner, ts->index+1, ulAddr, &res, sizeof(res) / sizeof(unsigned short));
	}
	
	ts->doecho = 0;
	return 0;
}

#ifndef ECHOCAN_CREATE
static int shtej_echocan(struct dahdi_chan *chan, int eclen)
{
#ifndef ECHO_ALWAYS_OPEN
	u32 ch;
	struct shtej_span *ts = chan->pvt;
	
	ch = chan->chanpos; 

	if (eclen)
		ts->echocan |= 1 << ch;
  else
		ts->echocan &= ~(1 << ch); 
	if(TEJMODE == 0)
		ts->echocan |= 0;
	else
		ts->echocan |= 1;
	
	ts->doecho = 1;
	
	//printk("shtej_echocan----[%s, %d, %x, %d]\n", chan->name, eclen, ts->echocan, ts->index);
#endif
	
	return 0;
}
#endif
static int shtej_spanconfig(struct dahdi_span *span, struct dahdi_lineconfig *lc)
{
	int i;
	struct shtej_span *ts = span->pvt;
	struct shtej *wc = ts->owner;

	if (debug)
	{
		printk("About to enter spanconfig\t[%s]!\n", span->name);
		printk("TEJ%dXXP: Configuring span %d\n", wc->numspans, span->spanno);
	}
	if (lc->sync < 0)
		lc->sync = 0;
	if (lc->sync > 4)
		lc->sync = 0;

	/* remove this span number from the current sync sources, if there */
	for(i = 0; i < wc->numspans; i++)
	{
		if (wc->tspans[i]->sync == span->spanno)
		{
			wc->tspans[i]->sync = 0;
			wc->tspans[i]->psync = 0;
		}
	}
	wc->tspans[span->offset]->syncpos = lc->sync;
	/* if a sync src, put it in proper place */
	if (lc->sync)
	{
		wc->tspans[lc->sync - 1]->sync = span->spanno;
		wc->tspans[lc->sync - 1]->psync = span->offset + 1;
	}

	wc->checktiming = 1; //add tej timer sync,09/03/30
	/* If we're already running, then go ahead and apply the changes */
	if (span->flags & DAHDI_FLAG_RUNNING)
		return shtej_startup(span);

	return 0;
}

static int shtej_chanconfig(struct dahdi_chan *chan, int sigtype)
{
	unsigned short val;
	struct shtej_span *ts = chan->pvt;
	struct shtej *wc = ts->owner;
	if (sigtype & DAHDI_SIG_EM && ts->sigset == 0)
	{
		ts->sigset = 1;
		shtej_WriteDS2652X(wc, 0,  ts->index*0x200+0x0118, 0xFF); // SSIE1, Enable Transmit Signaling for channels 1-8
		shtej_WriteDS2652X(wc, 0,  ts->index*0x200+0x0119, 0xFF); // SSIE2, Enable Transmit Signaling for channels 9-16
		shtej_WriteDS2652X(wc, 0,  ts->index*0x200+0x011A, 0xFF); // SSIE3, Enable Transmit Signaling for channels 17-24
	}
	
	if(sigtype & 0x300)
		val = 0;
	else
		val = 1;

	shtej_setdsp_data(wc, 1, (0x1380+chan->channo-1) * 2, &val, sizeof(val) / sizeof(unsigned short));
	
	return 0;
}

static void shtej_init_spans(struct shtej *wc)
{
	int x;
	int y;
	struct shtej_span *ts;

	for (x = 0; x < wc->numspans; x++)
	{
		ts =  wc->tspans[x];
		sprintf(ts->span.name, "TEJXAPCI/%d/%d", wc->num, x+1);
		sprintf(ts->span.desc, "TEJXAPCI Card %d Span %d", wc->num, x+1);
		ts->index = x;
		ts->echocan = 0;
		ts->doecho = 0;
		ts->sigset = 0;
		memset(ts->dorbs, 16 , 0);
		ts->span.manufacturer = "Synway";
		ts->span.spanconfig = shtej_spanconfig;
		ts->span.chanconfig = shtej_chanconfig;
		ts->span.startup = shtej_startup;
		ts->span.rbsbits = shtej_rbsbits;
		ts->span.maint = shtej_maint;
		ts->span.open = shtej_open;
		ts->span.close = shtej_close;
		ts->span.shutdown = shtej_shutdown;
		if (ts->spantype == TYPE_J1 || ts->spantype == TYPE_T1)
		{
			ts->span.channels = 24;
			ts->span.deflaw = DAHDI_LAW_MULAW;
			ts->span.linecompat = DAHDI_CONFIG_AMI | DAHDI_CONFIG_B8ZS | DAHDI_CONFIG_ESF | DAHDI_CONFIG_D4;
		}
		else
		{
			ts->span.channels = 31;
			ts->span.deflaw = DAHDI_LAW_ALAW;
			ts->span.linecompat = DAHDI_CONFIG_HDB3 | DAHDI_CONFIG_CCS | DAHDI_CONFIG_CRC4;
		}
		ts->span.chans = ts->_chans;
		ts->span.flags = DAHDI_FLAG_RBS;
		ts->span.ioctl = shtej_ioctl;
		#ifndef ECHOCAN_CREATE
		ts->span.echocan = shtej_echocan;
		#endif
		ts->span.pvt = ts;
		ts->owner = wc;
		ts->span.offset = x;
		init_waitqueue_head(&ts->span.maintq);
		for (y = 0; y < wc->tspans[x]->span.channels; y++)
		{
			struct dahdi_chan *mychans = ts->chans + y;	

			sprintf(mychans->name, "TEJXAPCI%d/%d/%d/%d", wc->numspans, wc->num, x + 1, y + 1);
			mychans->sigcap = DAHDI_SIG_EM | DAHDI_SIG_CLEAR | DAHDI_SIG_FXSLS | DAHDI_SIG_FXSGS | DAHDI_SIG_FXSKS | DAHDI_SIG_FXOLS | DAHDI_SIG_FXOGS | DAHDI_SIG_FXOKS | DAHDI_SIG_CAS | DAHDI_SIG_EM_E1 ;
			mychans->pvt = ts;
			mychans->chanpos = y + 1;
		}
	}
}


static struct dspbindata *shtej_open_dspbinfile(const char *filename)
{
	char fn[256];
	size_t read_len;
	struct file *fp = NULL;
	struct dspbindata *bin;
	unsigned long fileoffset;
	void *data;

	if (filename[0] == '/')
		strncpy(fn, filename, sizeof(fn)-1);
	else
		snprintf(fn, sizeof(fn), "%s/%s", BINFILE_DIR, filename);

	if (!(fp = filp_open(fn, O_RDONLY, S_IRUSR)) || !IS_ERR(fp))
	{

		bin = (struct dspbindata *)kmalloc(sizeof(struct dspbindata), GFP_KERNEL);	
		if (!bin)
		{
			printk("Memory Allocation Failure in  in function %s at line %d of %s\n",
								 __FILE__, __LINE__, __PRETTY_FUNCTION__);
			filp_close(fp , NULL);
			return NULL;
		}

		fileoffset = fp->f_dentry->d_inode->i_size;
		if (fileoffset < 1)
		{
			printk("file offset = NULL %s at line %d of %s\n",
						__FILE__, __LINE__, __PRETTY_FUNCTION__);
			filp_close(fp , NULL);
			kfree(bin);
			return NULL;	
		}

		data = kmalloc(fileoffset, GFP_KERNEL);
		if (!data)
		{
			printk("Memory Allocation Failure in  in function %s at line %d of %s\n",
								 __FILE__, __LINE__, __PRETTY_FUNCTION__);
			filp_close(fp , NULL);
			kfree(bin);
			return NULL;	
		}
			
		read_len = kernel_read(fp, 0, data, fileoffset);
		if (read_len != fileoffset)
		{
			printk("read_len != fileoffset in  in function %s at line %d of %s\n",  
								 __FILE__, __LINE__, __PRETTY_FUNCTION__);
			filp_close(fp , NULL);
			kfree(bin);
			kfree(data);
			return NULL;	
		}
		
		strncpy(bin->name, fn, sizeof(bin->name)-1);
		bin->dsp_data = data;
		filp_close(fp , NULL);
		return bin;


	}
	else
	{
		printk("No file to parse: %s\n", fn);
		return NULL;	
	}
}

static void shtej_close_dspbinfile(struct dspbindata *data)
{
	if (!data)
	{
		if (!data->dsp_data)
			kfree(data->dsp_data);
		kfree(data);
	}
}

static int shtej_init_dsp(struct shtej *wc, const char *filename, int Ndsp)
{
	binheader *h;
	binblock  *b;
	unsigned short i;
	unsigned short j;
	unsigned short res;
	unsigned short *p;
	unsigned short size;
	unsigned short *compare;
	struct dspbindata *bin_data;

	bin_data = shtej_open_dspbinfile(filename);

	if (!bin_data)
	{
		printk("shtej open dspbinfile fail\n");
		return -1;
	}

	h = (binheader *)bin_data->dsp_data;

	if (h->bin_type != 0x003b && h->bin_type != 0x00fe)
	{
		printk("Load Dsp bin Failure in  in function %s at line %d of %s\n",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
		shtej_close_dspbinfile(bin_data);
		return -1;

	}

	b = (binblock *)h->data;
	
	for (i = 0; i < h->block_num; i++)
	{

		if (b->block_size % 2)
			size = b->block_size + 1;
		else
			size = b->block_size;
		p = b->data;
		if (size)
		{

			shtej_setdsp_data(wc, Ndsp, b->block_addr, p, size/2);
			p = b->data;
			compare = (unsigned short *)kmalloc(size, GFP_KERNEL);
			memset(compare, 0x0, size);

			if (!compare)
			{
				printk("Load Dsp bin Failure in  in function %s at line %d of %s\n",
						__FILE__, __LINE__, __PRETTY_FUNCTION__);
				shtej_close_dspbinfile(bin_data);
				return -1;
			}
			shtej_getdsp_data(wc, Ndsp, b->block_addr, compare, size/2);
			for (j = 0; j < (size / 2); j++)
			{
				if (*p++ != *(compare + j))
				{
					printk("Load Dsp bin Failure in  in function %s at line %d of %s\n",
							__FILE__, __LINE__, __PRETTY_FUNCTION__);
					kfree(compare);
					shtej_close_dspbinfile(bin_data);
					return -1;
				}
			}
			kfree(compare);
		}
		p = b->data + size/2;
		b = (binblock *)p;

	}
	shtej_close_dspbinfile(bin_data);
	
	if (Ndsp<=MAXECDSPUSED)
		shtej_echoconf(wc, Ndsp);

	shtej_getdsp_data(wc, Ndsp, 0x60*2, &res, sizeof(res) / sizeof(unsigned short));
	if (res != 0)
	{
		printk("DSP %x is not in bootloader mode ! \n", Ndsp);
		return LOADER_MODE_ERROR;
	}
	
	shtej_getdsp_data(wc, Ndsp, 0x61*2, &res, sizeof(res) / sizeof(unsigned short));
	if (res != 0)
	{
		printk("DSP %x is not in bootloader mode ! \n", Ndsp);
		return LOADER_MODE_ERROR;
	}

	shtej_setdsp_data(wc, Ndsp, 0x61 * 2, &h->start_addr, sizeof(h->start_addr) / sizeof(unsigned short));
	shtej_getdsp_data(wc, Ndsp, 0x61*2, &res, sizeof(res) / sizeof(unsigned short));
	if (res != h->start_addr)
	{
		printk("DSP %x Write/Read err at 0x61. W:%x, R:%x ! \n", Ndsp, h->start_addr, res);
		return DSP_WNR_ERROR;
	}

	return 0;
}

static unsigned long plx_chip_type_get(struct shtej *wc)
{
	unsigned long val;
	unsigned long type;
	unsigned long reg_original;

	reg_original = plx_getreg(wc, 0x04);
	plx_setreg(wc, 0x4, reg_original | (0xff << 24));
	val = plx_getreg(wc, 0x4);
	plx_setreg(wc, 0x4, reg_original);

	if ((val >> 24) == 0)
	{
		printk("shtej: unkown chip type\n");
		return 0;
	}

	if ((val >> 28) == 0xf)
	{
		val = plx_getreg(wc, 0x70);
		if (val == 0x905410b5)
		{
			type = 0x9054;
			return type;
		}
	}

	return 0;
}

static int shtej_plx_init(struct shtej *wc)
{
	u8 mu_enabled;
	u8 eeprompresent;
	u32 regvalue;
	u32 reginterrupt;
	u32 reghotswap;
	u32 regpowermgmnt;


	wc->dsp_addr = 0x1;
	plx_setreg(wc, 0x4, wc->dsp_addr);

	if (plx_chip_type_get(wc) != 0x9054)
	{
		return -1;
	}

	pci_getconfig_dword(wc, 0x4, &regvalue);

	if (regvalue & (0xf8 << 24))
	{
		pci_setconfig_dword(wc, 0x4, regvalue);
	}

	regvalue = plx_getreg(wc, PCI9054_FIFO_CTRL_STAT);
	mu_enabled = (unsigned char)(regvalue & (1 << 0));
	regvalue = plx_getreg(wc, PCI9054_EEPROM_CTRL_STAT);

	regvalue &= ~((1 << 30) | (1 << 29));
	eeprompresent = (unsigned char)((regvalue >> 28) & (1 << 0));

	pci_getconfig_dword(wc, 0x3c, &reginterrupt);

	if (eeprompresent)
	{
		pci_getconfig_dword(wc, 0x48, &reghotswap);
		pci_getconfig_dword(wc, 0x44, &regpowermgmnt);
	}

	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue | (1 << 30));
	shtej_wait_ms(1000);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue | (1 << 29));
	shtej_wait_ms(100);
	plx_setreg(wc, PCI9054_EEPROM_CTRL_STAT, regvalue);
	if (mu_enabled)
	{
		regvalue = plx_getreg(wc, PCI9054_FIFO_CTRL_STAT);
		plx_setreg(wc, PCI9054_FIFO_CTRL_STAT, regvalue | (1 << 0));
	}

	pci_setconfig_dword(wc, 0x3c, reginterrupt);

	if (eeprompresent)
	{
		reghotswap &= ~((1 << 23) | (1 << 22) | (1 << 17));
		pci_setconfig_dword(wc, 0x48, reghotswap);
		regpowermgmnt &= ~(1 << 15);
		pci_getconfig_dword(wc, 0x44, &regpowermgmnt);
	}

	return 0;
}

static void shtej_card_reset(struct shtej *wc)
{
	unsigned short val;

	wc->dsp_addr = 0x1;
	plx_setreg(wc, 0x4, wc->dsp_addr);
	
	val = 0x0000;	//reset dsp
	shtej_setdsp_data(wc, 0, 0x20, &val, sizeof(val) / sizeof(unsigned short));
	shtej_wait_ms(1000);
	val = 0x0007;	//reset dsp
	shtej_setdsp_data(wc, 0, 0x20, &val, sizeof(val) / sizeof(unsigned short));
	shtej_wait_ms(1000);
	
	val = 0x0003;	//3--select dsp data port, 0--control reg
	shtej_setdsp_data(wc, 0, 0x40, &val, sizeof(val) / sizeof(unsigned short));

	val = 0x0000;
	shtej_setdsp_data(wc, 0, 0x24, &val, sizeof(val) / sizeof(unsigned short));
	shtej_wait_ms(1000);
	val = 0x0007;
	shtej_setdsp_data(wc, 0, 0x24, &val, sizeof(val) / sizeof(unsigned short));
	shtej_wait_ms(1000);
}

static int shtej_hardware_init(struct shtej *wc, unsigned int cardflags)
{
	unsigned short val;
	
	if (shtej_plx_init(wc))
	{
		printk("shtej: plx chip type is not 9054\n");
		return -1;
	}
	
	shtej_card_reset(wc);
	
	//Disable select BF561
	val = 0xf;
	shtej_setdsp_data(wc, 0, 0x0, &val, sizeof(val) / sizeof(unsigned short));

	if(TEJMODE == 0)
	{
		wc->t1e1 = 0xf;
		wc->cardtype = TYPE_E1;
	}
	else
	{
		wc->t1e1 = 0x0;
		wc->cardtype = TYPE_T1;
	}
	
	return 0;
}

static void Write_528(struct shtej *wc, u32 nSel, unsigned short RegAddr, unsigned short Data)
{
	shtej_setdsp_data(wc, 0, DS528_Addr, &RegAddr, sizeof(RegAddr) / sizeof(unsigned short));
	
	switch (nSel)
	{
		case 0:
			// write page number
			shtej_setdsp_data(wc, 0, DS528A_Data, &Data, sizeof(Data) / sizeof(unsigned short));
			break;
		case 1:
			shtej_setdsp_data(wc, 0, DS528B_Data, &Data, sizeof(Data) / sizeof(unsigned short));
			break;
	}
}

static void Read_528(struct shtej *wc, u32 nSel, unsigned short RegAddr, unsigned short * Data)
{
	shtej_setdsp_data(wc, 0, DS528_Addr, &RegAddr, sizeof(RegAddr) / sizeof(unsigned short));
	
	switch (nSel)
	{
		case 0:
			shtej_getdsp_data(wc, 0, DS528A_Data, Data, sizeof(*Data) / sizeof(unsigned short));
			break;
		case 1:
			shtej_getdsp_data(wc, 0, DS528B_Data, Data, sizeof(*Data) / sizeof(unsigned short));
			break;
	}
}

static void shtej_WriteDS2652X(struct shtej *wc, u32 nSel, unsigned short RegAddr, unsigned short Data)
{
	unsigned short Rdata;
	
	Write_528(wc, nSel, RegAddr, Data);
	Read_528(wc, nSel, RegAddr, &Rdata);
	
	//printk("==========\t\t[%x, %x, %x]\n", RegAddr, Data, Rdata&0xff);
}

static void e1_configure(struct shtej *wc, unsigned int ndsp, u32 nSel, int lineconfig)
{
	int i;
	unsigned short index;
	
	for(i=0;i<wc->numspans;i++)
	{
		if(wc->InitCard == 0 && i != ndsp)
			continue;
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x80, 0x81);
		
		if(lineconfig & DAHDI_CONFIG_CCS)
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x81, 0x60);
		else
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x81, 0x40);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x180, 0x81);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x83, 0x0);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x84, 0x14);	//Wdata=0x74;		//H.100 backplane
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x85, 0x05);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x85, 0x01);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x88, 0x28+(i%4));
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x164, 0x9b);	//E1TAF
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x165, 0xdf);	//E1TNAF
		
		if(lineconfig & DAHDI_CONFIG_CCS)
		{
			if(lineconfig & DAHDI_CONFIG_CRC4)
				shtej_WriteDS2652X(wc, nSel, i*0x200+0x181, 0x05);
			else
				shtej_WriteDS2652X(wc, nSel, i*0x200+0x181, 0x04);
		}
		else
		{
			if(lineconfig & DAHDI_CONFIG_CRC4)
				shtej_WriteDS2652X(wc, nSel, i*0x200+0x181, 0x45);
			else
				shtej_WriteDS2652X(wc, nSel, i*0x200+0x181, 0x44);
		}

		shtej_WriteDS2652X(wc, nSel, i*0x200+0x183, 0x00);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x184, 0x14);	//Wdata=0x74;	//H.100 backplane
	
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x0140, 0x0B); // Set CAS Multiframe Sync Header
		index = 0x0141;
		while(index < 0x0150)
		{
			shtej_WriteDS2652X(wc, nSel, i*0x200+index, 0xDD);
			index++;
		}
		
		if (!(lineconfig & DAHDI_CONFIG_CCS))
		{
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x0118, 0xFE); // SSIE1, Enable Transmit Signaling for channels 1-8
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x0119, 0xFF); // SSIE2, Enable Transmit Signaling for channels 9-16
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x011A, 0xFE); // SSIE3, Enable Transmit Signaling for channels 17-24
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x011B, 0xFF); // SSIE4, Enable Transmit Signaling for channels 18-32

			shtej_WriteDS2652X(wc, nSel, i*0x200+0x01C8, 0x00);
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x01C9, 0x00);
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x01CA, 0x00);
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x01CB, 0x00);

			shtej_WriteDS2652X(wc, nSel, i*0x200+0x0C8, 0x00);
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x0C9, 0x00);
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x0CA, 0x00);
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x0CB, 0x00);
		}


		shtej_WriteDS2652X(wc, nSel, i*0x200+0x185, 0x5);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x185, 0x1);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x188, 0x28+(i%4));//TIBOC	//4device(8M),Channel Interleave ,IBOEN,1~4 device on bus

		shtej_WriteDS2652X(wc, nSel, i*0x200+0x189, 0x0f);	//TDS0SEL	// select ch15 to 0x1bb
#ifdef E75OHM
		shtej_WriteDS2652X(wc, nSel, i*0x20+0x1001, 0);	
		shtej_WriteDS2652X(wc, nSel, i*0x20+0x1007, 0);
#else
		shtej_WriteDS2652X(wc, nSel, i*0x20+0x1001, 0x31);	//LTITSR	//120ohm
		
		shtej_WriteDS2652X(wc, nSel, i*0x20+0x1007, 0x30);	//LRISMR	//120ohm
#endif		
		//shtej_WriteDS2652X(wc, nSel, i*0x20+0x1002, 0x11);	//loop back
		shtej_WriteDS2652X(wc, nSel, i*0x20+0x1002, 0x1);	//LMCR	//LMCR, Enable Transmit Outputs
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x80, 0xc1);	//RMMR
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x180, 0xc1);	//RMMR
	}
}

static void t1_configure(struct shtej *wc, unsigned int ndsp, u32 nSel, int lineconfig)
{
	int i;
	unsigned short	udata;
	unsigned short	index;

	for(i=0;i<wc->numspans;i++)
	{
		if(wc->InitCard == 0 && i != ndsp)
			continue;
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x80, 0x80);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x180, 0x80);
		
		if(lineconfig & DAHDI_CONFIG_ESF)
		{
			if(TEJMODE == 2)
				shtej_WriteDS2652X(wc, nSel, i*0x200+0x81, 0xcc);
			else
				shtej_WriteDS2652X(wc, nSel, i*0x200+0x81, 0xc8);
		}
		else
		{
			if(TEJMODE == 2)
				shtej_WriteDS2652X(wc, nSel, i*0x200+0x81, 0xec);
			else
				shtej_WriteDS2652X(wc, nSel, i*0x200+0x81, 0xe8);
		}
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x84, 0x14);	//Wdata=0x74;		//H.100 backplane
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0xc0, 0x11);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0xc1, 0x11);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0xc2, 0x11);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0xc3, 0x11);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x85, 0x1);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x85, 0x5);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x88, 0x28+(i%4));	//RIBOC	//4device(8M),Channel Interleave ,IBOEN,1~4 device on bus

		if(TEJMODE == 2)
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x181, 0x94);
		else
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x181, 0x14);	//TCR1, Set Transmit T1 B8ZS
		
		if(lineconfig & DAHDI_CONFIG_ESF)
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x183, 0x0);
		else
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x183, 0x4);

		shtej_WriteDS2652X(wc, nSel, i*0x200+0x184, 0x14);	//Wdata=0x74;		//H.100 backplane
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x1c0, 0x11);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x1c1, 0x11);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x1c2, 0x11);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x1c3, 0x11);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x185, 0x1);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x185, 0x5);
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x188, 0x28+(i%4));	//TIBOC	//4device(8M),Channel Interleave ,IBOEN,1~4 device on bus

		shtej_WriteDS2652X(wc, nSel, i*0x200+0x162, 0x1c);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x164, 0x0);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x165, 0x10);
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x166, 0x80);
		
		index = 0x0140;
		while(index < 0x014c)
		{
			shtej_WriteDS2652X(wc, nSel,  i*0x200+index, 0xDD);
			index++;
		}
		
		/*if (!(lineconfig & DAHDI_CONFIG_ESF))
		{
			shtej_WriteDS2652X(wc, nSel,  i*0x200+0x0118, 0xFF); // SSIE1, Enable Transmit Signaling for channels 1-8
			shtej_WriteDS2652X(wc, nSel,  i*0x200+0x0119, 0xFF); // SSIE2, Enable Transmit Signaling for channels 9-16
			shtej_WriteDS2652X(wc, nSel,  i*0x200+0x011A, 0xFF); // SSIE3, Enable Transmit Signaling for channels 17-24
		}*/

		shtej_WriteDS2652X(wc, nSel, i*0x200+0x83, 0x0);

		//Configure LIU
		shtej_WriteDS2652X(wc, nSel, i*0x20+0x1000, 0x02);// LTRCR, T1/J1 Mode, ANSI T1.231, and Receive JA @ 128 bits

		if(TEJMODE == 2)
		{
			shtej_WriteDS2652X(wc, nSel, i*0x20+0x1001, 0x20);
			shtej_WriteDS2652X(wc, nSel, i*0x20+0x1007, 0x22);
		}
		else
		{
			shtej_WriteDS2652X(wc, nSel, i*0x20+0x1001, 0x10);      //LTITSR, T1 Mode 100 ohm 0dB CSU w/Transmit Z-Term
			shtej_WriteDS2652X(wc, nSel, i*0x20+0x1007, 0x12);      //LRISMR, T1 Mode 100 ohm Short Haul w/Receive Z-Match
		}
		
		
		shtej_WriteDS2652X(wc, nSel, i*0x20+0x1002, 0x1);	//LMCR	//LMCR, Enable Transmit Outputs
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x189, 0x0);

		
		if(i == 0)
		{
			udata = 0x0;
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x10, udata);//set channel 1.
			
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x87, 0x02);//set watermark 32 bytes.
			
			udata |= 0x40;
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x10, udata);//set channel 1.
			
			udata &= 0xbf;
			shtej_WriteDS2652X(wc, nSel, i*0x200+0x10, udata);//set channel 1.
		}
		
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x80, 0xc0);//RMMR
		shtej_WriteDS2652X(wc, nSel, i*0x200+0x180, 0xc0);//TMMR
	}
}


static int shtej_hardware_init26524(struct shtej *wc, u32 nSel)
{
	unsigned short	Rdata, Wdata, Addr, idQ50;
	unsigned int i;

 	//Check the Device ID in the IDR register ,Test if ds25528/24 exist and the type
	Addr=0xf8;	//IDR
	Read_528(wc, nSel, Addr, &Rdata);
	idQ50 = Rdata&0xff;
	if ((idQ50>>6) != 0x01)		//26524=01011xxx,26524=01100xxx
		return DS26524_DEVICE_ERROR;
	else
		idQ50 &= 0xf8;		//26524=01011000(58),26524=01100000(60)
	
	if (idQ50 == 0x58)
		wc->MaxSpans=0x08;
	else if (idQ50 == 0x60)
		wc->MaxSpans=0x04;
	else if (idQ50 == 0x70)
		wc->MaxSpans=0x01;
	else
	{
		wc->MaxSpans=0x00;
		printk("Read Device ID in the IDR register Error\n");
		return -1;
	}
	
	printk("Ver %x.\n",Rdata&0x07);
	
	Addr = 0xf6;	//GFSRR:Framer and BERT
	Wdata = 0xff;	//reset
	Write_528(wc, nSel, Addr, Wdata);
	Read_528(wc, nSel, Addr, &Rdata);
	if ((Rdata&0xff) != Wdata)
		return DS26524_WNR_ERROR;
	
	Addr = 0xf5;	//GLSRR, Global LIU
	Wdata = 0xff;	//reset
	Write_528(wc, nSel, Addr, Wdata);
	Read_528(wc, nSel, Addr, &Rdata);
	if ((Rdata&0xff) != Wdata)
		return DS26524_WNR_ERROR;
	
	shtej_wait_ms(10000);
	shtej_wait_ms(10000);
	
	Addr = 0xf6;	//GFSRR:Framer and BERT
	Wdata = 0x00;	//work normal
	Write_528(wc, nSel, Addr, Wdata);
	Read_528(wc, nSel, Addr, &Rdata);
	if ((Rdata&0xff) != Wdata)
		return DS26524_WNR_ERROR;
	
	Addr = 0xf5;	//GLSRR, Global LIU
	Wdata = 0x00;	//work normal
	Write_528(wc, nSel, Addr, Wdata);
	Read_528(wc, nSel, Addr, &Rdata);
	if ((Rdata&0xff) != Wdata)
		return DS26524_WNR_ERROR;
	
	shtej_wait_ms(200);
	
	if(wc->numspans != 1)
	{
		//GTCCR - lf
		Addr = 0xf3;
		if(wc->cardtype == TYPE_E1)
			Wdata = 0x90;	//e1,mclk-2m(0x90)	e1,rclk1-2m(0x00)	T1,mclk-2m(0x88)
		else
			Wdata = 0x88;
			
		Write_528(wc, nSel, Addr, Wdata);
		Read_528(wc, nSel, Addr, &Rdata);
		if ((Rdata&0xff) != Wdata)
			return DS26524_WNR_ERROR;
		
		shtej_wait_ms(200);
	}
	
	//Write the  entire  remainder of the register  space for each port  with  00h,
	// including reserved register locations.
	for(i=0;i<0x2000;i++)
	{
		if((i==0xf8)||(i==0xf3))	//IDR,GTCCR
			i++;
		Addr=i;
		Wdata = 0x00;
		Write_528(wc, nSel, Addr, Wdata);
	}
	
	if(wc->numspans == 1)
	{
		Addr = 0xf3;
		if(wc->cardtype == TYPE_E1)
			Wdata = 0x90;
		else
			Wdata = 0x88;
		Write_528(wc, nSel, Addr, Wdata);
		Read_528(wc, nSel, Addr, &Rdata);
		if ((Rdata&0xff) != Wdata)
			return DS26524_WNR_ERROR;
		shtej_wait_ms(200);
	}

	//GFCR - lf
	Addr = 0xf1;
	Wdata = 0xa0;	//4dev,bpclk=8M
	Write_528(wc, nSel, Addr, Wdata);
	Read_528(wc, nSel, Addr, &Rdata);
	if ((Rdata&0xff) != Wdata)
		return DS26524_WNR_ERROR;
	
	//GFCR2 - lf
	Addr = 0xf2;
	//Wdata = 0x00;
	Wdata = 0x02;	//tssyncio as out pin
	Write_528(wc, nSel, Addr, Wdata);
	Read_528(wc, nSel, Addr, &Rdata);
	if ((Rdata&0xff) != Wdata)
		return DS26524_WNR_ERROR;
	
	if(wc->cardtype == TYPE_E1)
		e1_configure(wc, wc->MaxSpans, nSel, DAHDI_CONFIG_CCS);
	else t1_configure(wc, wc->MaxSpans, nSel, DAHDI_CONFIG_ESF);
	
	return 0;
}


static int shtej_hardware_init_2(struct shtej *wc)
{
	int i;
	unsigned short wdata;
	
	shtej_disable_pci_interrupts(wc);
	shtej_disable_local_interrupts(wc);
	
	if(shtej_hardware_init26524(wc, 0))
	{
		printk("shtej: init 26524 fail\n");
		return -1;
	}

	for(i=0; i<MAXDSPUSED; ++i)
	{
		if( ((wc->numspans == 2) && (i == 2 || i== 3)) || ((wc->numspans == 1) && (i == 1 || i == 2 || i== 3)) )
			continue;
		
		if(i != MAXECDSPUSED)
		{
			if (shtej_init_dsp(wc, "tejxapciec.bin", i+1))
			{
				printk("shtej: init EC dsp %d fail\n", i+1);
				return -1;
			}
		}
		else
		{
			if (shtej_init_dsp(wc, "tejxapci.bin", i+1))
			{
				printk("shtej: init dsp %d fail\n", i+1);
				return -1;
			}
		}
	}
	
	for(i=0; i<MAXDSPUSED; ++i)
	{
		if( ((wc->numspans == 2) && (i == 2 || i== 3)) || ((wc->numspans == 1) && (i == 1 || i == 2 || i== 3)) )
			continue;
		
		wdata = 0x3000;
		shtej_setdsp_data(wc, i+1, 0x60 * 2, &wdata, sizeof(wdata) / sizeof(unsigned short));
	}
	
	shtej_wait_ms(10000);
	printk("DSP initialize complete\n");

	wc->dsp_addr = 0x1;
	plx_setreg(wc, 0x4, wc->dsp_addr);
	return 0;
}


static int __devinit shtej_launch(struct shtej *wc)
{
	if(wc->tspans[0]->span.flags & DAHDI_FLAG_REGISTERED)
		return 0;

	printk("SHTEJ: Launching card: %d\n", wc->serialno);

	if (dahdi_register(&wc->tspans[0]->span, 0))
	{
		printk(KERN_ERR "Unable to register span %s\n", wc->tspans[0]->span.name);
		return -1;
	}

	if(wc->numspans == 2 || wc->numspans == 4)
	{
		if (dahdi_register(&wc->tspans[1]->span, 0))
		{
			printk(KERN_ERR "Unable to register span %s\n", wc->tspans[1]->span.name);
			dahdi_unregister(&wc->tspans[0]->span);
			return -1;
		}
	}

	if (wc->numspans == 4)
	{
		if (dahdi_register(&wc->tspans[2]->span, 0))
		{
			printk(KERN_ERR "Unable to register span %s\n", wc->tspans[2]->span.name);
			dahdi_unregister(&wc->tspans[0]->span);
			dahdi_unregister(&wc->tspans[1]->span);
			return -1;
		}
		if (dahdi_register(&wc->tspans[3]->span, 0))
		{
			printk(KERN_ERR "Unable to register span %s\n", wc->tspans[3]->span.name);
			dahdi_unregister(&wc->tspans[0]->span);
			dahdi_unregister(&wc->tspans[1]->span);
			dahdi_unregister(&wc->tspans[2]->span);
			return -1;
		}
		printk(KERN_ERR "register span %s\n", wc->tspans[3]->span.name);
	}
	
	wc->checktiming = 1;//add TEJ timer sync,09/03/31
	
	return 0;
}

static void *plx_dmabuf_alloc(struct shtej *wc, struct plx_phys_mem_object *memobj)
{
	dma_addr_t busaddress;
	unsigned long virt_addr;

	memobj->kernelva = dma_alloc_coherent(&(wc->dev->dev), memobj->size, &busaddress, GFP_KERNEL | __GFP_NOWARN);

	if (!memobj->kernelva)
	{
		printk("dma buf alloc error\n");
		return NULL;
	}
	memobj->busphysical = (u64)busaddress;

	for (virt_addr = (unsigned long)memobj->kernelva; virt_addr < ((unsigned long)memobj->kernelva + memobj->size); virt_addr += PAGE_SIZE)
		SetPageReserved(virt_to_page(virt_addr));

	memobj->cpuphysical = virt_to_phys(memobj->kernelva);
	if (debug)
		printk("cpu phys addr: 0x%08lx, bus phys addr: 0x%08lx, kernel VA : 0x%p\n", (unsigned long)memobj->cpuphysical, (unsigned long)memobj->busphysical, memobj->kernelva);

	if (memobj->size >= (10 << 10))
	{
		if (debug)
			printk("(%d kb)\n", memobj->size >> 10);
	}
	else
	{
		if (debug)
			printk("(%d bytes)\n", memobj->size);
	}

	return memobj->kernelva;
}

static void plx_dmabuf_free(struct shtej *wc, struct plx_phys_mem_object *memobj)
{
	unsigned long virt_addr;

	for (virt_addr = (unsigned long)memobj->kernelva; virt_addr < ((unsigned long)memobj->kernelva + PAGE_ALIGN(memobj->size)); virt_addr += PAGE_SIZE)
		ClearPageReserved(virt_to_page(virt_addr));

	dma_free_coherent(&(wc->dev->dev), memobj->size, memobj->kernelva, (int)memobj->busphysical);
	if (debug)
		printk("released physical memory at 0x%08lx\n", (unsigned long) memobj->cpuphysical);

	if (memobj->size >= (10 << 10))
	{
		if (debug)
			printk("(%d kb)\n", memobj->size >> 10);
	}
	else
	{
		if (debug)
			printk("(%d bytes)\n", memobj->size);
	}

	memset(memobj, 0x0, sizeof(struct plx_phys_mem_object/*unsigned long*/));

}

static int plx_dmasgl_transfer_listadd(struct shtej *wc, DMA_CHANNEL channel, struct dma_transfer_element *dmadata)
{
	u8 i;
	u8 shift;
	u32 regvalue;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			shift = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			shift = 8;
			break;
		default:
			return -1;

	}

	regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
	if ((regvalue & ((1 << 4) << shift)) == 0)
	{
		if(debug > 66)
			printk("listadd: dma channel is carrently active\n");

		return -1;
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
	{
		if(debug > 66)
			printk("listadd: dma channel has not been opened for SGL DMA\n");

		return -1;
	}

	wc->dmainfo[i].dmadata[wc->dmainfo[i].dmadatacount] = *dmadata;

	if(wc->dmainfo[i].dmadatacount < PLAYBLOCK_NUM)
		wc->dmainfo[i].dmadatacount++;
	else
	{
		if(debug > 66)
			printk("DMA LIST IS FULL\n");
	}

	return 0;
}

static int plx_dmasgl_lockbuffer_and_build(struct shtej *wc, u8 dmaindex)
{
	u32 k = 0;
	u32 bussgl;
	u32 bussgloriginal;
	u32 sglsize;
	u32 totalpages;

	unsigned long vasgl;
	struct dma_transfer_element *dma;

	totalpages = wc->dmainfo[dmaindex].dmadatacount;

	sglsize = (totalpages * (4 * sizeof(u32))) + 16;

	if (wc->dmainfo[dmaindex].sglbuffer.kernelva)
	{
		if (wc->dmainfo[dmaindex].sglbuffer.size < sglsize)
		{
			printk("err:sglbuffer.size < sglsize return line %d\n",__LINE__);
			return 0;
		}
	}

	if (!wc->dmainfo[dmaindex].sglbuffer.kernelva)
	{
		printk("lockbuf: unable to allocate %d bytes for %d SGL descriptors\n", wc->dmainfo[dmaindex].sglbuffer.size, totalpages);
		return 0;
	}
	else
		vasgl = (unsigned long)wc->dmainfo[dmaindex].sglbuffer.kernelva;



	bussgl = (u32)wc->dmainfo[dmaindex].sglbuffer.busphysical;

	vasgl = (vasgl + 0xf) & ~(0xf);

	bussgl = (bussgl + 0xf) & ~(0xf);

	bussgloriginal = bussgl;


	for (k = 0; k < wc->dmainfo[dmaindex].dmadatacount; k++)
	{
		dma = &wc->dmainfo[dmaindex].dmadata[k];

		totalpages--;
		*(((u32*)vasgl) + 0) = dma->u.pciaddrlow;
		*(((u32*)vasgl) + 1) = dma->localaddr;
		*(((u32*)vasgl) + 2) = dma->transfercount;
		if (totalpages == 0)
		{
			*(((u32*)vasgl) + 3) = (dma->localtopcidma << 3) | (1 << 1) | (1 << 0);
		}
		else
		{
			bussgl += (4 * sizeof(u32));
			*(((u32*)vasgl) + 3) = bussgl | (dma->localtopcidma << 3) | (1 << 0);
			vasgl +=(4 * sizeof(u32));
		}

	}
	return bussgloriginal;
}

static int plx_dmasgl_transfer(struct shtej *wc, DMA_CHANNEL channel)
{
	u8 i;
	u8 shift;
	u16 offsetdesc;
	u32 regvalue;
	u32 sglpciaddress;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			shift = 0;
			offsetdesc = PCI9054_DMA0_DESC_PTR;
			break;
		case DMA_CHANNEL1:
			i = 1;
			shift = 8;
			offsetdesc = PCI9054_DMA1_DESC_PTR;
			break;
		default:
			return -1;

	}

	regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
	if ((regvalue & ((1 << 4) << shift)) == 0)
		return -1;

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
		return -1;

	sglpciaddress = plx_dmasgl_lockbuffer_and_build(wc, i);
	
	if (sglpciaddress == 0)
	{
		if(debug > 66)
			printk("unable to lock buffer and build SGL list\n");
		return -1;
	}

	plx_setreg(wc, offsetdesc, sglpciaddress | (1 << 0));
	regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
	plx_setreg(wc, PCI9054_DMA_COMMAND_STAT ,regvalue | ((1 << 0) << shift));	//enable channel to transfer data
	plx_setreg(wc, PCI9054_DMA_COMMAND_STAT, regvalue | (((1 << 0) | (1 << 1)) << shift)); // start dma

	wc->dmainfo[i].timeout = 0;
	return 0;
}

static int plx_checkdmaready(struct shtej *wc, DMA_CHANNEL channel)
{
	u8 i;
	u8 shift;
	u32 regvalue;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			shift = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			shift = 8;
			break;
		default:
			return -1;

	}

	regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
	if ((regvalue & ((1 << 4) << shift)) == 0)
	{
		if(debug > 66)
			printk("dmatransfer: dma channel is carrently active\n");

		return -1;
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
	{
		if(debug > 66)
			printk("dmatransfer: dma channel has not been opened for SGL DMA\n");

		return -1;
	}
	
	wc->dmainfo[i].dmadatacount = 0;
	
	return 0;
}

static inline void shtej_receiveprep(struct shtej *wc)
{
	int x, y, i, m, n, spanmaxch;
	struct dma_transfer_element el;
	unsigned short val;
	
	if(wc->cardtype == TYPE_E1)
		spanmaxch = 31;
	else
		spanmaxch = 24;
	
	if(wc->readhasdata[wc->readcurpos] == 1)
	{
		m = 0;
		n = 0;
		for (x = 0; x < spanmaxch; x++)
		{
			for (i = 0; i < wc->numspans; i++)
			{
				if (wc->tspans[i]->span.flags & DAHDI_FLAG_RUNNING)
				{
					for (y = 0; y < DAHDI_CHUNKSIZE; y++)
					{
						if(wc->cardtype == TYPE_E1)
							wc->tspans[i]->span.chans[x]->readchunk[y] = wc->readchunk[wc->readcurpos][4*x+4+i][y];
						else
							wc->tspans[i]->span.chans[x]->readchunk[y] = wc->readchunk[wc->readcurpos][4*m+4+i][y];
					}
					
					if(debug > 55)
					{
						if(i == 0 && (wc->filecount+DAHDI_CHUNKSIZE)<=0x20000)
						{
							for (y = 0; y < DAHDI_CHUNKSIZE; y++)
									wc->tmpwriterec[wc->Index][x][wc->filecount+y] = wc->readchunk[wc->readcurpos][4*x+4+i][y];
						}
					}

				}
			}
			if(wc->cardtype != TYPE_E1)
			{
				m++;
				if(T1S[n] == m)
				{
					m++;
					n++;
				}
			}
		}
		
#ifdef WORKQUEUE
		queue_work(wc->rec_span_wq, &wc->rec_span_work);
#else
		for (i = 0; i < wc->numspans; i++)
		{//耗时500ms
			if (wc->tspans[i]->span.flags & DAHDI_FLAG_RUNNING)
				dahdi_receive(&wc->tspans[i]->span);
		}
#endif
		
		if(debug > 66)
		{
			wc->tflags = (wc->tflags+1)&0xfff;
			if(wc->tflags != wc->readchunk_frame[wc->readcurpos]->readflag)
			{
				printk("R=====          [%x, %x]\n", wc->tflags, wc->readchunk_frame[wc->readcurpos]->readflag);
				wc->tflags = wc->readchunk_frame[wc->readcurpos]->readflag;
			}
		}
		wc->readhasdata[wc->readcurpos] = 0;
		wc->readcurpos = (wc->readcurpos+1)%RECBLOCK_BUF;
	}
	
	shtej_getdsp_data(wc, VOCDSP, RECBUF_FLAG * 2, &val, sizeof(val) / sizeof(unsigned short));
	
	if(wc->RECFlag_First == 1)
	{
		wc->RECFlag_First = 0;
		wc->RECFlag = val;
		wc->readfillpos = wc->RECFlag;
		wc->rpos = wc->readfillpos;
	}
	
	if(plx_checkdmaready(wc, DMA_CHANNEL1) == 0)
	{
		for(i = wc->readfillpos; i != wc->rpos;)
		{
			if(wc->readhasdata[i] == 0)
			{
				wc->readhasdata[wc->readfillpos] = 1;
				wc->readfillpos = (wc->readfillpos+1)%RECBLOCK_BUF;
			}
			else
				break;
				
			i = (i+1)%RECBLOCK_BUF;
		}
		
		if(((wc->readfillpos - wc->readcurpos + RECBLOCK_BUF)%RECBLOCK_BUF) > RECLIMIT_BUF)
		{
			wc->readcurpos	= 0;
			wc->readfillpos	= 0;
			wc->rpos				= 0;

			for(x = 0; x < RECBLOCK_BUF; x++)
				wc->readhasdata[x] = 0;
		}
		
		wc->rpos = wc->readfillpos;
		for(i = wc->RECFlag; i != val;)
		{
			el.localtopcidma = 1;
			el.localaddr = 0x10000*VOCDSP + (RECBLOCK_LEN * i + BASEADDR_RECBUF) * 2;
			el.u.pciaddrlow = wc->dmainfo[1].dspbuffer.busphysical+(NUM_CARDS * DAHDI_CHUNKSIZE + sizeof(unsigned short))*wc->rpos;
			el.transfercount = wc->dmainfo[1].dspbuffer.size/RECBLOCK_BUF;
			plx_dmasgl_transfer_listadd(wc, DMA_CHANNEL1, &el);
			
			i = (i+1)%RECBLOCK_NUM;
			wc->rpos = (wc->rpos+1)%RECBLOCK_BUF;
		}

		wc->RECFlag = val;
	}
}

static inline void shtej_transmitprep(struct shtej *wc)
{
	int x, y, i, m, n, spanmaxch;
	unsigned short val;
	struct dma_transfer_element el;
	
	if(wc->cardtype == TYPE_E1)
		spanmaxch = 31;
	else
		spanmaxch = 24;
	
	if(((wc->writefillpos - wc->writecurpos + PLAYBLOCK_BUF)%PLAYBLOCK_BUF) > PLAYLIMIT_BUF)
	{
		wc->writecurpos	= 0;
		wc->writefillpos	= 0;
		
		for(x = 0; x < PLAYBLOCK_BUF; x++)
			wc->writehasdata[x] = 0;
	}
	
	if(wc->writehasdata[wc->writefillpos] == 0)
	{
		for (i = 0; i < wc->numspans; i++)
		{
			if (wc->tspans[i]->span.flags & DAHDI_FLAG_RUNNING)
				dahdi_transmit(&wc->tspans[i]->span);
		}
		
		m = 0;
		n = 0;
		for (x = 0; x < spanmaxch; x++)
		{
			for (i = 0; i < wc->numspans; i++)
			{
				if (wc->tspans[i]->span.flags & DAHDI_FLAG_RUNNING)
				{
					for (y = 0; y < DAHDI_CHUNKSIZE; y++)
					{
						if(wc->cardtype == TYPE_E1)
							wc->writechunk[wc->writefillpos][x*4+4+i][y] = wc->tspans[i]->span.chans[x]->writechunk[y];
						else
							wc->writechunk[wc->writefillpos][m*4+4+i][y] = wc->tspans[i]->span.chans[x]->writechunk[y];
					}
					
					if(debug > 55)
					{
						if(i == 0 && (wc->filecount+DAHDI_CHUNKSIZE)<=0x20000)
						{
							for (y = 0; y < DAHDI_CHUNKSIZE; y++)
								wc->tmpwrite[wc->Index][x][wc->filecount+y] = wc->tspans[i]->span.chans[x]->writechunk[y];
						}
					}
					
				}
			}
					
			if(wc->cardtype != TYPE_E1)
			{
				m++;
				if(T1S[n] == m)
				{
					m++;
					n++;
				}
			}
		}
	
		wc->writehasdata[wc->writefillpos] = 1;
		wc->writefillpos = (wc->writefillpos+1)%PLAYBLOCK_BUF;
	}

	shtej_getdsp_data(wc, VOCDSP, PLAYBUF_FLAG * 2, &val, sizeof(val) / sizeof(unsigned short));

	if(wc->TRIFlag_First == 1)
	{
		wc->TRIFlag_First = 0;
		wc->TRIFlag = (val-1+PLAYBLOCK_NUM)%PLAYBLOCK_NUM;
	}
	
	if(plx_checkdmaready(wc, DMA_CHANNEL0) == 0)
	{
		for(i = wc->writecurpos; i != wc->writefillpos;)
		{
			if(wc->writehasdata[i] == 1)
			{
				if(val == wc->TRIFlag)
					break;
				
				wc->writehasdata[i] = 0;

				el.localtopcidma = 0;
				el.localaddr = 0x10000*VOCDSP + (PLAYBLOCK_LEN * wc->TRIFlag + BASEADDR_PLAYBUF) * 2;
				el.u.pciaddrlow = wc->dmainfo[0].dspbuffer.busphysical+(NUM_CARDS*DAHDI_CHUNKSIZE+sizeof(unsigned short))*i;
				el.transfercount = wc->dmainfo[0].dspbuffer.size/PLAYBLOCK_BUF;
				plx_dmasgl_transfer_listadd(wc, DMA_CHANNEL0, &el);
				
				wc->TRIFlag = (wc->TRIFlag+1)%PLAYBLOCK_NUM;
			}
			
			i = (i+1)%PLAYBLOCK_BUF;
		}
		wc->writecurpos = i;
		//wc->writecurpos = wc->writefillpos;
	}
}

static void __shtej_set_timing_source(struct shtej *wc, int unit, int master, int slave)
{
	unsigned short Wdata;
	int x;
	
	if (unit != wc->syncsrc) {
		if ((unit > -1) && (unit < 4)) {	
		 //set unit is timer source
				if(wc->cardtype == TYPE_E1)
					Wdata = 0x00+unit*0x10;
				else
					Wdata = 0x08+unit*0x10;
			Write_528(wc, 0, 0xf3, Wdata);
		} else {
		 //set unit timer source from board
		 		if(wc->cardtype == TYPE_E1)
					Wdata = 0x90;
		 		else
					Wdata = 0x88;
			Write_528(wc, 0, 0xf3, Wdata);
		}
		if (!master && !slave)
			wc->syncsrc = unit;
		if ((unit < 0) || (unit > 3))
			unit = 0;
		else
			unit++;
		if (!master && !slave) {
			for (x=0;x<wc->numspans;x++)
				wc->tspans[x]->span.syncsrc = unit;
		}
	} 
}


static void __shtej_set_timing_source_auto(struct shtej *wc)
{
	int x;
	
	wc->checktiming = 0;
	
	for (x=0;x<wc->numspans;x++) {
			if (wc->tspans[x]->sync) {
				if (/*(wc->tspans[wc->tspans[x]->psync - 1]->span.flags & DAHDI_FLAG_RUNNING) &&*/ 
					!(wc->tspans[wc->tspans[x]->psync - 1]->span.alarms & (DAHDI_ALARM_RED | DAHDI_ALARM_BLUE) )) {
						/* Valid timing source */
						__shtej_set_timing_source(wc, wc->tspans[x]->psync - 1, 0, 0);
						return;
				}
			}
		}
	__shtej_set_timing_source(wc, 4, 0, 0);
}

static void do_interrupts(void *p)
{
	u32 regvalue;
	unsigned long intsrc;
	struct shtej *wc = (struct shtej *)p;
	int i, j, k;
	unsigned short rxc, c, n;

	intsrc = wc->interruptsrc;

	if (intsrc & INTR_TYPE_LOCAL_1)
	{
#if 1
		i = wc->intcount & 15;

		if ((i < 3) && (wc->cardtype == TYPE_E1))
		{
			for (j = (i * 5); j < (i * 5) + 5; j++)
			{
				for (k = 0; k < wc->numspans; k++)
				{
					Read_528(wc, 0, k*0x200+0x41+j, &c);
					rxc = c & 0xf;
					if (rxc != wc->tspans[k]->span.chans[j + 16]->rxsig)
					{
						if (!(wc->tspans[k]->span.chans[j + 16]->sig & DAHDI_SIG_CLEAR))
							dahdi_rbsbits(wc->tspans[k]->span.chans[j + 16], rxc);
					}
					rxc = (c >> 4) & 0xf;
				if (rxc != wc->tspans[k]->span.chans[j]->rxsig)
					{
						if (!(wc->tspans[k]->span.chans[j]->sig & DAHDI_SIG_CLEAR))
							dahdi_rbsbits(wc->tspans[k]->span.chans[j], rxc);
					}
				}
			}
		}

		if ((i < 3) && (wc->cardtype == TYPE_T1 || wc->cardtype == TYPE_J1))
		{
			for (j = (i * 4); j < (i * 4) + 4; j++)
			{
				for (k = 0; k < wc->numspans; k++)
				{
					Read_528(wc, 0, k*0x200+0x40+j, &c);
					
					if(wc->tspans[k]->span.lineconfig & DAHDI_CONFIG_D4)
						c = c & 0xcc;
					
					rxc = c & 0xf;
					if (rxc != wc->tspans[k]->span.chans[j + 12]->rxsig)
					{
						if (!(wc->tspans[k]->span.chans[j + 12]->sig & DAHDI_SIG_CLEAR))
							dahdi_rbsbits(wc->tspans[k]->span.chans[j + 12], rxc);
					}
					rxc = (c >> 4) & 0xf;
				if (rxc != wc->tspans[k]->span.chans[j]->rxsig)
					{
						if (!(wc->tspans[k]->span.chans[j]->sig & DAHDI_SIG_CLEAR))
							dahdi_rbsbits(wc->tspans[k]->span.chans[j], rxc);
					}
				}
			}
		}
		
		for (i = 0; i < wc->numspans; i++)
		{
			if (wc->tspans[i]->alarmtimer)
			{
				if (!--wc->tspans[i]->alarmtimer)
				{
					wc->tspans[i]->span.alarms &= ~DAHDI_ALARM_RECOVER;
					if (wc->cardtype == TYPE_E1)
						shtej_WriteDS2652X(wc, 0, i*0x200+0x165, 0x5f);
					else
						shtej_WriteDS2652X(wc, 0, i*0x200+0x181, 0x10);
					dahdi_alarm_notify(&wc->tspans[i]->span);
				}
			}
		}
		
		i = wc->intcount & 15;
		if ( ( ((wc->numspans==4)&&(i >= 10)&&(i <= 13)) || ((wc->numspans==2)&&(i >= 10)&&(i <= 11)) || ((wc->numspans==1)&&(i == 10)) ) && !(wc->intcount & 0x30))
		{
			j = 0;
			i -= 10;

			Read_528(wc, 0, i*0x200+0xb0, &c);
			c &= 0xf;
			if ((c & 3) && !(wc->tspans[i]->alarm&DAHDI_ALARM_RED)) // if red alarm
			{
				j |= DAHDI_ALARM_RED;
				wc->tspans[i]->alarm = DAHDI_ALARM_RED;
			}
			if ((c == 0) && !(wc->tspans[i]->alarm&DAHDI_ALARM_BLUE)) // if blue alarm
			{
				j |= DAHDI_ALARM_BLUE;
				wc->tspans[i]->alarm = DAHDI_ALARM_BLUE;
			}

			wc->tspans[i]->span.alarms &= (DAHDI_ALARM_RED | DAHDI_ALARM_BLUE | DAHDI_ALARM_NOTOPEN);
			
			n = 1;
			if (wc->tspans[i]->span.lineconfig & DAHDI_CONFIG_NOTOPEN)
			{
				for (n = 0,k = 0; k < wc->tspans[i]->span.channels; k++)
				{
					if ((wc->tspans[i]->span.chans[k]->flags & DAHDI_FLAG_OPEN) ||
					(wc->tspans[i]->span.chans[k]->flags & DAHDI_FLAG_NETDEV)) n++;
				}

				if (!n) j |= DAHDI_ALARM_NOTOPEN;
			}


			//add TEJ timer sync,09/03/31
			//if (((!wc->tspans[i]->span.alarms) && (j)) || (wc->tspans[i]->span.alarms && (!j)))
			if(j) 
				wc->checktiming = 1;
			//end
			if ((!j) && wc->tspans[i]->span.alarms)
			{
				wc->tspans[i]->alarmtimer = DAHDI_ALARMSETTLE_TIME;
			}
			
			if (wc->tspans[i]->alarmtimer) j |= DAHDI_ALARM_RECOVER;

			if ((j) && (!wc->tspans[i]->span.alarms))
			{
				if (wc->cardtype == TYPE_E1)
					shtej_WriteDS2652X(wc, 0, i*0x200+0x165, 0x7f);
				else
					shtej_WriteDS2652X(wc, 0, i*0x200+0x181, 0x10);
			}
			
			if (wc->tspans[i]->span.maintstat || wc->tspans[i]->span.mainttimer) j |= DAHDI_ALARM_LOOPBACK;
			
			wc->tspans[i]->span.alarms = j;
			
			dahdi_alarm_notify(&wc->tspans[i]->span);
		}

#endif

		if(debug > 50)
			DebugInt(wc);

		if(debug > 55)
		{
			if(wc->complete == 1 && (wc->filecount+DAHDI_CHUNKSIZE) > 0x20000)
			{
				wc->Index = (wc->Index+1)%2;
				queue_work(wc->shtej_wq, &wc->shtej_work);
				wc->complete = 0;
				wc->filecount = 0;
			}
		}

		shtej_receiveprep(wc);
		shtej_transmitprep(wc);
		
		
		//add TEJ timer sync,09/03/30pm
		/*xnx_do_counters(wc);
		vcount = wc->intcount&15;
	  switch(x){
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  		xnx_check_sigbits(wc,vcount);
	  		break;
	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  		xnx_check_alarms(wc,vcount-4);
	  		break;
	 }	*/
	 if(wc->checktiming > 0)
	    __shtej_set_timing_source_auto(wc);
		//end
		
		if(debug > 55)
		{
			if((wc->filecount+DAHDI_CHUNKSIZE)<=0x20000)
				wc->filecount += DAHDI_CHUNKSIZE;
		}

		if (wc->dmainfo[0].dmadatacount)
		{
			plx_dmasgl_transfer(wc, DMA_CHANNEL0);
		}
		if (wc->dmainfo[1].dmadatacount)
		{
			plx_dmasgl_transfer(wc, DMA_CHANNEL1);
		}
		
		for (i = 0; i < wc->numspans; i++)
		{
			shtej_doecho(wc->tspans[i]);
			shtej_dorbsbits(wc->tspans[i]);
		}

		shtej_enable_cpld_interrupts(wc);
		shtej_enable_local_interrupts(wc);
		
		if((oldtiming & 1<<wc->num) != (timing & 1<<wc->num))
		{
			unsigned short	Rdata, Wdata;

			if(oldtiming & 1<<wc->num)
			{
				if(wc->cardtype == TYPE_E1)
					Wdata = 0x90;
				else
					Wdata = 0x88;
			}
			else
			{
				if(wc->cardtype == TYPE_E1)
					Wdata = 0x00;
				else
					Wdata = 0x08;
			}
			Write_528(wc, 0, 0xf3, Wdata);
			Read_528(wc, 0, 0xf3, &Rdata);
			oldtiming ^= 1<<wc->num;
		}
	}
	
	if (intsrc & INTR_TYPE_DOORBELL)
	{
		regvalue = plx_getreg(wc, 0x64);
		plx_setreg(wc, 0x64, regvalue);
	}

	if (intsrc & INTR_TYPE_PCI_ABORT)
	{
		pci_getconfig_dword(wc, 0x4, &regvalue);
		pci_setconfig_dword(wc, 0x4, regvalue);
	}

	if (intsrc & INTR_TYPE_OUTBOUND_POST)
		plx_setreg(wc, PCI9054_OUTPOST_INT_MASK, (1 << 3));
	
	shtej_enable_pci_interrupts(wc);
}


/*static void xnx_check_sigbits(struct shtej *wc, int span)
{
	
}
static void xnx_check_alarms(struct shtej *wc,int span)
{
	
}*/



DAHDI_IRQ_HANDLER(shtej_interrupt)
{
	unsigned long interruptreg;
	unsigned long interruptsrc;
	unsigned long regvalue;
	struct shtej *wc = dev_id;
	
	interruptreg = plx_getreg(wc, PCI9054_INT_CTRL_STAT);

	if (interruptreg == 0xffffffff)
		return IRQ_NONE;

	if ((interruptreg & (1 << 8)) == 0)
		return IRQ_NONE;

	interruptsrc = INTR_TYPE_NONE;

	if ((interruptreg & (1 << 13)) && (interruptreg & (1 << 9)))
		interruptsrc |= INTR_TYPE_DOORBELL;

	if ((interruptreg & (1 << 14)) && (interruptreg & (1 << 10)))
		interruptsrc |= INTR_TYPE_PCI_ABORT;

	if ((interruptreg & (1 << 15)) && (interruptreg & (1 << 11)))
	{
		interruptsrc |= INTR_TYPE_LOCAL_1;
		shtej_disable_cpld_interrupts(wc);
		shtej_disable_local_interrupts(wc);
		wc->intcount++;
	}

	if ((interruptreg & (1 << 21)) && (interruptreg & (1 << 18)))
	{
		regvalue = plx_getreg(wc, PCI9054_DMA0_MODE);
		if (regvalue & (1 << 17)) 
			interruptsrc |= INTR_TYPE_DMA_0;
	}

	if ((interruptreg & (1 << 22)) && (interruptreg & (1 << 19)))
	{
		regvalue = plx_getreg(wc, PCI9054_DMA1_MODE);
		if (regvalue & (1 << 17)) 
			interruptsrc |= INTR_TYPE_DMA_1;
	}

	regvalue = plx_getreg(wc, PCI9054_OUTPOST_INT_STAT);
	if (regvalue & (1 << 3))
	{
		regvalue = plx_getreg(wc, PCI9054_OUTPOST_INT_MASK);
		if ((regvalue & (1 << 3)) == 0)
			interruptsrc |= INTR_TYPE_OUTBOUND_POST;
	}


	if (interruptsrc == INTR_TYPE_NONE)
		return IRQ_NONE;

	wc->interruptsrc = interruptsrc;

	shtej_disable_pci_interrupts(wc);

	if (wc->irq_enable)
		do_interrupts((void *)wc);

	return IRQ_RETVAL(1);

}


static int plx_dmasgl_channel_open(struct shtej *wc, DMA_CHANNEL channel, struct dma_channel_desc *desc, void *owner)
{
	u8 i;
	u32 mode;
	u32 regvalue;
	u32 threshold;
	struct plx_reg_data regdata;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			break;
		default:
			printk("open:Invalid DMA channel\n");
			return -1;
	}

	if (wc->dmainfo[i].state != DMA_STATE_CLOSED)
	{
		printk("DMA channel already open\n");
		return -1;
	}

	wc->dmainfo[i].state = DMA_STATE_SGL;
	wc->dmainfo[i].owner = owner;

	regdata.wc = wc;
	regdata.offset = PCI9054_INT_CTRL_STAT;
	regdata.bitstoclear = 0;
	regdata.bitstoset = 0;

	regvalue = plx_getreg(wc, PCI9054_LOCAL_DMA_ARBIT);
	regvalue &= ~((1 << 20) | (1 << 19));

	switch (desc->dmachannelpriority)
	{
		case CHANNEL_0_HIGHEST:
			plx_setreg(wc, PCI9054_LOCAL_DMA_ARBIT, regvalue | (1 << 19));
			break;
		case CHANNEL_1_HIGHEST:
			plx_setreg(wc, PCI9054_LOCAL_DMA_ARBIT, regvalue | (1 << 20));
			break;
		case ROTATIONAL:
			plx_setreg(wc, PCI9054_LOCAL_DMA_ARBIT, regvalue);
			break;
		default:
			printk("dmachannelopen() invalid priority state\n");
			return -1;
	}

	threshold = 
			(desc->tholdforiopwrites << 0) |
			(desc->tholdforiopreads << 4) |
			(desc->tholdforpciwrites << 8) |
			(desc->tholdforpcireads << 12);
	mode =
			(1 << 9) |
			(1 << 10) |
			(1 << 17) |
			(0 << 18) |
			(desc->iopbuswidth << 0) |
			(desc->waitstates << 2) |
			(desc->enablereadyinput << 6) |
			(desc->enablebterminput << 7) |
			(desc->enableiopburst << 8) |
			(desc->holdiopaddrconst << 11) |
			(desc->demandmode << 12) |
			(desc->enablewriteinvalidmode << 13) |
			(desc->enabledmaeoppin << 14) |
			(desc->dmastoptransfermode << 15);
	if (desc->holdiopaddrconst)
		wc->dmainfo[i].localaddrconstant = 1;
	else 
		wc->dmainfo[i].localaddrconstant = 0;

	regvalue = plx_getreg(wc, PCI9054_DMA_THRESHOLD);
	if (channel == DMA_CHANNEL0)
	{
		plx_setreg(wc, PCI9054_DMA_THRESHOLD, (regvalue & 0xffff0000) | threshold);
		plx_setreg(wc, PCI9054_DMA0_MODE, mode);
		regdata.bitstoclear = (1 << 18);
		plx_sync_reg_modify(&regdata);
		plx_setreg(wc, PCI9054_DMA0_PCI_DAC, 0);
	}
	else
	{
		plx_setreg(wc, PCI9054_DMA_THRESHOLD, (regvalue & 0xffff0000) | (threshold << 16));
		plx_setreg(wc, PCI9054_DMA1_MODE, mode);
		regdata.bitstoclear = (1 << 19);
		plx_sync_reg_modify(&regdata);
		plx_setreg(wc, PCI9054_DMA1_PCI_DAC, 0);
	}

	return 0;
}

static int plx_dmasgl_channel_close(struct shtej *wc, DMA_CHANNEL channel, unsigned int checkinprogress)
{
	u8 i;
	u32 regvalue;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			break;
		default:
			printk("close:Invalid DMA channel\n");
			return -1;
	}

	if (checkinprogress)
	{
		regvalue = plx_getreg(wc, PCI9054_DMA_COMMAND_STAT);
		if (channel == DMA_CHANNEL1)
			regvalue = regvalue >> 8;

		if ((regvalue & (1 << 4)) == 0)
		{
			if (regvalue & (1 << 0))
			{
				printk("dma in progress\n");
				return -1;
			}
			else
			{
				printk("dma paused\n");
				return -1;
			}
		}
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
	{
		printk("--DMA channel has not been opened for SGL DMA\n");
		return -1;
	}

	wc->dmainfo[i].state = DMA_STATE_CLOSED;
	wc->dmainfo[i].owner = NULL;

	if (wc->dmainfo[i].sglbuffer.kernelva)
	{
		if (debug)
			printk("channel close: releasing memory used for SGL descript...\n");
		plx_dmabuf_free(wc, &wc->dmainfo[i].sglbuffer);
	}
	if (wc->dmainfo[i].dspbuffer.kernelva)
	{
		if (debug)
			printk("channel close: releasing memory used for DSP descript...\n");
		plx_dmabuf_free(wc, &wc->dmainfo[i].dspbuffer);
	}
	return 0;

}

static int shtej_stop_dma(struct shtej *wc)
{
	plx_dmasgl_channel_close(wc, DMA_CHANNEL0, 1);
	plx_dmasgl_channel_close(wc, DMA_CHANNEL1, 1);
	return 0;
}

static int shtej_init_dma(struct shtej *wc)
{
	int i;
	struct dma_channel_desc desc;

	for (i = 0; i < 2; i++)
	{
		wc->dmainfo[i].sglbuffer.size = 4096;
		plx_dmabuf_alloc(wc, &wc->dmainfo[i].sglbuffer);
	}

	memset(&desc, 0x0, sizeof(struct dma_channel_desc));
	desc.iopbuswidth = 1;
	desc.waitstates = 0;
	desc.enablereadyinput = 1;
	desc.enablebterminput = 0;
	desc.enableiopburst = 0;
	desc.holdiopaddrconst = 0;
	desc.demandmode = 0;
	desc.enablewriteinvalidmode = 0;
	desc.enabledmaeoppin = 0;
	desc.dmastoptransfermode = 1;
	desc.tholdforpciwrites = 15;
	desc.tholdforpcireads = 15;
	if (plx_dmasgl_channel_open(wc,DMA_CHANNEL0, &desc, wc))
		return -1;

	if (plx_dmasgl_channel_open(wc,DMA_CHANNEL1, &desc, wc))
		return -1;

	return 0;
}

static void shtej_init_readandwrite_chunk(struct shtej *wc)
{
	int i, j;
	
	wc->dmainfo[0].dspbuffer.size = (NUM_CARDS * DAHDI_CHUNKSIZE + sizeof(unsigned short))*PLAYBLOCK_BUF;
	plx_dmabuf_alloc(wc, &wc->dmainfo[0].dspbuffer);
	wc->dmainfo[1].dspbuffer.size = (NUM_CARDS * DAHDI_CHUNKSIZE + sizeof(unsigned short))*RECBLOCK_BUF;
	plx_dmabuf_alloc(wc, &wc->dmainfo[1].dspbuffer);

	for (j = 0; j<PLAYBLOCK_BUF; j++)
	{
		wc->writechunk_frame[j] = (struct dsp_writechunk_frame *)(wc->dmainfo[0].dspbuffer.kernelva+(NUM_CARDS * DAHDI_CHUNKSIZE + sizeof(unsigned short))*j);
		wc->writechunk_frame[j]->writeflag = 1;
		
		for (i = 0; i < NUM_CARDS; i++)
			wc->writechunk[j][i] = wc->writechunk_frame[j]->writechunkbuf + DAHDI_CHUNKSIZE * i;
	}
	
	for (j = 0; j<RECBLOCK_BUF; j++)
	{
		wc->readchunk_frame[j] = (struct dsp_readchunk_frame *)(wc->dmainfo[1].dspbuffer.kernelva+(NUM_CARDS * DAHDI_CHUNKSIZE + sizeof(unsigned short))*j);	
		wc->readchunk_frame[j]->readflag = 1;
		
		for (i = 0; i < NUM_CARDS; i++)
			wc->readchunk[j][i] = wc->readchunk_frame[j]->readchunkbuf + DAHDI_CHUNKSIZE * i;
	}
}


static int __devinit shtej_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int res;
	int x, y,xx;
	char qname[20];
	struct shtej *wc;
	struct shtej_desc *d = (struct shtej_desc *)ent->driver_data;

	if (pci_enable_device(pdev))
		res = -EIO;
	else
	{
		wc = kmalloc(sizeof(struct shtej), GFP_KERNEL);
		if (!wc)
			return -ENOMEM;
		memset(wc, 0x0, sizeof(struct shtej));
		spin_lock_init(&wc->lock);
		
#ifdef T1MODE
		TEJMODE = 1;
#endif
#ifdef J1MODE
		TEJMODE = 2;
#endif

		if(ent->device == 0x00fa)
			wc->numspans = 4;
		else if(ent->device == 0x00fb)
			wc->numspans = 2;
		else
			wc->numspans = 1;

		wc->variety = d->name;
		wc->dev = pdev;
		wc->irq_enable = 0;
		wc->InitCard = 1;
		
		wc->RECFlag = 0;
		wc->TRIFlag = 0;
		wc->RECFlag_First = 1;
		wc->TRIFlag_First = 1;

		if (pdev->irq < 1)
		{
			if (wc)
				kfree(wc);
			printk(KERN_ERR "No IRQ allocated for device\n");
			return -ENODEV;

		}
		
		wc->readcurpos	= 0;
		wc->readfillpos	= 0;
		wc->rpos				= 0;
		wc->writecurpos	= 0;
		wc->writefillpos	= 0;
		
		for(x = 0; x < PLAYBLOCK_BUF; x++)
			wc->writehasdata[x] = 0;

		for(x = 0; x < RECBLOCK_BUF; x++)
			wc->readhasdata[x] = 0;
		
		wc->plx_region = pci_resource_start(pdev,0);
		wc->plx_len = pci_resource_len(pdev,0);
		if (check_mem_region(wc->plx_region, wc->plx_len))
		{
			printk("plx region %lx-%lx already in use\n", wc->plx_region, wc->plx_region + wc->plx_len);
			if (wc) 
				kfree(wc);
			return -EIO;
		}

		wc->dsp_region = pci_resource_start(pdev,2);
		wc->dsp_len = pci_resource_len(pdev,2);
		if (check_mem_region(wc->dsp_region, wc->dsp_len))
		{
			printk("dsp region %lx-%lx already in use\n", wc->dsp_region, wc->dsp_region + wc->dsp_len);
			if (wc) 
				kfree(wc);
			return -EIO;
		}

		if (request_mem_region(wc->plx_region, wc->plx_len, "shtej"))
		{
			wc->freeregion_plx = 1;
		}

		wc->plx = ioremap(wc->plx_region, wc->plx_len);
		if (!wc->plx)
		{
			printk("Cannot ioremap memory space at 0x%lx\n", wc->plx_region);
			goto err_out_free_wc;
		}

		if (request_mem_region(wc->dsp_region, wc->dsp_len, "shtej"))
		{
			wc->freeregion_dsp = 1;
		}

		wc->dsp = ioremap(wc->dsp_region, wc->dsp_len);
		if (!wc->dsp)
		{
			printk("Cannot ioremap memory space at 0x%lx\n", wc->dsp_region);
			goto err_out_free_wc;
		}


		pci_set_master(pdev);

		pci_set_drvdata(pdev, wc);

		for(x = 0; x < MAX_CARDS; x++)
		{
			if (!cards[x]) break;
		}

		if (x >= MAX_CARDS)
		{
			printk("No card[] slot available!1\n");
			goto err_out_free_wc;
		}

		if (shtej_hardware_init(wc, d->flags))
		{
			printk("shtej: hardware init 1 fail\n");
			goto err_out_free_wc;
		}
		
		if (shtej_init_dma(wc))
		{
			printk("shtej: init dma fail\n");
			goto err_out_free_wc;
		}
		
		shtej_init_readandwrite_chunk(wc);

		wc->num = x;
		cards[x] = wc;

		for (x = 0; x < wc->numspans; x++)
		{
			if (wc->t1e1 & (1 << x))
			{

				wc->tspans[x] = kmalloc(sizeof(struct shtej_span) + sizeof(struct dahdi_chan) * 31, GFP_KERNEL);
				if (wc->tspans[x])
				{
					memset(wc->tspans[x], 0x0, sizeof(struct shtej_span) + sizeof(struct dahdi_chan) * 31);
					wc->tspans[x]->spantype = TYPE_E1;
				}
				for(xx=0;xx<31;xx++)
				{
				  wc->tspans[x]->_chans[xx]=&wc->tspans[x]->chans[xx];
				}
			}
			else
			{
				wc->tspans[x] = kmalloc(sizeof(struct shtej_span) + sizeof(struct dahdi_chan) * 24, GFP_KERNEL);
				if (wc->tspans[x])
				{
					memset(wc->tspans[x], 0x0, sizeof(struct shtej_span) + sizeof(struct dahdi_chan) * 24);
					wc->tspans[x]->spantype = TYPE_T1;
				}
				for(xx=0;xx<24;xx++)
				{
				  wc->tspans[x]->_chans[xx]=&wc->tspans[x]->chans[xx];
				}
			}

			if (!wc->tspans[x])
				goto err_out_free_wc;
			wc->tspans[x]->spanflags |= d->flags;
		}

#if 1
		if (request_irq(pdev->irq, shtej_interrupt, DAHDI_IRQ_SHARED_DISABLED, "shtej", wc))
		{
			printk("shtej: Unable to request IRQ %d\n", pdev->irq);
			goto err_out_free_wc;
		}
#endif
		
		if (shtej_hardware_init_2(wc))
		{
			free_irq(pdev->irq, wc);
			pci_set_drvdata(pdev, NULL);
			printk("shtej: hardware init 2 fail\n");
			goto err_out_free_wc;
		}

		shtej_init_spans(wc);
		shtej_SerialNo(wc);
		for (x = 0; cards[x]; x++)
		{
			if (shtej_launch(cards[x]))
			{
				free_irq(pdev->irq, wc);
				pci_set_drvdata(pdev, NULL);
				printk("shtej: launch cards fail\n");
				goto err_out_free_wc;
			}
		}
		
		wc->InitCard = 0;
		
		wc->BTime = 1;
		wc->complete = 1;
		wc->Index = 0;
		wc->filecount = 0;
		
		for(y=0; y<2; y++)
		{
			for(x=0; x<31; x++)
			{
				wc->tmpwriterec[y][x] = (unsigned char *)kmalloc(0x20000, GFP_KERNEL);
				wc->tmpwrite[y][x] = (unsigned char *)kmalloc(0x20000, GFP_KERNEL);
	
				if(wc->tmpwriterec[y][x] == NULL)
				{
					goto err_out_free_wc;
				}
			}
		}

		init_waitqueue_head(&wc->regq);
		sprintf(qname, "shtej_dsp%d", wc->num);
		wc->shtej_wq = create_singlethread_workqueue(qname);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
		INIT_WORK(&wc->shtej_work, wque_run);
#else
		INIT_WORK(&wc->shtej_work, wque_run, wc);
#endif

		sprintf(qname, "shtej%d", wc->num);
		init_waitqueue_head(&wc->rec_span_q);
		wc->rec_span_wq = create_singlethread_workqueue(qname);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
		INIT_WORK(&wc->rec_span_work, wque_rec_span);
#else
		INIT_WORK(&wc->rec_span_work, wque_rec_span, wc);
#endif
		
		wc->irq_enable = 1;
		shtej_enable_cpld_interrupts(wc);
		shtej_enable_local_interrupts(wc);
		shtej_enable_pci_interrupts(wc);

		res = 0;
	}

	return res;

err_out_free_wc:
	if (wc->freeregion_plx)
		release_mem_region(wc->plx_region, wc->plx_len);
	if (wc->freeregion_dsp)
		release_mem_region(wc->dsp_region, wc->dsp_len);
	if (wc->plx)
		iounmap(wc->plx);
	if (wc->dsp)
		iounmap(wc->dsp);
	if (wc)
	{
		for (x = 0; x < wc->numspans; x++) kfree(wc->tspans[x]);
		kfree(wc);
	}
	return -ENODEV;

}

static void shtej_release(struct shtej *wc)
{
	int x;
	
	if (wc->tspans[0]->span.flags & DAHDI_FLAG_REGISTERED)
		dahdi_unregister(&wc->tspans[0]->span);
	
	if(wc->numspans == 2 || wc->numspans == 4)
	{
		if (wc->tspans[1]->span.flags & DAHDI_FLAG_REGISTERED)
			dahdi_unregister(&wc->tspans[1]->span);
	}

	if (wc->numspans == 4)
	{
		if (wc->tspans[2]->span.flags & DAHDI_FLAG_REGISTERED)
			dahdi_unregister(&wc->tspans[2]->span);
		if (wc->tspans[3]->span.flags & DAHDI_FLAG_REGISTERED)
			dahdi_unregister(&wc->tspans[3]->span);
	}


	if (wc->freeregion_plx)
		release_mem_region(wc->plx_region, wc->plx_len);
	if (wc->freeregion_dsp)
		release_mem_region(wc->dsp_region, wc->dsp_len);

	if (wc->plx)
		iounmap(wc->plx);
	if (wc->dsp)
		iounmap(wc->dsp);
	
	if (wc)
	{
		for (x = 0; x < wc->numspans; x++) kfree(wc->tspans[x]);
		kfree(wc);
	}
	printk("Freed a TEJ/PCI\n");
}

static void __devexit shtej_remove_one(struct pci_dev *pdev)
{
	struct shtej *wc = pci_get_drvdata(pdev);
	int x, y;
	
	if (wc)
	{
		wc->irq_enable = 0;

		shtej_stop_dma(wc);
		shtej_wait_ms(5000);
		
		shtej_disable_pci_interrupts(wc);
		shtej_disable_local_interrupts(wc);
		
		shtej_card_reset(wc);
		
		free_irq(pdev->irq, wc);
		
		cards[wc->num] = NULL;

		destroy_workqueue(wc->rec_span_wq);

		destroy_workqueue(wc->shtej_wq);
		
		for(y=0; y<2; y++)
		{
			for(x=0; x<31; x++)
			{
				kfree(wc->tmpwriterec[y][x]);
				kfree(wc->tmpwrite[y][x]);
			}
		}

		shtej_release(wc);
	}
}

static struct pci_device_id shtej_pci_tbl[] = {

	{ 0x10b5, 0x00fa, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (unsigned long) &shtej[0]},
	{ 0x10b5, 0x00fb, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (unsigned long) &shtej[1]},
	{ 0x10b5, 0x00fc, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (unsigned long) &shtej[2]},
	{ 0 }
};

MODULE_DEVICE_TABLE(pci, shtej_pci_tbl);

static struct pci_driver shtej_driver = {
	name: 	"shtej",
	probe: 	shtej_init_one,
#ifdef LINUX26
	remove:	__devexit_p(shtej_remove_one),
#else
	remove:	shtej_remove_one,
#endif
	suspend: NULL,
	resume:	NULL,
	id_table: shtej_pci_tbl,
};

static int __init shtej_init(void)
{
	int res;
	
	res =dahdi_pci_module(&shtej_driver);
	if (res)
		return -ENODEV;
	return 0;
}

static void __exit shtej_cleanup(void)
{
	pci_unregister_driver(&shtej_driver);
}

module_param(debug, int, 0600);
module_param(timing, int, 0600);

MODULE_DESCRIPTION("TEJ-4a TEJ-2a TEJ-1a PCI Driver");
MODULE_AUTHOR("Xiong ce <liny@sanhuid.com>");
#if defined(MODULE_ALIAS)
MODULE_ALIAS("shtej");
#endif
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

module_init(shtej_init);
module_exit(shtej_cleanup);
