/*
 * FXM-16a/PCI card Interface Driver for synway Telephony interface
 *
 * Written by Crazy chen <cjd@sanhuid.com>, Unknown Xu <xxh@sanhuid.com> 
 * 
 * Copyright (C) 2007, Linux Support Services, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
*/

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#ifdef STANDALONE_ZAPATA
#include "../zaptel.h"
#else
#include <linux/zaptel.h>
#endif
#ifdef LINUX26
#include <linux/moduleparam.h>
#endif
#include "fxm16apcie.h"

static int battdebounce = DEFAULT_BATT_DEBOUNCE;
static int battthresh = DEFAULT_BATT_THRESH;
static int debug = 0;
static int alawoverride = 0;
static int ECHO_OPEN = 1;

static struct shfxm *cards[WC_MAX_IFACES];

static int plx_dmasgl_transfer_listadd(struct shfxm *wc, DMA_CHANNEL channel, struct dma_transfer_element *dmadata);
static int plx_dmasgl_transfer(struct shfxm *wc, DMA_CHANNEL channel);

static struct shfxm_desc shfxm = { "FXM-16a/PCIe Prototype", 0 };

int kernel_write(int count, unsigned char * addr, char * filename)
{
#if 1
	struct file *file;
	mm_segment_t old_fs;
	loff_t pos = 0;
	int  result = -ENOSYS;

	file   =   filp_open(filename,   O_RDWR   |   O_APPEND   |   O_CREAT, 0644);
	if (IS_ERR(file))
	{
		return   0;
	}

	if (!file->f_op->write)
	{
		return result;
	}
	old_fs = get_fs();
	set_fs(get_ds());
	result = file->f_op->write(file, addr, count, &pos);
	set_fs(old_fs);
	
	filp_close(file, NULL);

	return result;
#endif
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
static void wque_run(struct work_struct *work)
{
	struct shfxm *wc = container_of(work, struct shfxm, shfxm_work);
#else
static void wque_run(void *work_data)
{
	struct shfxm *wc = (struct shfxm *)work_data;
#endif
	char tmpname[50];
	int t;
	int r;
	int i;
	
	do_gettimeofday(&wc->qtime);
	
	t = (wc->Index+1)%2;
	for(i=0; i<NUM_CARDS; i++)
	{
			sprintf(tmpname, "/mnt/recf/Rec_EC%d_%d_%05d.pcm", wc->pos, i, (int)wc->qtime.tv_sec);
			r = kernel_write(0x20000, wc->tmpwriter[t][0][i], tmpname);
			/*sprintf(tmpname, "/mnt/recf/Rec%d_%d_%05d.pcm", wc->pos, i, (int)wc->qtime.tv_sec);
			r = kernel_write(0x20000, wc->tmpwriter[t][1][i], tmpname);
			sprintf(tmpname, "/mnt/recf/Play%d_%d_%05d.pcm", wc->pos, i, (int)wc->qtime.tv_sec);
			r = kernel_write(0x20000, wc->tmpwriter[t][2][i], tmpname);*/
	}

	for(i=0; i<NUM_CARDS; i++)
	{
			sprintf(tmpname, "/mnt/tri/T%d_%d_%05d.pcm", wc->pos, i, (int)wc->qtime.tv_sec);
			r = kernel_write(0x20000, wc->tmpwrite[t][i], tmpname);
	}

	wc->complete = 1;
}

static void DebugInt(struct shfxm *wc)
{
	int tmp;

	if(wc->BTime == 0)
	{
		do_gettimeofday(&wc->stime);
		wc->BTime = 1;
	}
	else
	{
		do_gettimeofday(&wc->etime);
		tmp = (wc->etime.tv_sec - wc->stime.tv_sec)*1000000 + (wc->etime.tv_usec - wc->stime.tv_usec);
		wc->BTime = 0;
		
		if(tmp > 1200)
			printk("FXM16APCIE Interrupt Time-----= %d\n", tmp);
	}
}

static inline int pci_getconfig_dword(struct shfxm *wc, int offset, u32 *val)
{
	return pci_read_config_dword(wc->dev, offset, val);
}

static inline int pci_setconfig_dword(struct shfxm *wc, int offset, u32 val)
{
	return pci_write_config_dword(wc->dev, offset, val);
}

static inline unsigned long plx_getreg(struct shfxm *wc, unsigned long offset)
{
	return	readl(wc->plx + offset);
}

static inline void plx_setreg(struct shfxm *wc, unsigned long offset, unsigned long val)
{
	writel(val, wc->plx + offset);
} 

static unsigned short dsp_getreg_word(struct shfxm *wc, int offset)
{
	return readw(wc->dsp + offset);
}

static void dsp_setreg_word(struct shfxm *wc, int offset, unsigned short val)
{
	writew(val, wc->dsp + offset);
}

static int shfxm_setdsp_data(struct shfxm *wc, int dspsel, unsigned short addr, unsigned short *data, unsigned long len)
{
	unsigned short highaddr;
	unsigned short lowaddr;
	unsigned long i;

	if (addr % 2 == 1 && dspsel > 0)
	{
		return -1;
	}

	for (i = 0; i < len; i++, addr += 2)
	{
		highaddr = addr & 0xff00;
		lowaddr = addr & 0xff;
		if (wc->dsp_addr != (0x10000 * dspsel + highaddr + 0x1))
		{
			wc->dsp_addr = 0x10000 * dspsel + highaddr + 0x1;
			plx_setreg(wc, 0x4, wc->dsp_addr);
		}
		dsp_setreg_word(wc, lowaddr, *(data + i));
	}
	return 0;

}

static int shfxm_getdsp_data(struct shfxm *wc, int dspsel, unsigned short addr, unsigned short *data, unsigned long len)
{
	unsigned short highaddr;
	unsigned short lowaddr;
	unsigned long i;

	if (addr % 2 == 1 && dspsel > 0)
	{
		return -1;
	}

	for (i = 0; i < len; i++, addr += 2)
	{
		highaddr = addr & 0xff00;
		lowaddr = addr & 0xff;
		if (wc->dsp_addr != (0x10000 * dspsel + highaddr + 0x1))
		{
			wc->dsp_addr = 0x10000 * dspsel + highaddr + 0x1;
			plx_setreg(wc, 0x4, wc->dsp_addr);
		}
		*(data + i) = dsp_getreg_word(wc, lowaddr);
	}

	return 0;
}

static inline void shfxm_wait_ms(long ms)
{
	udelay(ms);
}

static void shfxm_clear_cpld_interrupts(struct shfxm *wc)
{
	unsigned short val;

	shfxm_getdsp_data(wc, 0, 0x0c, &val, sizeof(val) / sizeof(unsigned short));
}

static void plx_sync_reg_modify(struct plx_reg_data *reg)
{
	unsigned long regval;

	regval = plx_getreg(reg->wc, reg->offset);
	regval |= reg->bitstoset;
	regval &= ~(reg->bitstoclear);
	plx_setreg(reg->wc, reg->offset, regval);

}

static void shfxm_enable_pci_interrupts(struct shfxm *wc)
{
	struct plx_reg_data regdata;
	unsigned long regval;

	regdata.wc = wc;
	regdata.offset = PCI9056_INT_CTRL_STAT;
	regdata.bitstoset = (1 << 8);
	regdata.bitstoclear = 0;

	plx_sync_reg_modify(&regdata);

	regval = plx_getreg(regdata.wc, regdata.offset);

}

static void shfxm_disable_pci_interrupts(struct shfxm *wc)
{
	struct plx_reg_data regdata;

	regdata.wc = wc;
	regdata.offset = PCI9056_INT_CTRL_STAT;
	regdata.bitstoset = 0;
	regdata.bitstoclear = (1 << 8);

	plx_sync_reg_modify(&regdata);

}

static void shfxm_enable_local_interrupts(struct shfxm *wc)
{
	struct plx_reg_data regdata;
	unsigned long regval;

	regdata.wc = wc;
	regdata.offset = PCI9056_INT_CTRL_STAT;
	regdata.bitstoset = (1 << 11);
	regdata.bitstoclear = 0;

	plx_sync_reg_modify(&regdata);

	regval = plx_getreg(regdata.wc, regdata.offset);

}

static void shfxm_disable_local_interrupts(struct shfxm *wc)
{
	struct plx_reg_data regdata;

	regdata.wc = wc;
	regdata.offset = PCI9056_INT_CTRL_STAT;
	regdata.bitstoset = 0;
	regdata.bitstoclear = (1 << 11);

	plx_sync_reg_modify(&regdata);

}

static inline void shfxm_fxo_ringcheck(struct shfxm *wc, int card)
{
	signed char b;
	unsigned char v;
	unsigned short reg;
	unsigned char lastvoltage;

	reg = 0x0;
	reg |= card & 0xf;
	lastvoltage = wc->mod[card].fxo.voltage;
	shfxm_setdsp_data(wc, 0, 0x04, &reg, sizeof(reg) / sizeof(unsigned short));
	shfxm_wait_ms(6);
	shfxm_getdsp_data(wc, 0, 0x02, &reg, sizeof(reg) / sizeof(unsigned short));
	wc->mod[card].fxo.voltage = (unsigned char)reg & 0xff;
	
	b = (wc->mod[card].fxo.voltage - 132) / 2;

	if (!wc->mod[card].fxo.offhook)
	{
		v = abs(wc->mod[card].fxo.voltage - lastvoltage);
		if (v > 15 && wc->mod[card].fxo.battery)
		{
			wc->mod[card].fxo.ringdebounce += (ZT_CHUNKSIZE * 8);
			if (wc->mod[card].fxo.ringdebounce >= ZT_CHUNKSIZE * 64)
			{
				if (!wc->mod[card].fxo.wasringing)
				{
					wc->mod[card].fxo.wasringing = 1;
					zt_hooksig(&wc->chans[card], ZT_RXSIG_RING);
					if (debug)
						printk("RING on %d/%d!\n", wc->span.spanno, card + 1);
				}
				wc->mod[card].fxo.ringdebounce = ZT_CHUNKSIZE * 64;
			}
		}
		else
		{
			wc->mod[card].fxo.ringdebounce -= ZT_CHUNKSIZE * 2;
			if (wc->mod[card].fxo.ringdebounce <= 0)
			{
				if (wc->mod[card].fxo.wasringing) {
					wc->mod[card].fxo.wasringing = 0;
					zt_hooksig(&wc->chans[card], ZT_RXSIG_OFFHOOK);
					if (debug)
						printk("NO RING on %d/%d!\n", wc->span.spanno, card + 1);
				}
				wc->mod[card].fxo.ringdebounce = 0;
			}
		}
	}

	if (abs(b) < battthresh)
	{
		wc->mod[card].fxo.nobatttimer++;

		if (wc->mod[card].fxo.battery && !wc->mod[card].fxo.battdebounce)
		{
			if (debug)
				printk("NO BATTERY on %d/%d!\n", wc->span.spanno, card + 1);
			wc->mod[card].fxo.battery =  0;
#ifdef	JAPAN
			if ((!wc->ohdebounce) && wc->offhook)
			{
				zt_hooksig(&wc->chans[card], ZT_RXSIG_ONHOOK);
				if (debug)
					printk("Signalled On Hook\n");
#ifdef	ZERO_BATT_RING
				wc->onhook++;
#endif
			}
#else
			zt_hooksig(&wc->chans[card], ZT_RXSIG_ONHOOK);
#endif
			wc->mod[card].fxo.battdebounce = battdebounce;
		}
		else if (!wc->mod[card].fxo.battery)
			wc->mod[card].fxo.battdebounce = battdebounce;
	}
	else if (abs(b) > battthresh)
	{
		if (!wc->mod[card].fxo.battery && !wc->mod[card].fxo.battdebounce)
		{
			if (debug)
				printk("BATTERY on %d/%d (%s)!\n", wc->span.spanno, card + 1, 
						(b < 0) ? "-" : "+");			    
#ifdef	ZERO_BATT_RING
			if (wc->onhook)
			{
				wc->onhook = 0;
				zt_hooksig(&wc->chans[card], ZT_RXSIG_OFFHOOK);
				if (debug)
					printk("Signalled Off Hook\n");
			}
#else
			zt_hooksig(&wc->chans[card], ZT_RXSIG_OFFHOOK);
#endif
			wc->mod[card].fxo.battery = 1;
			wc->mod[card].fxo.nobatttimer = 0;
			wc->mod[card].fxo.battdebounce = battdebounce;
		}
		else if (wc->mod[card].fxo.battery)
			wc->mod[card].fxo.battdebounce = battdebounce;

		if (wc->mod[card].fxo.lastpol >= 0)
		{
			if (b < 0)
			{
				wc->mod[card].fxo.lastpol = -1;
				wc->mod[card].fxo.polaritydebounce = POLARITY_DEBOUNCE;
			}
		} 
		if (wc->mod[card].fxo.lastpol <= 0)
		{
			if (b > 0)
			{
				wc->mod[card].fxo.lastpol = 1;
				wc->mod[card].fxo.polaritydebounce = POLARITY_DEBOUNCE;
			}
		}
	}
	else
	{
		wc->mod[card].fxo.battdebounce = battdebounce;
	}

	if (wc->mod[card].fxo.battdebounce)
		wc->mod[card].fxo.battdebounce--;
#if 0
	if (wc->mod[card].fxo.polaritydebounce)
	{
		wc->mod[card].fxo.polaritydebounce--;
		if (wc->mod[card].fxo.polaritydebounce < 1)
		{
			if (wc->mod[card].fxo.lastpol != wc->mod[card].fxo.polarity)
			{
				if (debug)
					printk("%lu Polarity reversed (%d -> %d)\n", jiffies, 
							wc->mod[card].fxo.polarity, 
							wc->mod[card].fxo.lastpol);
				if (wc->mod[card].fxo.polarity)
					zt_qevent_lock(&wc->chans[card], ZT_EVENT_POLARITY);
				wc->mod[card].fxo.polarity = wc->mod[card].fxo.lastpol;
			}
		}
	}
#endif

}

static inline void shfxm_fxs_hookcheck(struct shfxm *wc, int card)
{
	unsigned short val;
	int hook;

	shfxm_getdsp_data(wc, 0, 0x04, &val, sizeof(val) / sizeof(unsigned short));

	hook = ((~val >> card) & 1);

	if (hook != wc->mod[card].fxs.lastrxhook)
	{
		/* Reset the debounce (must be multiple of 4ms) */
		wc->mod[card].fxs.debounce = 8 * (2 * 8);
	}
	else
	{
		if (wc->mod[card].fxs.debounce > 0)
		{
			wc->mod[card].fxs.debounce-= 8 * ZT_CHUNKSIZE;
			if(!wc->mod[card].fxs.debounce)
			{
				wc->mod[card].fxs.debouncehook = hook;
			}
			if (!wc->mod[card].fxs.oldrxhook && wc->mod[card].fxs.debouncehook)
			{
				/* Off hook */
				if (debug)
					printk("shfxm: Card %d Going off hook\n", card);
				zt_hooksig(&wc->chans[card], ZT_RXSIG_OFFHOOK);

				wc->mod[card].fxs.oldrxhook = 1;

			}
			else if (wc->mod[card].fxs.oldrxhook && !wc->mod[card].fxs.debouncehook)
			{
				/* On hook */
				if (debug)
					printk("shfxm: Card %d Going on hook\n", card);
				zt_hooksig(&wc->chans[card], ZT_RXSIG_ONHOOK);
				wc->mod[card].fxs.oldrxhook = 0;
			}
		}
	}
	wc->mod[card].fxs.lastrxhook = hook;
}

static inline void shfxm_fxo_hookcheck(struct shfxm *wc, int card)
{
	unsigned short reg;

	if (wc->mod[card].fxo.offhook) 
		wc->regringhook |= (1 << card);
	else
		wc->regringhook &= ~(1 << card);

	if (card < 8)
	{
		reg = wc->regringhook & 0xff;
		shfxm_setdsp_data(wc, 0, 0x08, &reg, sizeof(reg) / sizeof(unsigned short));
	}
	else
	{
		reg = wc->regringhook >> 8 & 0xff;
		shfxm_setdsp_data(wc, 0, 0x0a, &reg, sizeof(reg) / sizeof(unsigned short));
	}
	wc->mod[card].fxo.flags &= ~FLAG_DO_OFFHOOK;
}

static inline void shfxm_fxs_sendring(struct shfxm *wc, int card)
{
	unsigned short reg;

	if (wc->mod[card].fxs.offhook) 
		wc->regringhook |= (1 << card);
	else
		wc->regringhook &= ~(1 << card);

	if (card < 8) {
		reg = wc->regringhook & 0xff;
		shfxm_setdsp_data(wc, 0, 0x08, &reg, sizeof(reg) / sizeof(unsigned short));
	} else {
		reg = wc->regringhook >> 8 & 0xff;
		shfxm_setdsp_data(wc, 0, 0x0a, &reg, sizeof(reg) / sizeof(unsigned short));
	}
	wc->mod[card].fxs.flags &= ~FLAG_DO_SENDRING;
}

static int plx_checkdmaready(struct shfxm *wc, DMA_CHANNEL channel)
{
	u8 i;
	u8 shift;
	u32 regvalue;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			shift = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			shift = 8;
			break;
		default:
			return -1;

	}

	regvalue = plx_getreg(wc, PCI9056_DMA_COMMAND_STAT);
	if ((regvalue & ((1 << 4) << shift)) == 0)
	{
		if(debug > 66)
			printk("checkdmaready: dma channel is carrently active\n");
		return -1;
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
	{
		if(debug > 66)
			printk("checkdmaready: dma channel has not been opened for SGL DMA\n");
		return -1;
	}
	
	wc->dmainfo[i].dmadatacount = 0;
	
	return 0;
}

static inline void shfxm_receiveprep(struct shfxm *wc)
{
	int x, i, j;
	struct dma_transfer_element el;
	unsigned short val;

	if(wc->readhasdata[wc->readcurpos] == 1)
	{
		for (i = 0; i < NUM_CARDS; i++)
		{
			if (wc->cardflag & (1 << i))
			{
				for (x = 0; x < ZT_CHUNKSIZE; x++)
					wc->chans[i].readchunk[x] = wc->readchunk[wc->readcurpos][0][i][x];
				
				if(debug > 55)
				{
					if((wc->filecount+ZT_CHUNKSIZE)<=0x20000)
					{
						for (j = 0; j < RECDATA_NUM; j++)
						{
							for (x = 0; x < ZT_CHUNKSIZE; x++)
								wc->tmpwriter[wc->Index][j][i][wc->filecount+x] = wc->readchunk[wc->readcurpos][j][i][x];
						}
					}
				}
			}
		}
		
		if(debug > 66)
		{
			wc->tflags = (wc->tflags+1)&0xfff;
			if(wc->tflags != wc->readchunk_frame[wc->readcurpos]->readflag)
			{
				printk("R=====\t[%x, %x]\n", wc->tflags, wc->readchunk_frame[wc->readcurpos]->readflag);
				wc->tflags = wc->readchunk_frame[wc->readcurpos]->readflag;
			}
		}
		
		wc->readhasdata[wc->readcurpos] = 0;
		wc->readcurpos = (wc->readcurpos+1)%RECBLOCK_BUF;

		zt_receive(&wc->span);
	}

	shfxm_getdsp_data(wc, 1, RECBUF_FLAG * 2, &val, sizeof(val) / sizeof(unsigned short));
	if(wc->RECFlag == val && debug > 66)
		printk("FXM Debug RECBUF_FLAG error [%x, %x]\n", val, wc->pos);

	if(wc->RECFlag_First == 1)
	{
		wc->RECFlag_First = 0;
		wc->RECFlag = val;
		wc->readfillpos = wc->RECFlag;
		wc->rpos = wc->readfillpos;
	}

	if(plx_checkdmaready(wc, DMA_CHANNEL1) == 0)
	{
		for(i = wc->readfillpos; i != wc->rpos;)
		{
			if(wc->readhasdata[i] == 0)
			{
				wc->readhasdata[wc->readfillpos] = 1;
				wc->readfillpos = (wc->readfillpos+1)%RECBLOCK_BUF;
			}
			else
				break;
				
			i = (i+1)%RECBLOCK_BUF;
		}
		
		if(((wc->readfillpos - wc->readcurpos + RECBLOCK_BUF)%RECBLOCK_BUF) > RECLIMIT_BUF)
		{
			wc->readcurpos	= 0;
			wc->readfillpos	= 0;
			wc->rpos				= 0;

			for(x = 0; x < RECBLOCK_BUF; x++)
				wc->readhasdata[x] = 0;
		}
		
		wc->rpos = wc->readfillpos;
		for(i = wc->RECFlag; i != val;)
		{
			el.localtopcidma = 1;
			el.localaddr = 0x10000 + (RECBLOCK_LEN * i + BASEADDR_RECBUF) * 2;
			el.u.pciaddrlow = wc->dmainfo[1].dspbuffer.busphysical+(NUM_CARDS * ZT_CHUNKSIZE * RECDATA_NUM + sizeof(unsigned short))*wc->rpos;
			el.transfercount = wc->dmainfo[1].dspbuffer.size/RECBLOCK_BUF;
			plx_dmasgl_transfer_listadd(wc, DMA_CHANNEL1, &el);
			
			i = (i+1)%RECBLOCK_NUM;
			wc->rpos = (wc->rpos+1)%RECBLOCK_BUF;
		}

		wc->RECFlag = val;
	}
}

static inline void shfxm_transmitprep(struct shfxm *wc)
{
	
	int x;
	int i;
	unsigned short val;
	struct dma_transfer_element el;
	
	if(((wc->writefillpos - wc->writecurpos + PLAYBLOCK_BUF)%PLAYBLOCK_BUF) > PLAYLIMIT_BUF)
	{
		wc->writecurpos	= 0;
		wc->writefillpos	= 0;
		
		for(x = 0; x < PLAYBLOCK_BUF; x++)
			wc->writehasdata[x] = 0;
	}
	
	zt_transmit(&wc->span);
	
	wc->writechunk_frame[wc->writefillpos]->writeflag = wc->Ttflags;
	wc->Ttflags = (wc->Ttflags + 1) & 0xfff;
	for (i = 0; i < NUM_CARDS; i++)
	{
		if (wc->cardflag & (1 << i))
		{
			for (x = 0; x < ZT_CHUNKSIZE; x++)
			{
				wc->writechunk[wc->writefillpos][i][x] = wc->chans[i].writechunk[x];
			}
			
#if 1
			if(debug > 55)
			{
				if((wc->filecount+ZT_CHUNKSIZE)<=0x20000)
				{
					for (x = 0; x < ZT_CHUNKSIZE; x++)
						wc->tmpwrite[wc->Index][i][wc->filecount+x] = wc->chans[i].writechunk[x];
				}
			}
#endif

		}
	}
	wc->writehasdata[wc->writefillpos] = 1;
	wc->writefillpos = (wc->writefillpos+1)%PLAYBLOCK_BUF;
	
	shfxm_getdsp_data(wc, 1, PLAYBUF_FLAG * 2, &val, sizeof(val) / sizeof(unsigned short));
	if(wc->TRIFlag == val && debug > 66)
		printk("FXM Debug PLAYBUF_FLAG error [%x, %x]\n", val, wc->pos);
	
	if(wc->TRIFlag_First == 1)
	{
		wc->TRIFlag_First = 0;
		wc->TRIFlag = (val-1+PLAYBLOCK_NUM)%PLAYBLOCK_NUM;
	}
	
	if(plx_checkdmaready(wc, DMA_CHANNEL0) == 0)
	{
		for(i = wc->writecurpos; i != wc->writefillpos;)
		{
			if(wc->writehasdata[i] == 1)
			{
				if(val == wc->TRIFlag)
					break;
				
				wc->writehasdata[i] = 0;
	
				el.localtopcidma = 0;
				el.localaddr = 0x10000 + (PLAYBLOCK_LEN * wc->TRIFlag + BASEADDR_PLAYBUF) * 2;
				el.u.pciaddrlow = wc->dmainfo[0].dspbuffer.busphysical+(NUM_CARDS*ZT_CHUNKSIZE+sizeof(unsigned short))*i;
				el.transfercount = wc->dmainfo[0].dspbuffer.size/PLAYBLOCK_BUF;
				plx_dmasgl_transfer_listadd(wc, DMA_CHANNEL0, &el);
				
				wc->TRIFlag = (wc->TRIFlag+1)%PLAYBLOCK_NUM;
			}
			
			i = (i+1)%PLAYBLOCK_BUF;
		}
		//wc->writecurpos = i;
		wc->writecurpos = wc->writefillpos;
	}
}

static void shfxm_do_echocan(struct shfxm *wc, int card)
{
	unsigned short val;

	if (wc->echocan & (1 << card) && ECHO_OPEN == 1)
		val = 1 << 12;
	else
		val = 0 << 12;

	shfxm_setdsp_data(wc, 1, (SET_CW2 + card) * 2,  &val, sizeof(val) / sizeof(unsigned short));

	if (wc->modtype[card] == MOD_TYPE_FXO)
		wc->mod[card].fxo.flags &= ~FLAG_DO_ECHOCAN;
	else
		wc->mod[card].fxs.flags &= ~FLAG_DO_ECHOCAN;
}


static void do_interrupts(void *p)
{
	u32 regvalue;
	unsigned long intsrc;
	struct shfxm *wc = (struct shfxm *)p;

	intsrc = wc->interruptsrc;

	if (intsrc & INTR_TYPE_LOCAL_1)
	{
		int x;
		int mode;

		mode = wc->intcount & 0x7;

		for (x = mode * 2; x <= mode * 2 + 1; x++)
		{
			if (wc->cardflag & (1 << x))
			{
				if (wc->modtype[x] == MOD_TYPE_FXO)
					shfxm_fxo_ringcheck(wc, x);
				else
					shfxm_fxs_hookcheck(wc, x);
			}
		}

		for (x = 0; x < NUM_CARDS; x++)
		{
			if (wc->cardflag & (1 << x))
			{
				if (wc->modtype[x] == MOD_TYPE_FXO)
				{
					if (wc->mod[x].fxo.flags & FLAG_DO_OFFHOOK)
						shfxm_fxo_hookcheck(wc, x);
					if (wc->mod[x].fxo.flags & FLAG_DO_ECHOCAN)
						shfxm_do_echocan(wc, x);
				}
				else
				{
					if (wc->mod[x].fxs.flags & FLAG_DO_SENDRING)
						shfxm_fxs_sendring(wc, x);
					if (wc->mod[x].fxs.flags & FLAG_DO_ECHOCAN)
						shfxm_do_echocan(wc, x);
				}
			}
		}
		
		if(debug > 50)
			DebugInt(wc);

		if(debug > 55)
		{
			if(wc->complete == 1 && (wc->filecount+ZT_CHUNKSIZE) > 0x20000)
			{
				wc->Index = (wc->Index+1)%2;
				queue_work(wc->shfxm_wq, &wc->shfxm_work);
				wc->complete = 0;
				wc->filecount = 0;
			}
		}

		shfxm_receiveprep(wc);
		shfxm_transmitprep(wc);
		
		if(debug > 55)
		{
			if((wc->filecount+ZT_CHUNKSIZE)<=0x20000)
				wc->filecount += ZT_CHUNKSIZE;
		}

		if (wc->dmainfo[0].dmadatacount)
			plx_dmasgl_transfer(wc, DMA_CHANNEL0);

		if (wc->dmainfo[1].dmadatacount)
			plx_dmasgl_transfer(wc, DMA_CHANNEL1);

		shfxm_enable_local_interrupts(wc);
	}

	if (intsrc & INTR_TYPE_DOORBELL)
	{
		regvalue = plx_getreg(wc, 0x64);
		plx_setreg(wc, 0x64, regvalue);
	}

	if (intsrc & INTR_TYPE_PCI_ABORT)
	{
		pci_getconfig_dword(wc, 0x4, &regvalue);
		pci_setconfig_dword(wc, 0x4, regvalue);
	}

	if (intsrc & INTR_TYPE_OUTBOUND_POST)
	{
		plx_setreg(wc, PCI9056_OUTPOST_INT_MASK, (1 << 3));
	}

	shfxm_enable_pci_interrupts(wc);
}

ZAP_IRQ_HANDLER(shfxm_interrupt)
{
	unsigned long interruptreg;
	unsigned long interruptsrc;
	unsigned long regvalue;
	struct shfxm *wc = dev_id;


	interruptreg = plx_getreg(wc, PCI9056_INT_CTRL_STAT);

	if (interruptreg == 0xffffffff)
	{
		return IRQ_NONE;
	}

	if ((interruptreg & (1 << 8)) == 0)
	{
		return IRQ_NONE;
	}

	interruptsrc = INTR_TYPE_NONE;

	if ((interruptreg & (1 << 13)) && (interruptreg & (1 << 9)))
	{
		interruptsrc |= INTR_TYPE_DOORBELL;
	}

	if ((interruptreg & (1 << 14)) && (interruptreg & (1 << 10)))
	{
		interruptsrc |= INTR_TYPE_PCI_ABORT;
	}

	if ((interruptreg & (1 << 15)) && (interruptreg & (1 << 11)))
	{
		interruptsrc |= INTR_TYPE_LOCAL_1;
		shfxm_clear_cpld_interrupts(wc);
		shfxm_disable_local_interrupts(wc);
		wc->intcount++;
	}

	if ((interruptreg & (1 << 21)) && (interruptreg & (1 << 18)))
	{
		regvalue = plx_getreg(wc, PCI9056_DMA0_MODE);
		if (regvalue & (1 << 17)) 
			interruptsrc |= INTR_TYPE_DMA_0;
	}

	if ((interruptreg & (1 << 22)) && (interruptreg & (1 << 19)))
	{
		regvalue = plx_getreg(wc, PCI9056_DMA1_MODE);
		if (regvalue & (1 << 17)) 
			interruptsrc |= INTR_TYPE_DMA_1;
	}

	regvalue = plx_getreg(wc, PCI9056_OUTPOST_INT_STAT);
	if (regvalue & (1 << 3))
	{
		regvalue = plx_getreg(wc, PCI9056_OUTPOST_INT_MASK);
		if ((regvalue & (1 << 3)) == 0)
			interruptsrc |= INTR_TYPE_OUTBOUND_POST;
	}	


	if (interruptsrc == INTR_TYPE_NONE)
	{
		return IRQ_NONE;
	}

	wc->interruptsrc = interruptsrc;

	shfxm_disable_pci_interrupts(wc);

	if (wc->irq_enable)
		do_interrupts((void *)wc);

#ifdef LINUX26
	return IRQ_RETVAL(1);
#endif		

}


static int __devinit shfxm_launch(struct shfxm *wc)
{

	if(wc->span.flags & ZT_FLAG_REGISTERED)
		return 0;
		
	printk("FXM16APCIe: Launching card: %d\n", wc->serialno);

	if (zt_register(&wc->span, 0))
	{
		printk("Unable to register span with zaptel\n");
		return -1;
	}
	return 0;


}


static struct dspbindata *shfxm_open_dspbinfile(const char *filename)
{
	char fn[256];
	size_t read_len;
	struct file *fp = NULL;
	struct dspbindata *bin;
	unsigned long fileoffset;
	void *data;

	if (filename[0] == '/')
	{
		strncpy(fn, filename, sizeof(fn)-1);
	}
	else
	{
		snprintf(fn, sizeof(fn), "%s/%s", BINFILE_DIR, filename);
	}

	if (!(fp = filp_open(fn, O_RDONLY, S_IRUSR)) || !IS_ERR(fp))
	{
		bin = (struct dspbindata *)kmalloc(sizeof(struct dspbindata), GFP_KERNEL);	
		if (!bin)
		{ 
			printk("Memory Allocation Failure in  in function %s at line %d of %s\n",
								 __FILE__, __LINE__, __PRETTY_FUNCTION__);
			filp_close(fp , NULL);
			return NULL;	
		}

		fileoffset = fp->f_dentry->d_inode->i_size;
		if (fileoffset < 1)
		{
			printk("file offset = NULL %s at line %d of %s\n",
						__FILE__, __LINE__, __PRETTY_FUNCTION__);
			filp_close(fp , NULL);
			kfree(bin);
			return NULL;	
		}

		data = kmalloc(fileoffset, GFP_KERNEL);	
		if (!data)
		{ 
			printk("Memory Allocation Failure in  in function %s at line %d of %s\n",
								 __FILE__, __LINE__, __PRETTY_FUNCTION__);
			filp_close(fp , NULL);
			kfree(bin);
			return NULL;	
		}
			
		read_len = kernel_read(fp, 0, data, fileoffset);
		if (read_len != fileoffset)
		{
			printk("read_len != fileoffset in  in function %s at line %d of %s\n",  
								 __FILE__, __LINE__, __PRETTY_FUNCTION__);
			filp_close(fp , NULL);
			kfree(bin);
			kfree(data);
			return NULL;	
		}
		
		strncpy(bin->name, fn, sizeof(bin->name)-1);
		bin->dsp_data = data;
		filp_close(fp , NULL);
		return bin;


	}
	else
	{
		printk("No file to parse: %s\n", fn);
		return NULL;	
	}
}

static void shfxm_close_dspbinfile(struct dspbindata *data)
{
	if (!data)
	{
		if (!data->dsp_data)
			kfree(data->dsp_data);
		kfree(data);
	}
}


static int shfxm_init_dsp(struct shfxm *wc, const char *filename)
{
	binheader *h;
	binblock  *b;
	unsigned short i;
	unsigned short j;
	unsigned short val;
	unsigned short *p;
	unsigned short size;
	unsigned short *compare;
	struct dspbindata *bin_data;

	bin_data = shfxm_open_dspbinfile(filename);

	if (!bin_data)
	{
		printk("shfxm open dspbinfile fail\n");
		return -1;
	}

	h = (binheader *)bin_data->dsp_data;

	if (h->bin_type != 0x00fe)
	{
		printk("Load Dsp bin Failure in  in function %s at line %d of %s\n",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
		shfxm_close_dspbinfile(bin_data);
		return -1;

	}

	b = (binblock *)h->data;
	
	for (i = 0; i < h->block_num; i++)
	{
		if (b->block_size % 2)
			size = b->block_size + 1;
		else 
			size = b->block_size;
		p = b->data;
		if (size)
		{
			shfxm_setdsp_data(wc, 1, b->block_addr, p, size/2);
			p = b->data;
			compare = (unsigned short *)kmalloc(size, GFP_KERNEL);
			memset(compare, 0x0, size);

			if (!compare)
			{
				printk("Load Dsp bin Failure in  in function %s at line %d of %s\n",
						__FILE__, __LINE__, __PRETTY_FUNCTION__);
				shfxm_close_dspbinfile(bin_data);
				return -1;
			}
			shfxm_getdsp_data(wc, 1, b->block_addr, compare, size/2);
			for (j = 0; j < (size / 2); j++)
			{
				if (*p++ != *(compare + j))
				{
					printk("Load Dsp bin Failure in  in function %s at line %d of %s\n",
							__FILE__, __LINE__, __PRETTY_FUNCTION__);
					kfree(compare);
					shfxm_close_dspbinfile(bin_data);
					return -1;
				}
			}
			kfree(compare);
		}
		p = b->data + size/2;
		b = (binblock *)p;

	}
	shfxm_close_dspbinfile(bin_data);

	val = 0x0;
	shfxm_setdsp_data(wc, 1, 0x03d0 * 2, &val, sizeof(val) / sizeof(unsigned short));
	shfxm_setdsp_data(wc, 1, 0x61 * 2, &h->start_addr, sizeof(h->start_addr) / sizeof(unsigned short));
	val = 0x1000;
	shfxm_setdsp_data(wc, 1, 0x60 * 2, &val, sizeof(val) / sizeof(unsigned short));
	shfxm_wait_ms(500);
	shfxm_getdsp_data(wc, 1, 0x03d0 * 2, &val, sizeof(val) / sizeof(unsigned short));
	if (val != 0x1)
	{
		printk("Load Dsp bin Failure in  in function %s at line %d of %s\n",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
		return -1;
	}

	return 0;

}

static int shfxm_hardware_init_2(struct shfxm *wc)
{
	unsigned short val;
	unsigned int chflag;
	int x;

	shfxm_disable_pci_interrupts(wc);
	shfxm_disable_local_interrupts(wc);

	if (shfxm_init_dsp(wc, "fxm16apcie.bin"))
	{
		printk("shfxm: init dsp fail\n");
		return -1;
	}

	shfxm_getdsp_data(wc, 0, 0x0a, &val, sizeof(val) / sizeof(unsigned short));

	chflag = 0x0;
	chflag |= val;
	chflag <<= 16;
	shfxm_getdsp_data(wc, 0, 0x08, &val, sizeof(val) / sizeof(unsigned short));
	chflag |= val;

	for (x = 0; x < NUM_CARDS; x++)
	{
		int chtype;

		chtype = (chflag >> (x * 2)) & 0x3;
		switch (chtype)
		{
			case 0:
				wc->cardflag |= (1 << x);
				wc->modtype[x] = MOD_TYPE_FXO;
				printk("Module %d: Installed -- AUTO FXO\n",x);
				break;
			case 2:
				wc->cardflag |= (1 << x);
				printk("Module %d: Installed -- AUTO FXS\n",x);
				break;
			case 3:
				printk("Module %d: Not installed\n",x);
				break;
			default:
				printk("Module %d: Not unknow type\n",x);
				break;
		}
	}

	for (x = 0; x < NUM_CARDS; x++)
	{
		if (wc->cardflag & (1 << x))
		{
			if (alawoverride) 
				val = 0x1011;
			else 
				val = 0x1022;
			shfxm_setdsp_data(wc, 1, (SET_RPCW + x) * 2,  &val, sizeof(val) / sizeof(unsigned short));
			val = 0x10 << 8;
			shfxm_setdsp_data(wc, 1, (SET_M3 + x) * 2,  &val, sizeof(val) / sizeof(unsigned short));

		}
	}



	return 0;
}


static int shfxm_echocan(struct zt_chan *chan, int ecval)
{
	unsigned int ch;
	struct shfxm *wc = chan->pvt;
	
	ch = chan->chanpos - 1;

	if (ecval)
		wc->echocan |= 1 << ch;
	else
		wc->echocan &= ~(1 << ch);

	if (wc->modtype[ch] == MOD_TYPE_FXO)
		wc->mod[ch].fxo.flags |= FLAG_DO_ECHOCAN;
	else 
		wc->mod[ch].fxs.flags |= FLAG_DO_ECHOCAN;

	return 0;
}


static int shfxm_ioctl(struct zt_chan *chan, unsigned int cmd, unsigned long data)
{
	switch (cmd)
	{
		case ZT_ONHOOKTRANSFER:
			if(debug)
				printk("shfxm_ioctl---1----------->\n");
			break;
		case ZT_SETPOLARITY:
			if(debug)
				printk("shfxm_ioctl---2----------->\n");
			break;
#if 0
		case WCTDM_GET_STATS:
			printk("shfxm_ioctl---3----------->\n");
			break;
		case WCTDM_GET_REGS:
			printk("shfxm_ioctl---4----------->\n");
			break;
		case WCTDM_SET_REG:
			printk("shfxm_ioctl---5----------->\n");
			break;
		case WCTDM_SET_ECHOTUNE:
			printk("shfxm_ioctl---6----------->\n");
			break;
#endif
		default:
			return -ENOTTY;
	}
	return 0;
}

static int shfxm_hooksig(struct zt_chan *chan, zt_txsig_t txsig)
{
	struct shfxm *wc = chan->pvt;

	if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXO) {
		/* XXX Enable hooksig for FXO XXX */
		switch(txsig) {
		case ZT_TXSIG_START:
		case ZT_TXSIG_OFFHOOK:
			wc->mod[chan->chanpos - 1].fxo.offhook = 1;
			wc->mod[chan->chanpos - 1].fxo.flags |= FLAG_DO_OFFHOOK;
			break;
		case ZT_TXSIG_ONHOOK:
			wc->mod[chan->chanpos - 1].fxo.offhook = 0;
			wc->mod[chan->chanpos - 1].fxo.flags |= FLAG_DO_OFFHOOK;
			break;
		default:
			if(debug)
				printk("wcfxo: Can't set tx state to %d\n", txsig);
		}
	}
	else
	{
		switch(txsig)
		{
			case ZT_TXSIG_ONHOOK:
				switch(chan->sig)
				{
					case ZT_SIG_EM:
					case ZT_SIG_FXOKS:
					case ZT_SIG_FXOLS:
					case ZT_SIG_FXOGS:
						wc->mod[chan->chanpos - 1].fxs.offhook = 0;
						wc->mod[chan->chanpos - 1].fxs.flags |= FLAG_DO_SENDRING;
						break;
				}
				break;
			case ZT_TXSIG_OFFHOOK:
				switch(chan->sig)
				{
					case ZT_SIG_EM:
						wc->mod[chan->chanpos - 1].fxs.offhook = 0;
						wc->mod[chan->chanpos - 1].fxs.flags |= FLAG_DO_SENDRING;
						break;
					default:
						wc->mod[chan->chanpos - 1].fxs.offhook = 0;
						wc->mod[chan->chanpos - 1].fxs.flags |= FLAG_DO_SENDRING;
						break;
				}
				break;
			case ZT_TXSIG_START:
				wc->mod[chan->chanpos - 1].fxs.offhook = 1;
				wc->mod[chan->chanpos - 1].fxs.flags |= FLAG_DO_SENDRING;
				break;
			case ZT_TXSIG_KEWL:
				break;
			default:
				if(debug)
					printk("wctdm: Can't set tx state to %d\n", txsig);
		}
	}
	return 0;

}

static void shfxm_release(struct shfxm *wc)
{
	zt_unregister(&wc->span);
	if (wc->freeregion_plx)
		release_mem_region(wc->plx_region, wc->plx_len);
	if (wc->freeregion_dsp)
		release_mem_region(wc->dsp_region, wc->dsp_len);
	if (wc->plx)
		iounmap(wc->plx);
	if (wc->dsp)
		iounmap(wc->dsp);
	if (wc) 
		kfree(wc);
	printk("Freed a FXM-16a/PCI\n");

}

static int shfxm_close(struct zt_chan *chan)
{
	struct shfxm *wc = chan->pvt;

	wc->usecount--;

	module_put(THIS_MODULE);

	/* If we're dead, release us now */
	if (!wc->usecount && wc->dead) 
		shfxm_release(wc);
	return 0;
}

static int shfxm_open(struct zt_chan *chan)
{

	struct shfxm *wc = chan->pvt;

	if (!(wc->cardflag & (1 << (chan->chanpos - 1))))
		return -ENODEV;
	if (wc->dead)
		return -ENODEV;
	wc->usecount++;

	try_module_get(THIS_MODULE);

	return 0;
}


static int shfxm_init_span(struct shfxm *wc)
{
	int x;

	sprintf(wc->span.name, "FXM16APCIE/%d", wc->pos);
	sprintf(wc->span.desc, "%s Board %d", wc->variety, wc->pos + 1);
	if (alawoverride)
	{
		printk("ALAW override parameter detected.  Device will be operating in ALAW\n");
		wc->span.deflaw = ZT_LAW_ALAW;
	}
	else
		wc->span.deflaw = ZT_LAW_MULAW;
	for (x = 0; x < NUM_CARDS; x++)
	{
		sprintf(wc->chans[x].name, "FXM16APCIE/%d/%d", wc->pos, x);
		wc->chans[x].sigcap = 0;
		wc->chans[x].chanpos = x+1;
		wc->chans[x].pvt = wc;
	}
	wc->span.chans = wc->chans;
	wc->span.channels = NUM_CARDS;
	wc->span.flags = ZT_FLAG_RBS;
	wc->span.hooksig = shfxm_hooksig;
	wc->span.open = shfxm_open;
	wc->span.close = shfxm_close;
	wc->span.ioctl = shfxm_ioctl;
	wc->span.echocan = shfxm_echocan;
#if 0
	wc->span.watchdog = wctdm_watchdog;
#endif
	init_waitqueue_head(&wc->span.maintq);

	wc->span.pvt = wc;

	return 0;
}

static void shfxm_card_reset(struct shfxm *wc)
{
	unsigned short reg;

	reg = 0x0;
	shfxm_setdsp_data(wc, 0, 0x00, &reg, sizeof(reg) / sizeof(unsigned short));
	shfxm_wait_ms(10);
	reg = 0x1;
	shfxm_setdsp_data(wc, 0, 0x00, &reg, sizeof(reg) / sizeof(unsigned short));
	shfxm_wait_ms(10);

}

static unsigned long plx_chip_type_get(struct shfxm *wc)
{
	unsigned long val;
	unsigned long type;
	unsigned long reg_original;

	reg_original = plx_getreg(wc, 0x04);
	plx_setreg(wc, 0x4, reg_original | (0xff << 24));
	val = plx_getreg(wc, 0x4);
	plx_setreg(wc, 0x4, reg_original);

	if ((val >> 24) == 0)
	{
		printk("shfxm: unkown chip type\n");
		return -1;
	}

	if ((val >> 28) == 0xf)
	{
		val = plx_getreg(wc, 0x70);
		if (val == 0x905610b5)
		{
			type = 0x9056;
			return type;
		}
	}
	return -1;

}

static int shfxm_plx_init(struct shfxm *wc)
{
	u8 mu_enabled;
	u8 eeprompresent;
	u32 regvalue;
	u32 reginterrupt;
	u32 reghotswap;
	u32 regpowermgmnt;

	wc->dsp_addr = 0x1;
	plx_setreg(wc, 0x4, wc->dsp_addr);

	if (plx_chip_type_get(wc) != 0x9056)
	{
		return -1;
	}

	pci_getconfig_dword(wc, 0x4, &regvalue);

	if (regvalue & (0xf8 << 24))
	{
		pci_setconfig_dword(wc, 0x4, regvalue);
	}

	regvalue = plx_getreg(wc, PCI9056_FIFO_CTRL_STAT);
	mu_enabled = (unsigned char)(regvalue & (1 << 0));
	regvalue = plx_getreg(wc, PCI9056_EEPROM_CTRL_STAT);

	regvalue &= ~((1 << 30) | (1 << 29));
	eeprompresent = (unsigned char)((regvalue >> 28) & (1 << 0));

	pci_getconfig_dword(wc, 0x3c, &reginterrupt);

	if (eeprompresent)
	{
		pci_getconfig_dword(wc, 0x48, &reghotswap);
		pci_getconfig_dword(wc, 0x44, &regpowermgmnt);
	}


	plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue | (1 << 30));
	shfxm_wait_ms(100);
	plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue);
	plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue | (1 << 29));
	shfxm_wait_ms(10);
	plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue);
	if (mu_enabled)
	{
		regvalue = plx_getreg(wc, PCI9056_FIFO_CTRL_STAT);
		plx_setreg(wc, PCI9056_FIFO_CTRL_STAT, regvalue | (1 << 0));
	}

	pci_setconfig_dword(wc, 0x3c, reginterrupt);

	if (eeprompresent)
	{
		reghotswap &= ~((1 << 23) | (1 << 22) | (1 << 17));
		pci_setconfig_dword(wc, 0x48, reghotswap);
		regpowermgmnt &= ~(1 << 15);
		pci_getconfig_dword(wc, 0x44, &regpowermgmnt);
	}
	

	return 0;
}

static int shfxm_hardware_init_1(struct shfxm *wc)
{
	unsigned short val;


	if (shfxm_plx_init(wc))
	{
		printk("shfxm: plx chip type is not 9056\n");
		return -1;
	}

	shfxm_card_reset(wc);

	val = 0x101;
	shfxm_setdsp_data(wc, 0, 0x18, &val, sizeof(val) / sizeof(unsigned short));
	shfxm_getdsp_data(wc, 0, 0x18, &val, sizeof(val) / sizeof(unsigned short));
	if (val != 0x109)
	{
		printk("fxm16apcie 0x18 = %x\n", val);
		return -1;
	}
	
	val = 0xff;
	shfxm_setdsp_data(wc, 0, 0x10, &val, sizeof(val) / sizeof(unsigned short));
	shfxm_setdsp_data(wc, 0, 0x12, &val, sizeof(val) / sizeof(unsigned short));
	
	return 0;
}


static void shfxm_post_initialize(struct shfxm *wc)
{
	int x;
	/* Finalize signalling  */
	for (x = 0; x < NUM_CARDS; x++)
	{
		if (wc->cardflag & (1 << x))
		{
			if (wc->modtype[x] == MOD_TYPE_FXO)
			{
				wc->chans[x].sigcap = ZT_SIG_FXSKS | ZT_SIG_FXSLS | ZT_SIG_SF | ZT_SIG_CLEAR;
				//wc->mod[x].fxo.regctrl0 = 0xf;
			} else
				wc->chans[x].sigcap = ZT_SIG_FXOKS | ZT_SIG_FXOLS | ZT_SIG_FXOGS | ZT_SIG_SF | ZT_SIG_EM | ZT_SIG_CLEAR;
		}
	}
}

static void *plx_dmabuf_alloc(struct shfxm *wc, struct plx_phys_mem_object *memobj)
{
	dma_addr_t busaddress;
	unsigned long virt_addr;

	memobj->kernelva = dma_alloc_coherent(&(wc->dev->dev), memobj->size, &busaddress, GFP_KERNEL | __GFP_NOWARN);

	if (!memobj->kernelva)
		return NULL;

	memobj->busphysical = (u64)busaddress;

	for (virt_addr = (unsigned long)memobj->kernelva; virt_addr < ((unsigned long)memobj->kernelva + memobj->size); virt_addr += PAGE_SIZE)
	{
		SetPageReserved(virt_to_page(virt_addr));
	}

	memobj->cpuphysical = virt_to_phys(memobj->kernelva);
	if (debug)
		printk("cpu phys addr: 0x%08lx, bus phys addr: 0x%08lx, kernel VA : 0x%p\n", (unsigned long)memobj->cpuphysical, (unsigned long)memobj->busphysical, memobj->kernelva);

	if (memobj->size >= (10 << 10))
	{
		if (debug)
			printk("(%d kb)\n", memobj->size >> 10);
	}
	else
	{
		if (debug)
			printk("(%d bytes)\n", memobj->size);
	}

	return memobj->kernelva;
}

static void plx_dmabuf_free(struct shfxm *wc, struct plx_phys_mem_object *memobj)
{
	unsigned long virt_addr;

	for (virt_addr = (unsigned long)memobj->kernelva; virt_addr < ((unsigned long)memobj->kernelva + PAGE_ALIGN(memobj->size)); virt_addr += PAGE_SIZE)
	{
		ClearPageReserved(virt_to_page(virt_addr));
	}
	dma_free_coherent(&(wc->dev->dev), memobj->size, memobj->kernelva, (int)memobj->busphysical);
	if (debug)
		printk("released physical memory at 0x%08lx\n", (unsigned long) memobj->cpuphysical);

	if (memobj->size >= (10 << 10))
	{
		if (debug)
			printk("(%d kb)\n", memobj->size >> 10);
	}
	else
	{
		if (debug)
			printk("(%d bytes)\n", memobj->size);
	}

	memset(memobj, 0x0, sizeof(unsigned long));

}

static int plx_dmasgl_transfer_listadd(struct shfxm *wc, DMA_CHANNEL channel, struct dma_transfer_element *dmadata)
{
	u8 i;
	u8 shift;
	u32 regvalue;

	switch (channel)
	{
		case DMA_CHANNEL0:
			i = 0;
			shift = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			shift = 8;
			break;
		default:
			return -1;

	}

	regvalue = plx_getreg(wc, PCI9056_DMA_COMMAND_STAT);
	if ((regvalue & ((1 << 4) << shift)) == 0)
	{
		if(debug > 66)
			printk("listadd: dma channel is carrently active\n");
		return -1;
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
	{
		if(debug > 66)
			printk("listadd: dma channel has not been opened for SGL DMA\n");
		return -1;
	}

	wc->dmainfo[i].dmadata[wc->dmainfo[i].dmadatacount] = *dmadata;
	
	if(wc->dmainfo[i].dmadatacount < PLAYBLOCK_NUM)
		wc->dmainfo[i].dmadatacount++;
	else
	{
		if(debug > 66)
			printk("DMA LIST IS FULL\n");
	}

	return 0;
}

static int plx_dmasgl_lockbuffer_and_build(struct shfxm *wc, u8 dmaindex)
{
	u32 k = 0;
	u32 bussgl;
	u32 bussgloriginal;
	u32 sglsize;
	u32 totalpages;
	unsigned long vasgl;
	struct dma_transfer_element *dma;

	totalpages = wc->dmainfo[dmaindex].dmadatacount;

	sglsize = (totalpages * (4 * sizeof(u32))) + 16;

	if (wc->dmainfo[dmaindex].sglbuffer.kernelva)
	{
		if (wc->dmainfo[dmaindex].sglbuffer.size < sglsize)
		{
			printk("err:sglbuffer.size < sglsize return line %d\n",__LINE__);
			return 0;
		}
	}

	if (!wc->dmainfo[dmaindex].sglbuffer.kernelva)
	{
		printk("lockbuf: unable to allocate %d bytes for %d SGL descriptors\n", wc->dmainfo[dmaindex].sglbuffer.size, totalpages);
		return 0;
	}
	else
	{
		vasgl = (unsigned long)wc->dmainfo[dmaindex].sglbuffer.kernelva;
	}


	bussgl = (u32)wc->dmainfo[dmaindex].sglbuffer.busphysical;

	vasgl = (vasgl + 0xf) & ~(0xf);

	bussgl = (bussgl + 0xf) & ~(0xf);

	bussgloriginal = bussgl;


	for (k = 0; k < wc->dmainfo[dmaindex].dmadatacount; k++)
	{
		dma = &wc->dmainfo[dmaindex].dmadata[k];

		totalpages--;
		*(((u32*)vasgl) + 0) = dma->u.pciaddrlow;
		*(((u32*)vasgl) + 1) = dma->localaddr;
		*(((u32*)vasgl) + 2) = dma->transfercount;
		if (totalpages == 0)
		{
			*(((u32*)vasgl) + 3) = (dma->localtopcidma << 3) | (1 << 1) | (1 << 0);
		}
		else
		{
			bussgl += (4 * sizeof(u32));
			*(((u32*)vasgl) + 3) = bussgl | (dma->localtopcidma << 3) | (1 << 0);
			vasgl +=(4 * sizeof(u32));
		}

	}
	return bussgloriginal;
}

static int plx_dmasgl_transfer(struct shfxm *wc, DMA_CHANNEL channel)
{
	u8 i;
	u8 shift;
	u16 offsetdesc;
	u32 regvalue;
	u32 sglpciaddress;

	switch (channel) {
		case DMA_CHANNEL0:
			i = 0;
			shift = 0;
			offsetdesc = PCI9056_DMA0_DESC_PTR;
			break;
		case DMA_CHANNEL1:
			i = 1;
			shift = 8;
			offsetdesc = PCI9056_DMA1_DESC_PTR;
			break;
		default:
			return -1;

	}

	regvalue = plx_getreg(wc, PCI9056_DMA_COMMAND_STAT);
	if ((regvalue & ((1 << 4) << shift)) == 0)
		return -1;

	if (wc->dmainfo[i].state != DMA_STATE_SGL)
		return -1;

	sglpciaddress = plx_dmasgl_lockbuffer_and_build(wc, i);
	
	if (sglpciaddress == 0)
	{
		if(debug > 66)
			printk("unable to lock buffer and build SGL list\n");
		return -1;
	}

	plx_setreg(wc, offsetdesc, sglpciaddress | (1 << 0));
	regvalue = plx_getreg(wc, PCI9056_DMA_COMMAND_STAT);
	plx_setreg(wc, PCI9056_DMA_COMMAND_STAT ,regvalue | ((1 << 0) << shift));

	plx_setreg(wc, PCI9056_DMA_COMMAND_STAT, regvalue | (((1 << 0) | (1 << 1)) << shift)); // start dma

	wc->dmainfo[i].timeout = 0;

	return 0;
}

static int plx_dmasgl_channel_open(struct shfxm *wc, DMA_CHANNEL channel, struct dma_channel_desc *desc, void *owner)
{
	u8 i;
	u32 mode;
	u32 regvalue;
	u32 threshold;
	struct plx_reg_data regdata;


	switch (channel) {
		case DMA_CHANNEL0:
			i = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			break;
		default:
			return -1;

	}

	if (wc->dmainfo[i].state != DMA_STATE_CLOSED)
	{
		printk("DMA channel already open\n");
		return -1;
	}

	wc->dmainfo[i].state = DMA_STATE_SGL;
	wc->dmainfo[i].owner = owner;

	regdata.wc = wc;
	regdata.offset = PCI9056_INT_CTRL_STAT;
	regdata.bitstoclear = 0;
	regdata.bitstoset = 0;

	regvalue = plx_getreg(wc, PCI9056_LOCAL_DMA_ARBIT);
	regvalue &= ~((1 << 20) | (1 << 19));

	switch (desc->dmachannelpriority)
	{
		case CHANNEL_0_HIGHEST:
			plx_setreg(wc, PCI9056_LOCAL_DMA_ARBIT, regvalue | (1 << 19));
			break;
		case CHANNEL_1_HIGHEST:
			plx_setreg(wc, PCI9056_LOCAL_DMA_ARBIT, regvalue | (1 << 20));
			break;
		case ROTATIONAL:
			plx_setreg(wc, PCI9056_LOCAL_DMA_ARBIT, regvalue);
			break;
		default:
			printk("dmachannelopen() invalid priority state\n");
			return -1;
	}

	threshold = 
		(desc->tholdforiopwrites << 0) |
		(desc->tholdforiopreads << 4) |
		(desc->tholdforpciwrites << 8) |
		(desc->tholdforpcireads << 12);
	mode =
		(1 << 9) |
		(1 << 10) |
		(1 << 17) |
		(0 << 18) |
		(desc->iopbuswidth << 0) |
		(desc->waitstates << 2) |
		(desc->enablereadyinput << 6) |
		(desc->enablebterminput << 7) |
		(desc->enableiopburst << 8) |
		(desc->holdiopaddrconst << 11) |
		(desc->demandmode << 12) |
		(desc->enablewriteinvalidmode << 13) |
		(desc->enabledmaeoppin << 14) |
		(desc->dmastoptransfermode << 15);
	if (desc->holdiopaddrconst)
		wc->dmainfo[i].localaddrconstant = 1;
	else 
		wc->dmainfo[i].localaddrconstant = 0;

	regvalue = plx_getreg(wc, PCI9056_DMA_THRESHOLD);
	if (channel == DMA_CHANNEL0) {
		plx_setreg(wc, PCI9056_DMA_THRESHOLD, (regvalue & 0xffff0000) | threshold);
		plx_setreg(wc, PCI9056_DMA0_MODE, mode);
		regdata.bitstoclear = (0 << 18);
		plx_sync_reg_modify(&regdata);
		plx_setreg(wc, PCI9056_DMA0_PCI_DAC, 0);
	} else {
		plx_setreg(wc, PCI9056_DMA_THRESHOLD, (regvalue & 0xffff0000) | (threshold << 16));
		plx_setreg(wc, PCI9056_DMA1_MODE, mode);
		regdata.bitstoclear = (0 << 19);
		plx_sync_reg_modify(&regdata);
		plx_setreg(wc, PCI9056_DMA1_PCI_DAC, 0);
	}


	return 0;
}


static int plx_dmasgl_channel_close(struct shfxm *wc, DMA_CHANNEL channel, unsigned int checkinprogress)
{
	u8 i;
	u32 regvalue;

	switch (channel) {
		case DMA_CHANNEL0:
			i = 0;
			break;
		case DMA_CHANNEL1:
			i = 1;
			break;
		default:
			printk("close:Invalid DMA channel\n");
			return -1;
	}

	if (checkinprogress) {
		regvalue = plx_getreg(wc, PCI9056_DMA_COMMAND_STAT);
		if (channel == DMA_CHANNEL1)
			regvalue = regvalue >> 8;

		if ((regvalue & (1 << 4)) == 0) {
			if (regvalue & (1 << 0)) {
				printk("dma in progress\n");
				return -1;
			} else {
				printk("dma paused\n");
				return -1;
			}
		}
	}

	if (wc->dmainfo[i].state != DMA_STATE_SGL) {
		printk("--DMA channel has not been opened for SGL DMA\n");
		return -1;
	}

	wc->dmainfo[i].state = DMA_STATE_CLOSED;
	wc->dmainfo[i].owner = NULL;

	if (wc->dmainfo[i].sglbuffer.kernelva) {
		if (debug)
			printk("channel close: releasing memory used for SGL descript...\n");
		plx_dmabuf_free(wc, &wc->dmainfo[i].sglbuffer);
	}
	if (wc->dmainfo[i].dspbuffer.kernelva) {
		if (debug)
			printk("channel close: releasing memory used for DSP descript...\n");
		plx_dmabuf_free(wc, &wc->dmainfo[i].dspbuffer);
	}
	return 0;

}

static int shfxm_stop_dma(struct shfxm *wc)
{
	plx_dmasgl_channel_close(wc, DMA_CHANNEL0, 1);
	plx_dmasgl_channel_close(wc, DMA_CHANNEL1, 1);
	return 0;
}

static int shfxm_init_dma(struct shfxm *wc)
{
	int i;
	struct dma_channel_desc desc;

	for (i = 0; i < 2; i++) {
		wc->dmainfo[i].sglbuffer.size = 4096;
		plx_dmabuf_alloc(wc, &wc->dmainfo[i].sglbuffer);
	}

	memset(&desc, 0x0, sizeof(struct dma_channel_desc));
	desc.iopbuswidth = 1;
	desc.waitstates = 0;
	desc.enablereadyinput = 1;
	desc.enablebterminput = 0;
	desc.enableiopburst = 0;
	desc.holdiopaddrconst = 0;
	desc.demandmode = 0;
	desc.enablewriteinvalidmode = 0;
	desc.enabledmaeoppin = 0;
	desc.dmastoptransfermode = 1;
	desc.tholdforpciwrites = 15;
	desc.tholdforpcireads = 15;
	if (plx_dmasgl_channel_open(wc,DMA_CHANNEL0, &desc, wc)) {
		return -1;
	}
	if (plx_dmasgl_channel_open(wc,DMA_CHANNEL1, &desc, wc)) {
		return -1;
	}
	return 0;
}

static void shfxm_init_readandwrite_chunk(struct shfxm *wc)
{
	int i, j, k;

	wc->dmainfo[0].dspbuffer.size = (NUM_CARDS * ZT_CHUNKSIZE + sizeof(unsigned short))*PLAYBLOCK_BUF;
	plx_dmabuf_alloc(wc, &wc->dmainfo[0].dspbuffer);
	wc->dmainfo[1].dspbuffer.size = (NUM_CARDS * ZT_CHUNKSIZE * RECDATA_NUM + sizeof(unsigned short))*RECBLOCK_BUF;
	plx_dmabuf_alloc(wc, &wc->dmainfo[1].dspbuffer);

	for (j = 0; j<PLAYBLOCK_BUF; j++)
	{
		wc->writechunk_frame[j] = (struct dsp_writechunk_frame *)(wc->dmainfo[0].dspbuffer.kernelva+(NUM_CARDS * ZT_CHUNKSIZE + sizeof(unsigned short))*j);
		wc->writechunk_frame[j]->writeflag = 1;
		
		for (i = 0; i < NUM_CARDS; i++)
			wc->writechunk[j][i] = wc->writechunk_frame[j]->writechunkbuf + ZT_CHUNKSIZE * i;
	}
	
	for (j = 0; j<RECBLOCK_BUF; j++)
	{
		wc->readchunk_frame[j] = (struct dsp_readchunk_frame *)(wc->dmainfo[1].dspbuffer.kernelva+(NUM_CARDS * ZT_CHUNKSIZE * RECDATA_NUM + sizeof(unsigned short))*j);
		wc->readchunk_frame[j]->readflag = 1;
		
		for (k = 0; k < RECDATA_NUM; k++)
		{
			for (i = 0; i < NUM_CARDS; i++)
				wc->readchunk[j][k][i] = wc->readchunk_frame[j]->readchunkbuf + NUM_CARDS * ZT_CHUNKSIZE * k + ZT_CHUNKSIZE * i;
		}
	}
}

static void shfxm_SerialNo(struct shfxm *wc)
{
	s8 BitPos, count;
	u32 regvalue, EepromCommand, tmpreg;
	
	regvalue = plx_getreg(wc, PCI9056_EEPROM_CTRL_STAT);
	
	regvalue &= ~(0xF << 24);
	plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue);
	regvalue |= (1 << 25);
	plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue);
	
	EepromCommand = (0x0180) | (0x008/2);
	for (BitPos = 8; BitPos >= 0; BitPos--)
	{
		if (EepromCommand & (1 << BitPos))
			plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue | (1 << 26));
		else
			plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue);
		
		tmpreg = plx_getreg(wc, PCI9056_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, tmpreg | (1 << 24));
		for (count=0; count<20; count++)
			tmpreg = plx_getreg(wc, PCI9056_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, tmpreg & ~(1 << 24));
	}
	
	regvalue = plx_getreg(wc, PCI9056_EEPROM_CTRL_STAT);
	regvalue |= (1 << 31);
	plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue | (1 << 26));

	for (BitPos = 0; BitPos < 32; BitPos++)
	{
		tmpreg = plx_getreg(wc, PCI9056_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, tmpreg | (1 << 24));
		for (count=0; count<20; count++)
			tmpreg = plx_getreg(wc, PCI9056_EEPROM_CTRL_STAT);
		plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, tmpreg & ~(1 << 24));

		for (count=0; count <20; count++)
			regvalue = plx_getreg(wc, PCI9056_EEPROM_CTRL_STAT);

		if (regvalue & (1 << 27))
			wc->serialno = (wc->serialno << 1) | 1;
		else
			wc->serialno = (wc->serialno << 1);
	}
	
	regvalue &= ~(1 << 31);
	plx_setreg(wc, PCI9056_EEPROM_CTRL_STAT, regvalue & ~(0xF << 24));
	
	wc->serialno = (wc->serialno >> 16 | wc->serialno << 16);
}

static int __devinit shfxm_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int x;
	int y;
	int res;
	char qname[20];
	struct shfxm *wc;
	struct shfxm_desc *d = (struct shfxm_desc *)ent->driver_data;

	for (x=0;x<WC_MAX_IFACES;x++)
		if (!cards[x]) break;
	if (x >= WC_MAX_IFACES)
	{
		printk("Too many interfaces\n");
		return -EIO;
	}

	if (pci_enable_device(pdev))
	{
		res = -EIO;
	}
	else
	{
		int cardcount = 0;

		wc = kmalloc(sizeof(struct shfxm), GFP_KERNEL);
		if (!wc) return -ENOMEM;
		memset(wc, 0x0, sizeof(struct shfxm));

		if (pdev->irq < 1) {
			if (wc) 
				kfree(wc);
			printk(KERN_ERR "No IRQ allocated for device\n");
			return -ENODEV;

		}

		spin_lock_init(&wc->lock);
		cards[x] = wc;
		wc->curcard = -1;
		wc->irq_enable = 0;
		wc->dev = pdev;
		wc->pos = x;
		wc->variety = d->name;
		wc->Ttflags = 0;

		wc->RECFlag = 0;
		wc->TRIFlag = 0;
		wc->RECFlag_First = 1;
		wc->TRIFlag_First = 1;
		wc->RECCount = 0;
		wc->HasToRec = 0;

		for (y=0;y<NUM_CARDS;y++)
			wc->flags[y] = d->flags;
		
		wc->readcurpos	= 0;
		wc->readfillpos	= 0;
		wc->rpos				= 0;
		wc->writecurpos	= 0;
		wc->writefillpos	= 0;

		for(x = 0; x < PLAYBLOCK_BUF; x++)
			wc->writehasdata[x] = 0;

		for(x = 0; x < RECBLOCK_BUF; x++)
			wc->readhasdata[x] = 0;

		wc->plx_region = pci_resource_start(pdev,0);
		wc->plx_len = pci_resource_len(pdev,0);

		if (check_mem_region(wc->plx_region, wc->plx_len))
		{
			printk("plx region %lx-%lx already in use\n", wc->plx_region, wc->plx_region + wc->plx_len);
			if (wc) 
				kfree(wc);
			return -EIO;
		}
		
		wc->dsp_region = pci_resource_start(pdev,2);
		wc->dsp_len = pci_resource_len(pdev,2);


		if (check_mem_region(wc->dsp_region, wc->dsp_len))
		{
			printk("dsp region %lx-%lx already in use\n", wc->dsp_region, wc->dsp_region + wc->dsp_len);
			if (wc) 
				kfree(wc);
			return -EIO;
		}

		if (request_mem_region(wc->plx_region, wc->plx_len, "shtej"))
		{
			wc->freeregion_plx = 1;
		}

		wc->plx = ioremap(wc->plx_region, wc->plx_len);
		if (!wc->plx)
		{
			printk("Cannot ioremap memory space at 0x%lx\n", wc->plx_region);
			goto err_out_free_wc;
		}

		if (request_mem_region(wc->dsp_region, wc->dsp_len, "shtej"))
		{
			wc->freeregion_dsp = 1;
		}

		wc->dsp = ioremap(wc->dsp_region, wc->dsp_len);
		if (!wc->dsp)
		{
			printk("Cannot ioremap memory space at 0x%lx\n", wc->dsp_region);
			goto err_out_free_wc;
		}

		if (shfxm_hardware_init_1(wc))
		{
			printk("shfxm: hardware init 1 fail\n");
			goto err_out_free_wc;
		}

		if (shfxm_init_dma(wc))
		{
			printk("shfxm: init dma fail\n");
			goto err_out_free_wc;
		}

		pci_set_master(pdev);

		pci_set_drvdata(pdev, wc);

		shfxm_init_span(wc);

		shfxm_init_readandwrite_chunk(wc);

		if (request_irq(pdev->irq, shfxm_interrupt, ZAP_IRQ_SHARED_DISABLED, "shfxm", wc))
		{
			printk("shfxm: Unable to request IRQ %d\n", pdev->irq);
			pci_set_drvdata(pdev, NULL);
			goto err_out_free_wc;
		}

		if (shfxm_hardware_init_2(wc))
		{
			free_irq(pdev->irq, wc);
			pci_set_drvdata(pdev, NULL);
			printk("shfxm: hardware init 2 fail\n");
			goto err_out_free_wc;
		}
		
		wc->BTime = 1;
		wc->complete = 1;
		wc->Index = 0;
		wc->filecount = 0;
		for(x=0; x<NUM_CARDS; x++)
		{			
			wc->tmpwrite[0][x] = (unsigned char *)kmalloc(0x20000, GFP_KERNEL);
			wc->tmpwrite[1][x] = (unsigned char *)kmalloc(0x20000, GFP_KERNEL);
			if(wc->tmpwrite[0][x] == NULL || wc->tmpwrite[1][x] == NULL)
			{
				goto err_out_free_wc;
			}
			for(y=0; y<RECDATA_NUM; y++)
			{
				wc->tmpwriter[0][y][x] = (unsigned char *)kmalloc(0x20000, GFP_KERNEL);
				wc->tmpwriter[1][y][x] = (unsigned char *)kmalloc(0x20000, GFP_KERNEL);
				if(wc->tmpwriter[0][y][x] == NULL || wc->tmpwriter[1][y][x] == NULL)
				{
					goto err_out_free_wc;
				}
			}
		}
		init_waitqueue_head(&wc->regq);
		/* Initialize the work queue */
		sprintf(qname, "fxm8apci%d", wc->pos);
		wc->shfxm_wq = create_singlethread_workqueue(qname);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
		INIT_WORK(&wc->shfxm_work, wque_run);
#else
		INIT_WORK(&wc->shfxm_work, wque_run, wc);
#endif

		shfxm_SerialNo(wc);

		
		shfxm_clear_cpld_interrupts(wc);
		


		shfxm_post_initialize(wc);
		
		for (x = 0; cards[x]; x++)
		{
			if (shfxm_launch(cards[x]))
			{
				free_irq(pdev->irq, wc);
				pci_set_drvdata(pdev, NULL);
				printk("shfxm: launch cards fail\n");
				goto err_out_free_wc;
			}
		}
		shfxm_enable_local_interrupts(wc);
		shfxm_enable_pci_interrupts(wc);
		wc->irq_enable = 1;
		for (x = 0; x < NUM_CARDS; x++)
		{
			if (wc->cardflag & (1 << x))
				cardcount++;
		}

		printk("Found a TDM: %s (%d modules)\n", wc->variety, cardcount);
		res = 0;

	}

	return res;

err_out_free_wc:
	cards[wc->pos] = NULL;
	if (wc->freeregion_plx)
		release_mem_region(wc->plx_region, wc->plx_len);
	if (wc->freeregion_dsp)
		release_mem_region(wc->dsp_region, wc->dsp_len);
	if (wc->plx)
		iounmap(wc->plx);
	if (wc->dsp)
		iounmap(wc->dsp);
	if (wc) 
		kfree(wc);
	return -EIO;

}


static void __devexit shfxm_remove_one(struct pci_dev *pdev)
{
	struct shfxm *wc = pci_get_drvdata(pdev);
	unsigned short reg;
	int x, j;
	
	if (wc)
	{
		reg = 0x0;
		wc->irq_enable = 0;
		shfxm_stop_dma(wc);
		shfxm_wait_ms(5000);
		shfxm_disable_pci_interrupts(wc);
		shfxm_disable_local_interrupts(wc);
		shfxm_setdsp_data(wc, 0, 0x00, &reg, sizeof(reg) / sizeof(unsigned short));
		free_irq(wc->dev->irq, wc);
		cards[wc->pos] = NULL;
		
		destroy_workqueue(wc->shfxm_wq);
		for(x=0; x<NUM_CARDS; x++)
		{
			kfree(wc->tmpwrite[0][x]);
			kfree(wc->tmpwrite[1][x]);
			for (j = 0; j < RECDATA_NUM; j++)
			{
				kfree(wc->tmpwriter[0][j][x]);
				kfree(wc->tmpwriter[1][j][x]);
			}
		}

		if (!wc->usecount)
			shfxm_release(wc);
		else
			wc->dead = 1;
	}
}

static struct pci_device_id shfxm_pci_tbl[] = {

	{ 0x10b5, 0x00fe, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (unsigned long) &shfxm},
	{ 0 }
};

MODULE_DEVICE_TABLE(pci, shfxm_pci_tbl);

static struct pci_driver shfxm_driver = {
	name: 	"shfxm16apcie",
	probe: 	shfxm_init_one,
#ifdef LINUX26
	remove:	__devexit_p(shfxm_remove_one),
#else
	remove:	shfxm_remove_one,
#endif
	suspend: NULL,
	resume:	NULL,
	id_table: shfxm_pci_tbl,
};

static int __init shfxm_init(void)
{
	int res;
	

	res = zap_pci_module(&shfxm_driver);
	if (res)
		return -ENODEV;
	return 0;
}

static void __exit shfxm_cleanup(void)
{
	pci_unregister_driver(&shfxm_driver);
}


module_param(debug, int, 0600);
module_param(battthresh, int, 0600);
module_param(alawoverride, int, 0600);
module_param(ECHO_OPEN, int, 0600);

MODULE_DESCRIPTION("FXM-16a/PCI Driver");
MODULE_AUTHOR("Chen Jindong <cjd@sanhuid.com>");
#if defined(MODULE_ALIAS)
MODULE_ALIAS("shfxm");
#endif
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

module_init(shfxm_init);
module_exit(shfxm_cleanup);
